/* Generated by Cython 0.29.21 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "define_macros": [
            [
                "UNICODE",
                "1"
            ],
            [
                "_UNICODE",
                "1"
            ],
            [
                "SECURITY_WIN32",
                "1"
            ]
        ],
        "depends": [],
        "libraries": [
            "Secur32"
        ],
        "name": "spnego._sspi_raw.sspi",
        "sources": [
            "spnego\\_sspi_raw\\sspi.pyx"
        ]
    },
    "module_name": "spnego._sspi_raw.sspi"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__spnego___sspi_raw__sspi
#define __PYX_HAVE_API__spnego___sspi_raw__sspi
/* Early includes */
#include <stddef.h>
#include "Windows.h"
#include "Security.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "stringsource",
  "spnego\\_sspi_raw\\sspi.pyx",
  "spnego\\_sspi_raw\\sspi.pxd",
  "spnego\\_sspi_raw\\text.pxd",
  "type.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar;
struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential;
struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext;
struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc;
struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer;
struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase;
struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity;
struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__;

/* "spnego/_sspi_raw/text.pxd":9
 * 
 * 
 * cdef class WideChar:             # <<<<<<<<<<<<<<
 *     cdef LPWSTR buffer
 *     cdef int length
 */
struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar {
  PyObject_HEAD
  LPWSTR buffer;
  int length;
};


/* "spnego/_sspi_raw/sspi.pxd":19
 * 
 * 
 * cdef class Credential:             # <<<<<<<<<<<<<<
 * 
 *     cdef CredHandle handle
 */
struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential {
  PyObject_HEAD
  CredHandle handle;
  unsigned PY_LONG_LONG expiry;
};


/* "spnego/_sspi_raw/sspi.pxd":25
 * 
 * 
 * cdef class SecurityContext:             # <<<<<<<<<<<<<<
 * 
 *     cdef CtxtHandle handle
 */
struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext {
  PyObject_HEAD
  CtxtHandle handle;
  unsigned long context_attr;
  unsigned PY_LONG_LONG expiry;
  unsigned long init;
};


/* "spnego/_sspi_raw/sspi.pxd":33
 * 
 * 
 * cdef class SecBufferDesc:             # <<<<<<<<<<<<<<
 *     cdef NativeSecBufferDesc _c_value
 *     cdef list _buffers
 */
struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc {
  PyObject_HEAD
  struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_vtab;
  SecBufferDesc _c_value;
  PyObject *_buffers;
};


/* "spnego/_sspi_raw/sspi.pxd":40
 * 
 * 
 * cdef class SecBuffer:             # <<<<<<<<<<<<<<
 *     cdef PSecBuffer c_value
 *     cdef object sys_alloc
 */
struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer {
  PyObject_HEAD
  struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_vtab;
  PSecBuffer c_value;
  PyObject *sys_alloc;
  PyObject *_is_owner;
  void *_p_buffer;
};


/* "spnego/_sspi_raw/sspi.pxd":49
 * 
 * 
 * cdef class _AuthIdentityBase:             # <<<<<<<<<<<<<<
 * 
 *     cdef void *__c_value__(_AuthIdentityBase self)
 */
struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_vtab;
};


/* "spnego/_sspi_raw/sspi.pxd":54
 * 
 * 
 * cdef class WinNTAuthIdentity(_AuthIdentityBase):             # <<<<<<<<<<<<<<
 * 
 *     cdef WideChar _username
 */
struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase __pyx_base;
  struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *_username;
  struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *_domain;
  struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *_password;
  SEC_WINNT_AUTH_IDENTITY_W c_value;
};


/* "spnego/_sspi_raw/sspi.pyx":433
 *         value.replace_ptr(&self._c_value.pBuffers[key])
 * 
 *     def __iter__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         for val in self._buffers:
 *             yield val
 */
struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ {
  PyObject_HEAD
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self;
  PyObject *__pyx_v_val;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};



/* "spnego/_sspi_raw/sspi.pyx":407
 * 
 * 
 * cdef class SecBufferDesc:             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(SecBufferDesc self, list buffers not None, unsigned long version=SECBUFFER_VERSION):
 */

struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc {
  PSecBufferDesc (*__pyx___c_value__)(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *);
};
static struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_vtabptr_6spnego_9_sspi_raw_4sspi_SecBufferDesc;


/* "spnego/_sspi_raw/sspi.pyx":459
 * 
 * 
 * cdef class SecBuffer:             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(SecBuffer self, unsigned long buffer_type, bytes buffer=None, unsigned long length=0):
 */

struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBuffer {
  PyObject *(*replace_ptr)(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *, PSecBuffer);
};
static struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_vtabptr_6spnego_9_sspi_raw_4sspi_SecBuffer;


/* "spnego/_sspi_raw/sspi.pyx":558
 * 
 * 
 * cdef class _AuthIdentityBase:             # <<<<<<<<<<<<<<
 * 
 *     cdef void *__c_value__(_AuthIdentityBase self):
 */

struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi__AuthIdentityBase {
  void *(*__pyx___c_value__)(struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *);
};
static struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_vtabptr_6spnego_9_sspi_raw_4sspi__AuthIdentityBase;


/* "spnego/_sspi_raw/sspi.pyx":564
 * 
 * 
 * cdef class WinNTAuthIdentity(_AuthIdentityBase):             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(WinNTAuthIdentity self, unicode username, unicode domain, unicode password):
 */

struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity {
  struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi__AuthIdentityBase __pyx_base;
};
static struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_vtabptr_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_LONG(LONG value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_short(unsigned short value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PSecBufferDesc __pyx_f_6spnego_9_sspi_raw_4sspi_13SecBufferDesc___c_value__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6spnego_9_sspi_raw_4sspi_9SecBuffer_replace_ptr(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, PSecBuffer __pyx_v_ptr); /* proto*/
static void *__pyx_f_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase___c_value__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_v_self); /* proto*/
static void *__pyx_f_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity___c_value__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self); /* proto*/

/* Module declarations from 'libc.stddef' */

/* Module declarations from 'spnego._sspi_raw.windows' */

/* Module declarations from 'spnego._sspi_raw.security' */

/* Module declarations from 'spnego._sspi_raw.text' */
static PyTypeObject *__pyx_ptype_6spnego_9_sspi_raw_4text_WideChar = 0;
static PyObject *(*__pyx_f_6spnego_9_sspi_raw_4text_u16_to_text)(LPWSTR, size_t); /*proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'spnego._sspi_raw.sspi' */
static PyTypeObject *__pyx_ptype_6spnego_9_sspi_raw_4sspi_Credential = 0;
static PyTypeObject *__pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext = 0;
static PyTypeObject *__pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBufferDesc = 0;
static PyTypeObject *__pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBuffer = 0;
static PyTypeObject *__pyx_ptype_6spnego_9_sspi_raw_4sspi__AuthIdentityBase = 0;
static PyTypeObject *__pyx_ptype_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity = 0;
static PyTypeObject *__pyx_ptype_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ = 0;
static PyObject *__pyx_f_6spnego_9_sspi_raw_4sspi___pyx_unpickle__AuthIdentityBase__set_state(struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "spnego._sspi_raw.sspi"
extern int __pyx_module_is_main_spnego___sspi_raw__sspi;
int __pyx_module_is_main_spnego___sspi_raw__sspi = 0;

/* Implementation of 'spnego._sspi_raw.sspi' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_NotImplementedError;
static const char __pyx_k_s[] = "%s\\";
static const char __pyx_k_0_1[] = "{0}{1}";
static const char __pyx_k__11[] = "";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_qop[] = "qop";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_both[] = "both";
static const char __pyx_k_cred[] = "cred";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_http[] = "http";
static const char __pyx_k_info[] = "info";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_extra[] = "extra";
static const char __pyx_k_input[] = "input";
static const char __pyx_k_names[] = "names";
static const char __pyx_k_rpcid[] = "rpcid";
static const char __pyx_k_sizes[] = "sizes";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_token[] = "token";
static const char __pyx_k_buffer[] = "buffer";
static const char __pyx_k_domain[] = "domain";
static const char __pyx_k_expiry[] = "expiry";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_length[] = "length";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_native[] = "native";
static const char __pyx_k_output[] = "output";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_seq_no[] = "seq_no";
static const char __pyx_k_stream[] = "stream";
static const char __pyx_k_target[] = "target";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_0_1_2_3[] = "<{0}.{1} {2}{3}>";
static const char __pyx_k_SSPIQoP[] = "SSPIQoP";
static const char __pyx_k_buffers[] = "buffers";
static const char __pyx_k_comment[] = "comment";
static const char __pyx_k_context[] = "context";
static const char __pyx_k_inbound[] = "inbound";
static const char __pyx_k_message[] = "message";
static const char __pyx_k_missing[] = "missing";
static const char __pyx_k_network[] = "network";
static const char __pyx_k_package[] = "package";
static const char __pyx_k_padding[] = "padding";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_to_text[] = "to_text";
static const char __pyx_k_version[] = "version";
static const char __pyx_k_SEC_E_OK[] = "SEC_E_OK";
static const char __pyx_k_attrmask[] = "attrmask";
static const char __pyx_k_cBuffers[] = "cBuffers";
static const char __pyx_k_datagram[] = "datagram";
static const char __pyx_k_delegate[] = "delegate";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_identify[] = "identify";
static const char __pyx_k_mechlist[] = "mechlist";
static const char __pyx_k_no_token[] = "no_token";
static const char __pyx_k_outbound[] = "outbound";
static const char __pyx_k_pBuffers[] = "pBuffers";
static const char __pyx_k_password[] = "password";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_raw_info[] = "raw_info";
static const char __pyx_k_readonly[] = "readonly";
static const char __pyx_k_reserved[] = "reserved";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_username[] = "username";
static const char __pyx_k_SecBuffer[] = "SecBuffer";
static const char __pyx_k_SecStatus[] = "SecStatus";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_attribute[] = "attribute";
static const char __pyx_k_auth_data[] = "auth_data";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_from_text[] = "from_text";
static const char __pyx_k_integrity[] = "integrity";
static const char __pyx_k_licensing[] = "licensing";
static const char __pyx_k_max_token[] = "max_token";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_principal[] = "principal";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_to_native[] = "to_native";
static const char __pyx_k_ulVersion[] = "ulVersion";
static const char __pyx_k_w_package[] = "w_package";
static const char __pyx_k_Credential[] = "Credential";
static const char __pyx_k_SecPkgAttr[] = "SecPkgAttr";
static const char __pyx_k_SecPkgInfo[] = "SecPkgInfo";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_block_size[] = "block_size";
static const char __pyx_k_call_level[] = "call_level";
static const char __pyx_k_connection[] = "connection";
static const char __pyx_k_credential[] = "credential";
static const char __pyx_k_namedtuple[] = "namedtuple";
static const char __pyx_k_pkg_params[] = "pkg_params";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_buffer_type[] = "buffer_type";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_context_req[] = "context_req";
static const char __pyx_k_mutual_auth[] = "mutual_auth";
static const char __pyx_k_p_auth_data[] = "p_auth_data";
static const char __pyx_k_session_key[] = "session_key";
static const char __pyx_k_target_host[] = "target_host";
static const char __pyx_k_target_name[] = "target_name";
static const char __pyx_k_w_principal[] = "w_principal";
static const char __pyx_k_alloc_buffer[] = "_alloc_buffer";
static const char __pyx_k_capabilities[] = "capabilities";
static const char __pyx_k_input_buffer[] = "input_buffer";
static const char __pyx_k_no_integrity[] = "no_integrity";
static const char __pyx_k_null_session[] = "null_session";
static const char __pyx_k_package_info[] = "package_info";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_spnego__text[] = "spnego._text";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_CredentialUse[] = "CredentialUse";
static const char __pyx_k_SecBufferDesc[] = "SecBufferDesc";
static const char __pyx_k_SecBufferType[] = "SecBufferType";
static const char __pyx_k_TargetDataRep[] = "TargetDataRep";
static const char __pyx_k_fragment_only[] = "fragment_only";
static const char __pyx_k_input_context[] = "input_context";
static const char __pyx_k_max_signature[] = "max_signature";
static const char __pyx_k_output_buffer[] = "output_buffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_replay_detect[] = "replay_detect";
static const char __pyx_k_stream_header[] = "stream_header";
static const char __pyx_k_use_dce_style[] = "use_dce_style";
static const char __pyx_k_w_target_name[] = "w_target_name";
static const char __pyx_k_credential_use[] = "credential_use";
static const char __pyx_k_dealloc_buffer[] = "_dealloc_buffer";
static const char __pyx_k_extended_error[] = "extended_error";
static const char __pyx_k_make_signature[] = "make_signature";
static const char __pyx_k_proxy_bindings[] = "proxy_bindings";
static const char __pyx_k_stream_trailer[] = "stream_trailer";
static const char __pyx_k_used_dce_style[] = "used_dce_style";
static const char __pyx_k_SEC_E_NOT_OWNER[] = "SEC_E_NOT_OWNER";
static const char __pyx_k_SecPkgAttrSizes[] = "SecPkgAttrSizes";
static const char __pyx_k_SecurityContext[] = "SecurityContext";
static const char __pyx_k_allocate_memory[] = "allocate_memory";
static const char __pyx_k_confidentiality[] = "confidentiality";
static const char __pyx_k_decrypt_message[] = "decrypt_message";
static const char __pyx_k_encrypt_message[] = "encrypt_message";
static const char __pyx_k_fragment_to_fit[] = "fragment_to_fit";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_sequence_detect[] = "sequence_detect";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_target_data_rep[] = "target_data_rep";
static const char __pyx_k_use_session_key[] = "use_session_key";
static const char __pyx_k_used_http_style[] = "used_http_style";
static const char __pyx_k_wrap_no_encrypt[] = "wrap_no_encrypt";
static const char __pyx_k_AuthIdentityBase[] = "_AuthIdentityBase";
static const char __pyx_k_ClientContextReq[] = "ClientContextReq";
static const char __pyx_k_ServerContextReq[] = "ServerContextReq";
static const char __pyx_k_allocated_memory[] = "allocated_memory";
static const char __pyx_k_channel_bindings[] = "channel_bindings";
static const char __pyx_k_negotiation_info[] = "negotiation_info";
static const char __pyx_k_reauthentication[] = "reauthentication";
static const char __pyx_k_security_trailer[] = "security_trailer";
static const char __pyx_k_third_leg_failed[] = "third_leg_failed";
static const char __pyx_k_verify_signature[] = "verify_signature";
static const char __pyx_k_ClientContextAttr[] = "ClientContextAttr";
static const char __pyx_k_ServerContextAttr[] = "ServerContextAttr";
static const char __pyx_k_WinNTAuthIdentity[] = "WinNTAuthIdentity";
static const char __pyx_k_fragment_supplied[] = "fragment_supplied";
static const char __pyx_k_SEC_E_LOGON_DENIED[] = "SEC_E_LOGON_DENIED";
static const char __pyx_k_allow_null_session[] = "allow_null_session";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_mechlist_signature[] = "mechlist_signature";
static const char __pyx_k_use_supplied_creds[] = "use_supplied_creds";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_SEC_E_INVALID_TOKEN[] = "SEC_E_INVALID_TOKEN";
static const char __pyx_k_forward_credentials[] = "forward_credentials";
static const char __pyx_k_intermediate_return[] = "intermediate_return";
static const char __pyx_k_no_additional_token[] = "no_additional_token";
static const char __pyx_k_query_context_names[] = "_query_context_names";
static const char __pyx_k_query_context_sizes[] = "_query_context_sizes";
static const char __pyx_k_used_supplied_creds[] = "used_supplied_creds";
static const char __pyx_k_SEC_E_INTERNAL_ERROR[] = "SEC_E_INTERNAL_ERROR";
static const char __pyx_k_SEC_E_INVALID_HANDLE[] = "SEC_E_INVALID_HANDLE";
static const char __pyx_k_SEC_E_NO_CREDENTIALS[] = "SEC_E_NO_CREDENTIALS";
static const char __pyx_k_SEC_E_TARGET_UNKNOWN[] = "SEC_E_TARGET_UNKNOWN";
static const char __pyx_k_SecBufferDesc___iter[] = "SecBufferDesc.__iter__";
static const char __pyx_k_allow_context_replay[] = "allow_context_replay";
static const char __pyx_k_change_pass_response[] = "change_pass_response";
static const char __pyx_k_used_collected_creds[] = "used_collected_creds";
static const char __pyx_k_SEC_E_CONTEXT_EXPIRED[] = "SEC_E_CONTEXT_EXPIRED";
static const char __pyx_k_SEC_E_MESSAGE_ALTERED[] = "SEC_E_MESSAGE_ALTERED";
static const char __pyx_k_SEC_E_OUT_OF_SEQUENCE[] = "SEC_E_OUT_OF_SEQUENCE";
static const char __pyx_k_SEC_E_WRONG_PRINCIPAL[] = "SEC_E_WRONG_PRINCIPAL";
static const char __pyx_k_SEC_I_COMPLETE_NEEDED[] = "SEC_I_COMPLETE_NEEDED";
static const char __pyx_k_SEC_I_CONTINUE_NEEDED[] = "SEC_I_CONTINUE_NEEDED";
static const char __pyx_k_allow_missing_binding[] = "allow_missing_binding";
static const char __pyx_k_allow_non_user_logons[] = "allow_non_user_logons";
static const char __pyx_k_spnego__sspi_raw_sspi[] = "spnego._sspi_raw.sspi";
static const char __pyx_k_SEC_E_BUFFER_TOO_SMALL[] = "SEC_E_BUFFER_TOO_SMALL";
static const char __pyx_k_SEC_E_SECPKG_NOT_FOUND[] = "SEC_E_SECPKG_NOT_FOUND";
static const char __pyx_k_manual_cred_validation[] = "manual_cred_validation";
static const char __pyx_k_readonly_with_checksum[] = "readonly_with_checksum";
static const char __pyx_k_SEC_E_QOP_NOT_SUPPORTED[] = "SEC_E_QOP_NOT_SUPPORTED";
static const char __pyx_k_accept_security_context[] = "accept_security_context";
static const char __pyx_k_0_ulVersion_1_cBuffers_2[] = "{0}(ulVersion={1}, cBuffers={2})";
static const char __pyx_k_SEC_E_INCOMPLETE_MESSAGE[] = "SEC_E_INCOMPLETE_MESSAGE";
static const char __pyx_k_query_context_attributes[] = "query_context_attributes";
static const char __pyx_k_SEC_E_INSUFFICIENT_MEMORY[] = "SEC_E_INSUFFICIENT_MEMORY";
static const char __pyx_k_SEC_E_UNKNOWN_CREDENTIALS[] = "SEC_E_UNKNOWN_CREDENTIALS";
static const char __pyx_k_query_context_session_key[] = "_query_context_session_key";
static const char __pyx_k_spnego__sspi_raw_sspi_pyx[] = "spnego\\_sspi_raw\\sspi.pyx";
static const char __pyx_k_0_1_ulVersion_2_cBuffers_3[] = "<{0}.{1}(ulVersion={2}, cBuffers={3})>";
static const char __pyx_k_SEC_E_UNSUPPORTED_FUNCTION[] = "SEC_E_UNSUPPORTED_FUNCTION";
static const char __pyx_k_acquire_credentials_handle[] = "acquire_credentials_handle";
static const char __pyx_k_query_context_package_info[] = "_query_context_package_info";
static const char __pyx_k_SEC_E_CRYPTO_SYSTEM_INVALID[] = "SEC_E_CRYPTO_SYSTEM_INVALID";
static const char __pyx_k_SEC_I_COMPLETE_AND_CONTINUE[] = "SEC_I_COMPLETE_AND_CONTINUE";
static const char __pyx_k_initialize_security_context[] = "initialize_security_context";
static const char __pyx_k_SEC_I_INCOMPLETE_CREDENTIALS[] = "SEC_I_INCOMPLETE_CREDENTIALS";
static const char __pyx_k_Cannot_malloc_SecBuffer_buffer[] = "Cannot malloc SecBuffer buffer";
static const char __pyx_k_pyx_unpickle__AuthIdentityBase[] = "__pyx_unpickle__AuthIdentityBase";
static const char __pyx_k_0_1_cbBuffer_2_BufferType_3_pvB[] = "<{0}.{1}(cbBuffer={2}, BufferType={3}, pvBuffer={4})>";
static const char __pyx_k_0_cbBuffer_1_BufferType_2_pvBuf[] = "{0}(cbBuffer={1}, BufferType={2}, pvBuffer={3!r})";
static const char __pyx_k_Only_names_package_info_session[] = "Only names, package_info, session_key, or sizes is implemented";
static const char __pyx_k_Cannot_malloc_SecBufferDesc_buff[] = "Cannot malloc SecBufferDesc buffer";
static const char __pyx_k_Incompatible_checksums_s_vs_0xd4[] = "Incompatible checksums (%s vs 0xd41d8cd = ())";
static const char __pyx_k_Only_an_empty_buffer_can_be_crea[] = "Only an empty buffer can be created with length";
static const char __pyx_k_SEC_E_NO_AUTHENTICATING_AUTHORIT[] = "SEC_E_NO_AUTHENTICATING_AUTHORITY";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static PyObject *__pyx_kp_u_0_1;
static PyObject *__pyx_kp_u_0_1_2_3;
static PyObject *__pyx_kp_u_0_1_cbBuffer_2_BufferType_3_pvB;
static PyObject *__pyx_kp_u_0_1_ulVersion_2_cBuffers_3;
static PyObject *__pyx_kp_u_0_cbBuffer_1_BufferType_2_pvBuf;
static PyObject *__pyx_kp_u_0_ulVersion_1_cBuffers_2;
static PyObject *__pyx_n_s_AuthIdentityBase;
static PyObject *__pyx_kp_u_Cannot_malloc_SecBufferDesc_buff;
static PyObject *__pyx_kp_u_Cannot_malloc_SecBuffer_buffer;
static PyObject *__pyx_n_s_ClientContextAttr;
static PyObject *__pyx_n_s_ClientContextReq;
static PyObject *__pyx_n_s_Credential;
static PyObject *__pyx_n_s_CredentialUse;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xd4;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_kp_u_Only_an_empty_buffer_can_be_crea;
static PyObject *__pyx_kp_u_Only_names_package_info_session;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_SEC_E_BUFFER_TOO_SMALL;
static PyObject *__pyx_n_s_SEC_E_CONTEXT_EXPIRED;
static PyObject *__pyx_n_s_SEC_E_CRYPTO_SYSTEM_INVALID;
static PyObject *__pyx_n_s_SEC_E_INCOMPLETE_MESSAGE;
static PyObject *__pyx_n_s_SEC_E_INSUFFICIENT_MEMORY;
static PyObject *__pyx_n_s_SEC_E_INTERNAL_ERROR;
static PyObject *__pyx_n_s_SEC_E_INVALID_HANDLE;
static PyObject *__pyx_n_s_SEC_E_INVALID_TOKEN;
static PyObject *__pyx_n_s_SEC_E_LOGON_DENIED;
static PyObject *__pyx_n_s_SEC_E_MESSAGE_ALTERED;
static PyObject *__pyx_n_s_SEC_E_NOT_OWNER;
static PyObject *__pyx_n_s_SEC_E_NO_AUTHENTICATING_AUTHORIT;
static PyObject *__pyx_n_s_SEC_E_NO_CREDENTIALS;
static PyObject *__pyx_n_s_SEC_E_OK;
static PyObject *__pyx_n_s_SEC_E_OUT_OF_SEQUENCE;
static PyObject *__pyx_n_s_SEC_E_QOP_NOT_SUPPORTED;
static PyObject *__pyx_n_s_SEC_E_SECPKG_NOT_FOUND;
static PyObject *__pyx_n_s_SEC_E_TARGET_UNKNOWN;
static PyObject *__pyx_n_s_SEC_E_UNKNOWN_CREDENTIALS;
static PyObject *__pyx_n_s_SEC_E_UNSUPPORTED_FUNCTION;
static PyObject *__pyx_n_s_SEC_E_WRONG_PRINCIPAL;
static PyObject *__pyx_n_s_SEC_I_COMPLETE_AND_CONTINUE;
static PyObject *__pyx_n_s_SEC_I_COMPLETE_NEEDED;
static PyObject *__pyx_n_s_SEC_I_CONTINUE_NEEDED;
static PyObject *__pyx_n_s_SEC_I_INCOMPLETE_CREDENTIALS;
static PyObject *__pyx_n_s_SSPIQoP;
static PyObject *__pyx_n_s_SecBuffer;
static PyObject *__pyx_n_s_SecBufferDesc;
static PyObject *__pyx_n_s_SecBufferDesc___iter;
static PyObject *__pyx_n_s_SecBufferType;
static PyObject *__pyx_n_s_SecPkgAttr;
static PyObject *__pyx_n_s_SecPkgAttrSizes;
static PyObject *__pyx_n_u_SecPkgAttrSizes;
static PyObject *__pyx_n_s_SecPkgInfo;
static PyObject *__pyx_n_u_SecPkgInfo;
static PyObject *__pyx_n_s_SecStatus;
static PyObject *__pyx_n_s_SecurityContext;
static PyObject *__pyx_n_s_ServerContextAttr;
static PyObject *__pyx_n_s_ServerContextReq;
static PyObject *__pyx_n_s_TargetDataRep;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_WinNTAuthIdentity;
static PyObject *__pyx_kp_b__11;
static PyObject *__pyx_kp_u__11;
static PyObject *__pyx_n_s_accept_security_context;
static PyObject *__pyx_n_s_acquire_credentials_handle;
static PyObject *__pyx_n_s_alloc_buffer;
static PyObject *__pyx_n_s_allocate_memory;
static PyObject *__pyx_n_s_allocated_memory;
static PyObject *__pyx_n_s_allow_context_replay;
static PyObject *__pyx_n_s_allow_missing_binding;
static PyObject *__pyx_n_s_allow_non_user_logons;
static PyObject *__pyx_n_s_allow_null_session;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_attribute;
static PyObject *__pyx_n_s_attrmask;
static PyObject *__pyx_n_s_auth_data;
static PyObject *__pyx_n_u_block_size;
static PyObject *__pyx_n_s_both;
static PyObject *__pyx_n_s_buffer;
static PyObject *__pyx_n_s_buffer_type;
static PyObject *__pyx_n_s_buffers;
static PyObject *__pyx_n_s_cBuffers;
static PyObject *__pyx_n_s_call_level;
static PyObject *__pyx_n_u_capabilities;
static PyObject *__pyx_n_s_change_pass_response;
static PyObject *__pyx_n_s_channel_bindings;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_u_comment;
static PyObject *__pyx_n_s_confidentiality;
static PyObject *__pyx_n_s_connection;
static PyObject *__pyx_n_s_context;
static PyObject *__pyx_n_s_context_req;
static PyObject *__pyx_n_s_cred;
static PyObject *__pyx_n_s_credential;
static PyObject *__pyx_n_s_credential_use;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_datagram;
static PyObject *__pyx_n_s_dealloc_buffer;
static PyObject *__pyx_n_s_decrypt_message;
static PyObject *__pyx_n_s_delegate;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_domain;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encrypt_message;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_expiry;
static PyObject *__pyx_n_s_extended_error;
static PyObject *__pyx_n_s_extra;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_forward_credentials;
static PyObject *__pyx_n_s_fragment_only;
static PyObject *__pyx_n_s_fragment_supplied;
static PyObject *__pyx_n_s_fragment_to_fit;
static PyObject *__pyx_n_s_from_text;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_http;
static PyObject *__pyx_n_s_identify;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inbound;
static PyObject *__pyx_n_s_info;
static PyObject *__pyx_n_s_initialize_security_context;
static PyObject *__pyx_n_s_input;
static PyObject *__pyx_n_s_input_buffer;
static PyObject *__pyx_n_s_input_context;
static PyObject *__pyx_n_s_integrity;
static PyObject *__pyx_n_s_intermediate_return;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_length;
static PyObject *__pyx_n_s_licensing;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_make_signature;
static PyObject *__pyx_n_s_manual_cred_validation;
static PyObject *__pyx_n_u_max_signature;
static PyObject *__pyx_n_u_max_token;
static PyObject *__pyx_n_s_mechlist;
static PyObject *__pyx_n_s_mechlist_signature;
static PyObject *__pyx_n_s_message;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_missing;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_mutual_auth;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_u_name_2;
static PyObject *__pyx_n_s_namedtuple;
static PyObject *__pyx_n_s_names;
static PyObject *__pyx_n_s_native;
static PyObject *__pyx_n_s_negotiation_info;
static PyObject *__pyx_n_s_network;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_no_additional_token;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_no_integrity;
static PyObject *__pyx_n_s_no_token;
static PyObject *__pyx_n_s_null_session;
static PyObject *__pyx_n_s_outbound;
static PyObject *__pyx_n_s_output;
static PyObject *__pyx_n_s_output_buffer;
static PyObject *__pyx_n_s_pBuffers;
static PyObject *__pyx_n_s_p_auth_data;
static PyObject *__pyx_n_s_package;
static PyObject *__pyx_n_s_package_info;
static PyObject *__pyx_n_s_padding;
static PyObject *__pyx_n_s_password;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pkg_params;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_principal;
static PyObject *__pyx_n_s_proxy_bindings;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle__AuthIdentityBase;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qop;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_query_context_attributes;
static PyObject *__pyx_n_s_query_context_names;
static PyObject *__pyx_n_s_query_context_package_info;
static PyObject *__pyx_n_s_query_context_session_key;
static PyObject *__pyx_n_s_query_context_sizes;
static PyObject *__pyx_n_s_raw_info;
static PyObject *__pyx_n_s_readonly;
static PyObject *__pyx_n_s_readonly_with_checksum;
static PyObject *__pyx_n_s_reauthentication;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_replay_detect;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_reserved;
static PyObject *__pyx_n_u_rpcid;
static PyObject *__pyx_kp_u_s;
static PyObject *__pyx_n_u_security_trailer;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_seq_no;
static PyObject *__pyx_n_s_sequence_detect;
static PyObject *__pyx_n_s_session_key;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_sizes;
static PyObject *__pyx_n_s_spnego__sspi_raw_sspi;
static PyObject *__pyx_kp_s_spnego__sspi_raw_sspi_pyx;
static PyObject *__pyx_n_s_spnego__text;
static PyObject *__pyx_n_s_stream;
static PyObject *__pyx_n_s_stream_header;
static PyObject *__pyx_n_s_stream_trailer;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_target;
static PyObject *__pyx_n_s_target_data_rep;
static PyObject *__pyx_n_s_target_host;
static PyObject *__pyx_n_s_target_name;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_third_leg_failed;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_to_native;
static PyObject *__pyx_n_s_to_text;
static PyObject *__pyx_n_s_token;
static PyObject *__pyx_n_s_ulVersion;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_use_dce_style;
static PyObject *__pyx_n_s_use_session_key;
static PyObject *__pyx_n_s_use_supplied_creds;
static PyObject *__pyx_n_s_used_collected_creds;
static PyObject *__pyx_n_s_used_dce_style;
static PyObject *__pyx_n_s_used_http_style;
static PyObject *__pyx_n_s_used_supplied_creds;
static PyObject *__pyx_n_s_username;
static PyObject *__pyx_n_s_verify_signature;
static PyObject *__pyx_n_s_version;
static PyObject *__pyx_n_u_version;
static PyObject *__pyx_n_s_w_package;
static PyObject *__pyx_n_s_w_principal;
static PyObject *__pyx_n_s_w_target_name;
static PyObject *__pyx_n_s_wrap_no_encrypt;
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential___cinit__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_self); /* proto */
static void __pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_2__dealloc__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_6expiry___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext___cinit__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self); /* proto */
static void __pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_2__dealloc__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_12context_attr___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_6expiry___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc___cinit__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self, PyObject *__pyx_v_buffers, unsigned long __pyx_v_version); /* proto */
static Py_ssize_t __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_2__len__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_4__getitem__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_6__setitem__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self, PyObject *__pyx_v_key, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_8__iter__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_11__repr__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_13__str__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self); /* proto */
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self, unsigned long __pyx_v_value); /* proto */
static void __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_15__dealloc__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_17__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_19__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer___cinit__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, unsigned long __pyx_v_buffer_type, PyObject *__pyx_v_buffer, unsigned long __pyx_v_length); /* proto */
static Py_ssize_t __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_2__len__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_4__repr__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_6__str__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self); /* proto */
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, unsigned long __pyx_v_value); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self); /* proto */
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_8_alloc_buffer(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, unsigned long __pyx_v_length); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_10_dealloc_buffer(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self); /* proto */
static void __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_12__dealloc__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase___reduce_cython__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase_2__setstate_cython__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity___cinit__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self, PyObject *__pyx_v_username, PyObject *__pyx_v_domain, PyObject *__pyx_v_password); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_2__repr__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_4__str__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self); /* proto */
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self); /* proto */
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self); /* proto */
static int __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static void __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6__dealloc__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_accept_security_context(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_credential, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_input_buffer, unsigned long __pyx_v_context_req, unsigned long __pyx_v_target_data_rep, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_output_buffer); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_2acquire_credentials_handle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_principal, PyObject *__pyx_v_package, unsigned long __pyx_v_credential_use, struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_v_auth_data); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_4decrypt_message(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message, unsigned long __pyx_v_seq_no); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_6encrypt_message(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message, unsigned long __pyx_v_seq_no, unsigned long __pyx_v_qop); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_8initialize_security_context(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_credential, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, PyObject *__pyx_v_target_name, unsigned long __pyx_v_context_req, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_input_buffer, unsigned long __pyx_v_target_data_rep, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_output_buffer); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_10make_signature(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, unsigned long __pyx_v_qop, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message, unsigned long __pyx_v_seq_no); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_12query_context_attributes(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, unsigned long __pyx_v_attribute); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_14_query_context_names(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_16_query_context_package_info(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_18_query_context_session_key(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_20_query_context_sizes(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_22verify_signature(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message, unsigned long __pyx_v_seq_no); /* proto */
static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_24__pyx_unpickle__AuthIdentityBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi_Credential(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi_SecurityContext(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi_SecBufferDesc(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi_SecBuffer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi__AuthIdentityBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_222419149;
static PyObject *__pyx_int_268435456;
static unsigned long __pyx_k__5;
static unsigned long __pyx_k__16;
static unsigned long __pyx_k__17;
static unsigned long __pyx_k__18;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__41;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__45;
/* Late includes */

/* "spnego/_sspi_raw/sspi.pyx":387
 * cdef class Credential:
 * 
 *     def __cinit__(Credential self):             # <<<<<<<<<<<<<<
 *         self.expiry = 0
 * 
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential___cinit__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential___cinit__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":388
 * 
 *     def __cinit__(Credential self):
 *         self.expiry = 0             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(Credential self):
 */
  __pyx_v_self->expiry = 0;

  /* "spnego/_sspi_raw/sspi.pyx":387
 * cdef class Credential:
 * 
 *     def __cinit__(Credential self):             # <<<<<<<<<<<<<<
 *         self.expiry = 0
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":390
 *         self.expiry = 0
 * 
 *     def __dealloc__(Credential self):             # <<<<<<<<<<<<<<
 *         if self.expiry:
 *             FreeCredentialsHandle(&self.handle)
 */

/* Python wrapper */
static void __pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_2__dealloc__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_2__dealloc__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":391
 * 
 *     def __dealloc__(Credential self):
 *         if self.expiry:             # <<<<<<<<<<<<<<
 *             FreeCredentialsHandle(&self.handle)
 * 
 */
  __pyx_t_1 = (__pyx_v_self->expiry != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":392
 *     def __dealloc__(Credential self):
 *         if self.expiry:
 *             FreeCredentialsHandle(&self.handle)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (void)(FreeCredentialsHandle((&__pyx_v_self->handle)));

    /* "spnego/_sspi_raw/sspi.pyx":391
 * 
 *     def __dealloc__(Credential self):
 *         if self.expiry:             # <<<<<<<<<<<<<<
 *             FreeCredentialsHandle(&self.handle)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":390
 *         self.expiry = 0
 * 
 *     def __dealloc__(Credential self):             # <<<<<<<<<<<<<<
 *         if self.expiry:
 *             FreeCredentialsHandle(&self.handle)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "spnego/_sspi_raw/sspi.pxd":22
 * 
 *     cdef CredHandle handle
 *     cdef readonly unsigned long long expiry             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_6expiry_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_6expiry_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_6expiry___get__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_6expiry___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_PY_LONG_LONG(__pyx_v_self->expiry); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.Credential.expiry.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_4__reduce_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.Credential.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_6__setstate_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_10Credential_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.Credential.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":397
 * cdef class SecurityContext:
 * 
 *     def __cinit__(SecurityContext self):             # <<<<<<<<<<<<<<
 *         self.context_attr = 0
 *         self.expiry = 0
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext___cinit__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext___cinit__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":398
 * 
 *     def __cinit__(SecurityContext self):
 *         self.context_attr = 0             # <<<<<<<<<<<<<<
 *         self.expiry = 0
 *         self.init = 0
 */
  __pyx_v_self->context_attr = 0;

  /* "spnego/_sspi_raw/sspi.pyx":399
 *     def __cinit__(SecurityContext self):
 *         self.context_attr = 0
 *         self.expiry = 0             # <<<<<<<<<<<<<<
 *         self.init = 0
 * 
 */
  __pyx_v_self->expiry = 0;

  /* "spnego/_sspi_raw/sspi.pyx":400
 *         self.context_attr = 0
 *         self.expiry = 0
 *         self.init = 0             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(SecurityContext self):
 */
  __pyx_v_self->init = 0;

  /* "spnego/_sspi_raw/sspi.pyx":397
 * cdef class SecurityContext:
 * 
 *     def __cinit__(SecurityContext self):             # <<<<<<<<<<<<<<
 *         self.context_attr = 0
 *         self.expiry = 0
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":402
 *         self.init = 0
 * 
 *     def __dealloc__(SecurityContext self):             # <<<<<<<<<<<<<<
 *         if self.init:
 *             DeleteSecurityContext(&self.handle)
 */

/* Python wrapper */
static void __pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_2__dealloc__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_2__dealloc__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":403
 * 
 *     def __dealloc__(SecurityContext self):
 *         if self.init:             # <<<<<<<<<<<<<<
 *             DeleteSecurityContext(&self.handle)
 * 
 */
  __pyx_t_1 = (__pyx_v_self->init != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":404
 *     def __dealloc__(SecurityContext self):
 *         if self.init:
 *             DeleteSecurityContext(&self.handle)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (void)(DeleteSecurityContext((&__pyx_v_self->handle)));

    /* "spnego/_sspi_raw/sspi.pyx":403
 * 
 *     def __dealloc__(SecurityContext self):
 *         if self.init:             # <<<<<<<<<<<<<<
 *             DeleteSecurityContext(&self.handle)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":402
 *         self.init = 0
 * 
 *     def __dealloc__(SecurityContext self):             # <<<<<<<<<<<<<<
 *         if self.init:
 *             DeleteSecurityContext(&self.handle)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "spnego/_sspi_raw/sspi.pxd":28
 * 
 *     cdef CtxtHandle handle
 *     cdef readonly unsigned long context_attr             # <<<<<<<<<<<<<<
 *     cdef readonly unsigned long long expiry
 *     cdef unsigned long init
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_12context_attr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_12context_attr_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_12context_attr___get__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_12context_attr___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->context_attr); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecurityContext.context_attr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pxd":29
 *     cdef CtxtHandle handle
 *     cdef readonly unsigned long context_attr
 *     cdef readonly unsigned long long expiry             # <<<<<<<<<<<<<<
 *     cdef unsigned long init
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_6expiry_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_6expiry_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_6expiry___get__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_6expiry___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_PY_LONG_LONG(__pyx_v_self->expiry); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecurityContext.expiry.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_4__reduce_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecurityContext.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_6__setstate_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_15SecurityContext_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecurityContext.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":409
 * cdef class SecBufferDesc:
 * 
 *     def __cinit__(SecBufferDesc self, list buffers not None, unsigned long version=SECBUFFER_VERSION):             # <<<<<<<<<<<<<<
 *         self._c_value = NativeSecBufferDesc(version, <unsigned long>len(buffers), NULL)
 *         self._buffers = buffers
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_buffers = 0;
  unsigned long __pyx_v_version;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_buffers,&__pyx_n_s_version,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffers)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_version);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 409, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_buffers = ((PyObject*)values[0]);
    if (values[1]) {
      __pyx_v_version = __Pyx_PyInt_As_unsigned_long(values[1]); if (unlikely((__pyx_v_version == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 409, __pyx_L3_error)
    } else {
      __pyx_v_version = __pyx_k__5;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 409, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBufferDesc.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_buffers), (&PyList_Type), 0, "buffers", 1))) __PYX_ERR(1, 409, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc___cinit__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self), __pyx_v_buffers, __pyx_v_version);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc___cinit__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self, PyObject *__pyx_v_buffers, unsigned long __pyx_v_version) {
  Py_ssize_t __pyx_v_idx;
  PyObject *__pyx_v_buffer = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  SecBufferDesc __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":410
 * 
 *     def __cinit__(SecBufferDesc self, list buffers not None, unsigned long version=SECBUFFER_VERSION):
 *         self._c_value = NativeSecBufferDesc(version, <unsigned long>len(buffers), NULL)             # <<<<<<<<<<<<<<
 *         self._buffers = buffers
 * 
 */
  __pyx_t_1.ulVersion = __pyx_v_version;
  __pyx_t_2 = PyList_GET_SIZE(__pyx_v_buffers); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 410, __pyx_L1_error)
  __pyx_t_1.cBuffers = ((unsigned long)__pyx_t_2);
  __pyx_t_1.pBuffers = NULL;
  __pyx_v_self->_c_value = __pyx_t_1;

  /* "spnego/_sspi_raw/sspi.pyx":411
 *     def __cinit__(SecBufferDesc self, list buffers not None, unsigned long version=SECBUFFER_VERSION):
 *         self._c_value = NativeSecBufferDesc(version, <unsigned long>len(buffers), NULL)
 *         self._buffers = buffers             # <<<<<<<<<<<<<<
 * 
 *         if buffers:
 */
  __Pyx_INCREF(__pyx_v_buffers);
  __Pyx_GIVEREF(__pyx_v_buffers);
  __Pyx_GOTREF(__pyx_v_self->_buffers);
  __Pyx_DECREF(__pyx_v_self->_buffers);
  __pyx_v_self->_buffers = __pyx_v_buffers;

  /* "spnego/_sspi_raw/sspi.pyx":413
 *         self._buffers = buffers
 * 
 *         if buffers:             # <<<<<<<<<<<<<<
 *             self._c_value.pBuffers = <PSecBuffer>malloc(sizeof(NativeSecBuffer) * len(buffers))
 *             if not self._c_value.pBuffers:
 */
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_buffers) != 0);
  if (__pyx_t_3) {

    /* "spnego/_sspi_raw/sspi.pyx":414
 * 
 *         if buffers:
 *             self._c_value.pBuffers = <PSecBuffer>malloc(sizeof(NativeSecBuffer) * len(buffers))             # <<<<<<<<<<<<<<
 *             if not self._c_value.pBuffers:
 *                 raise MemoryError("Cannot malloc SecBufferDesc buffer")
 */
    __pyx_t_2 = PyList_GET_SIZE(__pyx_v_buffers); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 414, __pyx_L1_error)
    __pyx_v_self->_c_value.pBuffers = ((PSecBuffer)malloc(((sizeof(SecBuffer)) * __pyx_t_2)));

    /* "spnego/_sspi_raw/sspi.pyx":415
 *         if buffers:
 *             self._c_value.pBuffers = <PSecBuffer>malloc(sizeof(NativeSecBuffer) * len(buffers))
 *             if not self._c_value.pBuffers:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("Cannot malloc SecBufferDesc buffer")
 * 
 */
    __pyx_t_3 = ((!(__pyx_v_self->_c_value.pBuffers != 0)) != 0);
    if (unlikely(__pyx_t_3)) {

      /* "spnego/_sspi_raw/sspi.pyx":416
 *             self._c_value.pBuffers = <PSecBuffer>malloc(sizeof(NativeSecBuffer) * len(buffers))
 *             if not self._c_value.pBuffers:
 *                 raise MemoryError("Cannot malloc SecBufferDesc buffer")             # <<<<<<<<<<<<<<
 * 
 *             for idx, buffer in enumerate(buffers):
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(1, 416, __pyx_L1_error)

      /* "spnego/_sspi_raw/sspi.pyx":415
 *         if buffers:
 *             self._c_value.pBuffers = <PSecBuffer>malloc(sizeof(NativeSecBuffer) * len(buffers))
 *             if not self._c_value.pBuffers:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("Cannot malloc SecBufferDesc buffer")
 * 
 */
    }

    /* "spnego/_sspi_raw/sspi.pyx":418
 *                 raise MemoryError("Cannot malloc SecBufferDesc buffer")
 * 
 *             for idx, buffer in enumerate(buffers):             # <<<<<<<<<<<<<<
 *                 self[idx] = buffer
 * 
 */
    __pyx_t_2 = 0;
    __pyx_t_4 = __pyx_v_buffers; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    for (;;) {
      if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(1, 418, __pyx_L1_error)
      #else
      __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 418, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_buffer, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_v_idx = __pyx_t_2;
      __pyx_t_2 = (__pyx_t_2 + 1);

      /* "spnego/_sspi_raw/sspi.pyx":419
 * 
 *             for idx, buffer in enumerate(buffers):
 *                 self[idx] = buffer             # <<<<<<<<<<<<<<
 * 
 *     cdef PSecBufferDesc __c_value__(SecBufferDesc self):
 */
      if (unlikely(__Pyx_SetItemInt(((PyObject *)__pyx_v_self), __pyx_v_idx, __pyx_v_buffer, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1) < 0)) __PYX_ERR(1, 419, __pyx_L1_error)

      /* "spnego/_sspi_raw/sspi.pyx":418
 *                 raise MemoryError("Cannot malloc SecBufferDesc buffer")
 * 
 *             for idx, buffer in enumerate(buffers):             # <<<<<<<<<<<<<<
 *                 self[idx] = buffer
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "spnego/_sspi_raw/sspi.pyx":413
 *         self._buffers = buffers
 * 
 *         if buffers:             # <<<<<<<<<<<<<<
 *             self._c_value.pBuffers = <PSecBuffer>malloc(sizeof(NativeSecBuffer) * len(buffers))
 *             if not self._c_value.pBuffers:
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":409
 * cdef class SecBufferDesc:
 * 
 *     def __cinit__(SecBufferDesc self, list buffers not None, unsigned long version=SECBUFFER_VERSION):             # <<<<<<<<<<<<<<
 *         self._c_value = NativeSecBufferDesc(version, <unsigned long>len(buffers), NULL)
 *         self._buffers = buffers
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBufferDesc.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":421
 *                 self[idx] = buffer
 * 
 *     cdef PSecBufferDesc __c_value__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         return &self._c_value
 * 
 */

static PSecBufferDesc __pyx_f_6spnego_9_sspi_raw_4sspi_13SecBufferDesc___c_value__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self) {
  PSecBufferDesc __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__c_value__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":422
 * 
 *     cdef PSecBufferDesc __c_value__(SecBufferDesc self):
 *         return &self._c_value             # <<<<<<<<<<<<<<
 * 
 *     def __len__(SecBufferDesc self):
 */
  __pyx_r = (&__pyx_v_self->_c_value);
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":421
 *                 self[idx] = buffer
 * 
 *     cdef PSecBufferDesc __c_value__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         return &self._c_value
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":424
 *         return &self._c_value
 * 
 *     def __len__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         return self._c_value.cBuffers
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_3__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_2__len__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_2__len__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":425
 * 
 *     def __len__(SecBufferDesc self):
 *         return self._c_value.cBuffers             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(SecBufferDesc self, key):
 */
  __pyx_r = __pyx_v_self->_c_value.cBuffers;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":424
 *         return &self._c_value
 * 
 *     def __len__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         return self._c_value.cBuffers
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":427
 *         return self._c_value.cBuffers
 * 
 *     def __getitem__(SecBufferDesc self, key):             # <<<<<<<<<<<<<<
 *         return self._buffers[key]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_4__getitem__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_4__getitem__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":428
 * 
 *     def __getitem__(SecBufferDesc self, key):
 *         return self._buffers[key]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(SecBufferDesc self, key, SecBuffer value not None):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->_buffers == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 428, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->_buffers, __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":427
 *         return self._c_value.cBuffers
 * 
 *     def __getitem__(SecBufferDesc self, key):             # <<<<<<<<<<<<<<
 *         return self._buffers[key]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBufferDesc.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":430
 *         return self._buffers[key]
 * 
 *     def __setitem__(SecBufferDesc self, key, SecBuffer value not None):             # <<<<<<<<<<<<<<
 *         value.replace_ptr(&self._c_value.pBuffers[key])
 * 
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBuffer, 0, "value", 0))) __PYX_ERR(1, 430, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_6__setitem__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self), ((PyObject *)__pyx_v_key), ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_6__setitem__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self, PyObject *__pyx_v_key, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":431
 * 
 *     def __setitem__(SecBufferDesc self, key, SecBuffer value not None):
 *         value.replace_ptr(&self._c_value.pBuffers[key])             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(SecBufferDesc self):
 */
  __pyx_t_1 = __Pyx_PyIndex_AsSsize_t(__pyx_v_key); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 431, __pyx_L1_error)
  __pyx_t_2 = ((struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_value->__pyx_vtab)->replace_ptr(__pyx_v_value, (&(__pyx_v_self->_c_value.pBuffers[__pyx_t_1]))); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":430
 *         return self._buffers[key]
 * 
 *     def __setitem__(SecBufferDesc self, key, SecBuffer value not None):             # <<<<<<<<<<<<<<
 *         value.replace_ptr(&self._c_value.pBuffers[key])
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBufferDesc.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_10generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "spnego/_sspi_raw/sspi.pyx":433
 *         value.replace_ptr(&self._c_value.pBuffers[key])
 * 
 *     def __iter__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         for val in self._buffers:
 *             yield val
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_9__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_9__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_8__iter__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_8__iter__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ *)__pyx_tp_new_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__(__pyx_ptype_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 433, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_10generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_SecBufferDesc___iter, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!gen)) __PYX_ERR(1, 433, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBufferDesc.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_10generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(1, 433, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":434
 * 
 *     def __iter__(SecBufferDesc self):
 *         for val in self._buffers:             # <<<<<<<<<<<<<<
 *             yield val
 * 
 */
  if (unlikely(__pyx_cur_scope->__pyx_v_self->_buffers == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 434, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->_buffers; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 434, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_val);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_val, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "spnego/_sspi_raw/sspi.pyx":435
 *     def __iter__(SecBufferDesc self):
 *         for val in self._buffers:
 *             yield val             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(SecBufferDesc self):
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_val);
    __pyx_r = __pyx_cur_scope->__pyx_v_val;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(1, 435, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":434
 * 
 *     def __iter__(SecBufferDesc self):
 *         for val in self._buffers:             # <<<<<<<<<<<<<<
 *             yield val
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "spnego/_sspi_raw/sspi.pyx":433
 *         value.replace_ptr(&self._c_value.pBuffers[key])
 * 
 *     def __iter__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         for val in self._buffers:
 *             yield val
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":437
 *             yield val
 * 
 *     def __repr__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         return "<{0}.{1}(ulVersion={2}, cBuffers={3})>".format(type(self).__module__, type(self).__name__,
 *             self._c_value.ulVersion, self._c_value.cBuffers)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_12__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_12__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_11__repr__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_11__repr__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":438
 * 
 *     def __repr__(SecBufferDesc self):
 *         return "<{0}.{1}(ulVersion={2}, cBuffers={3})>".format(type(self).__module__, type(self).__name__,             # <<<<<<<<<<<<<<
 *             self._c_value.ulVersion, self._c_value.cBuffers)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_1_ulVersion_2_cBuffers_3, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_module); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "spnego/_sspi_raw/sspi.pyx":439
 *     def __repr__(SecBufferDesc self):
 *         return "<{0}.{1}(ulVersion={2}, cBuffers={3})>".format(type(self).__module__, type(self).__name__,
 *             self._c_value.ulVersion, self._c_value.cBuffers)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(SecBufferDesc self):
 */
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->_c_value.ulVersion); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->_c_value.cBuffers); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 438, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 438, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":437
 *             yield val
 * 
 *     def __repr__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         return "<{0}.{1}(ulVersion={2}, cBuffers={3})>".format(type(self).__module__, type(self).__name__,
 *             self._c_value.ulVersion, self._c_value.cBuffers)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBufferDesc.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":441
 *             self._c_value.ulVersion, self._c_value.cBuffers)
 * 
 *     def __str__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         return "{0}(ulVersion={1}, cBuffers={2})".format(type(self).__name__, self._c_value.ulVersion,
 *             self._c_value.cBuffers)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_14__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_14__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_13__str__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_13__str__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":442
 * 
 *     def __str__(SecBufferDesc self):
 *         return "{0}(ulVersion={1}, cBuffers={2})".format(type(self).__name__, self._c_value.ulVersion,             # <<<<<<<<<<<<<<
 *             self._c_value.cBuffers)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_ulVersion_1_cBuffers_2, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->_c_value.ulVersion); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "spnego/_sspi_raw/sspi.pyx":443
 *     def __str__(SecBufferDesc self):
 *         return "{0}(ulVersion={1}, cBuffers={2})".format(type(self).__name__, self._c_value.ulVersion,
 *             self._c_value.cBuffers)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->_c_value.cBuffers); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 442, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 442, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":441
 *             self._c_value.ulVersion, self._c_value.cBuffers)
 * 
 *     def __str__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         return "{0}(ulVersion={1}, cBuffers={2})".format(type(self).__name__, self._c_value.ulVersion,
 *             self._c_value.cBuffers)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBufferDesc.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":446
 * 
 *     @property
 *     def version(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         return self._c_value.ulVersion
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version___get__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":447
 *     @property
 *     def version(SecBufferDesc self):
 *         return self._c_value.ulVersion             # <<<<<<<<<<<<<<
 * 
 *     @version.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->_c_value.ulVersion); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":446
 * 
 *     @property
 *     def version(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         return self._c_value.ulVersion
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBufferDesc.version.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":450
 * 
 *     @version.setter
 *     def version(SecBufferDesc self, unsigned long value):             # <<<<<<<<<<<<<<
 *         self._c_value.ulVersion = value
 * 
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  unsigned long __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyInt_As_unsigned_long(__pyx_arg_value); if (unlikely((__pyx_v_value == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 450, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBufferDesc.version.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version_2__set__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self), ((unsigned long)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self, unsigned long __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":451
 *     @version.setter
 *     def version(SecBufferDesc self, unsigned long value):
 *         self._c_value.ulVersion = value             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(SecBufferDesc self):
 */
  __pyx_v_self->_c_value.ulVersion = __pyx_v_value;

  /* "spnego/_sspi_raw/sspi.pyx":450
 * 
 *     @version.setter
 *     def version(SecBufferDesc self, unsigned long value):             # <<<<<<<<<<<<<<
 *         self._c_value.ulVersion = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":453
 *         self._c_value.ulVersion = value
 * 
 *     def __dealloc__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         if self._c_value.pBuffers:
 *             free(self._c_value.pBuffers)
 */

/* Python wrapper */
static void __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_16__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_16__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_15__dealloc__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_15__dealloc__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":454
 * 
 *     def __dealloc__(SecBufferDesc self):
 *         if self._c_value.pBuffers:             # <<<<<<<<<<<<<<
 *             free(self._c_value.pBuffers)
 *             self._c_value.pBuffers = NULL
 */
  __pyx_t_1 = (__pyx_v_self->_c_value.pBuffers != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":455
 *     def __dealloc__(SecBufferDesc self):
 *         if self._c_value.pBuffers:
 *             free(self._c_value.pBuffers)             # <<<<<<<<<<<<<<
 *             self._c_value.pBuffers = NULL
 * 
 */
    free(__pyx_v_self->_c_value.pBuffers);

    /* "spnego/_sspi_raw/sspi.pyx":456
 *         if self._c_value.pBuffers:
 *             free(self._c_value.pBuffers)
 *             self._c_value.pBuffers = NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_self->_c_value.pBuffers = NULL;

    /* "spnego/_sspi_raw/sspi.pyx":454
 * 
 *     def __dealloc__(SecBufferDesc self):
 *         if self._c_value.pBuffers:             # <<<<<<<<<<<<<<
 *             free(self._c_value.pBuffers)
 *             self._c_value.pBuffers = NULL
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":453
 *         self._c_value.ulVersion = value
 * 
 *     def __dealloc__(SecBufferDesc self):             # <<<<<<<<<<<<<<
 *         if self._c_value.pBuffers:
 *             free(self._c_value.pBuffers)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_18__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_18__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_17__reduce_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_17__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBufferDesc.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_20__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_20__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_19__setstate_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_19__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBufferDesc.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":461
 * cdef class SecBuffer:
 * 
 *     def __cinit__(SecBuffer self, unsigned long buffer_type, bytes buffer=None, unsigned long length=0):             # <<<<<<<<<<<<<<
 *         if buffer and length:
 *             raise ValueError("Only an empty buffer can be created with length")
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  unsigned long __pyx_v_buffer_type;
  PyObject *__pyx_v_buffer = 0;
  unsigned long __pyx_v_length;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_buffer_type,&__pyx_n_s_buffer,&__pyx_n_s_length,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject*)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_buffer);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_length);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 461, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_buffer_type = __Pyx_PyInt_As_unsigned_long(values[0]); if (unlikely((__pyx_v_buffer_type == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 461, __pyx_L3_error)
    __pyx_v_buffer = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_length = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_length == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 461, __pyx_L3_error)
    } else {
      __pyx_v_length = ((unsigned long)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 461, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_buffer), (&PyBytes_Type), 1, "buffer", 1))) __PYX_ERR(1, 461, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer___cinit__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self), __pyx_v_buffer_type, __pyx_v_buffer, __pyx_v_length);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer___cinit__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, unsigned long __pyx_v_buffer_type, PyObject *__pyx_v_buffer, unsigned long __pyx_v_length) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":462
 * 
 *     def __cinit__(SecBuffer self, unsigned long buffer_type, bytes buffer=None, unsigned long length=0):
 *         if buffer and length:             # <<<<<<<<<<<<<<
 *             raise ValueError("Only an empty buffer can be created with length")
 * 
 */
  __pyx_t_2 = (__pyx_v_buffer != Py_None)&&(PyBytes_GET_SIZE(__pyx_v_buffer) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_length != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "spnego/_sspi_raw/sspi.pyx":463
 *     def __cinit__(SecBuffer self, unsigned long buffer_type, bytes buffer=None, unsigned long length=0):
 *         if buffer and length:
 *             raise ValueError("Only an empty buffer can be created with length")             # <<<<<<<<<<<<<<
 * 
 *         self.sys_alloc = 0
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 463, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":462
 * 
 *     def __cinit__(SecBuffer self, unsigned long buffer_type, bytes buffer=None, unsigned long length=0):
 *         if buffer and length:             # <<<<<<<<<<<<<<
 *             raise ValueError("Only an empty buffer can be created with length")
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":465
 *             raise ValueError("Only an empty buffer can be created with length")
 * 
 *         self.sys_alloc = 0             # <<<<<<<<<<<<<<
 *         self._is_owner = 1
 * 
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GOTREF(__pyx_v_self->sys_alloc);
  __Pyx_DECREF(__pyx_v_self->sys_alloc);
  __pyx_v_self->sys_alloc = __pyx_int_0;

  /* "spnego/_sspi_raw/sspi.pyx":466
 * 
 *         self.sys_alloc = 0
 *         self._is_owner = 1             # <<<<<<<<<<<<<<
 * 
 *         self.c_value = <PSecBuffer>calloc(1, sizeof(NativeSecBuffer))
 */
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GOTREF(__pyx_v_self->_is_owner);
  __Pyx_DECREF(__pyx_v_self->_is_owner);
  __pyx_v_self->_is_owner = __pyx_int_1;

  /* "spnego/_sspi_raw/sspi.pyx":468
 *         self._is_owner = 1
 * 
 *         self.c_value = <PSecBuffer>calloc(1, sizeof(NativeSecBuffer))             # <<<<<<<<<<<<<<
 *         if not self.c_value:
 *             raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover
 */
  __pyx_v_self->c_value = ((PSecBuffer)calloc(1, (sizeof(SecBuffer))));

  /* "spnego/_sspi_raw/sspi.pyx":469
 * 
 *         self.c_value = <PSecBuffer>calloc(1, sizeof(NativeSecBuffer))
 *         if not self.c_value:             # <<<<<<<<<<<<<<
 *             raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->c_value != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "spnego/_sspi_raw/sspi.pyx":470
 *         self.c_value = <PSecBuffer>calloc(1, sizeof(NativeSecBuffer))
 *         if not self.c_value:
 *             raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover             # <<<<<<<<<<<<<<
 * 
 *         self.buffer_type = buffer_type
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 470, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":469
 * 
 *         self.c_value = <PSecBuffer>calloc(1, sizeof(NativeSecBuffer))
 *         if not self.c_value:             # <<<<<<<<<<<<<<
 *             raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":472
 *             raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover
 * 
 *         self.buffer_type = buffer_type             # <<<<<<<<<<<<<<
 *         if buffer:
 *             self.buffer = buffer
 */
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_v_buffer_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_buffer_type, __pyx_t_3) < 0) __PYX_ERR(1, 472, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":473
 * 
 *         self.buffer_type = buffer_type
 *         if buffer:             # <<<<<<<<<<<<<<
 *             self.buffer = buffer
 *         elif length:
 */
  __pyx_t_1 = (__pyx_v_buffer != Py_None)&&(PyBytes_GET_SIZE(__pyx_v_buffer) != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":474
 *         self.buffer_type = buffer_type
 *         if buffer:
 *             self.buffer = buffer             # <<<<<<<<<<<<<<
 *         elif length:
 *             self._alloc_buffer(length)
 */
    if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_buffer, __pyx_v_buffer) < 0) __PYX_ERR(1, 474, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":473
 * 
 *         self.buffer_type = buffer_type
 *         if buffer:             # <<<<<<<<<<<<<<
 *             self.buffer = buffer
 *         elif length:
 */
    goto __pyx_L7;
  }

  /* "spnego/_sspi_raw/sspi.pyx":475
 *         if buffer:
 *             self.buffer = buffer
 *         elif length:             # <<<<<<<<<<<<<<
 *             self._alloc_buffer(length)
 * 
 */
  __pyx_t_1 = (__pyx_v_length != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":476
 *             self.buffer = buffer
 *         elif length:
 *             self._alloc_buffer(length)             # <<<<<<<<<<<<<<
 * 
 *     cdef replace_ptr(SecBuffer self, PSecBuffer ptr):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_alloc_buffer); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "spnego/_sspi_raw/sspi.pyx":475
 *         if buffer:
 *             self.buffer = buffer
 *         elif length:             # <<<<<<<<<<<<<<
 *             self._alloc_buffer(length)
 * 
 */
  }
  __pyx_L7:;

  /* "spnego/_sspi_raw/sspi.pyx":461
 * cdef class SecBuffer:
 * 
 *     def __cinit__(SecBuffer self, unsigned long buffer_type, bytes buffer=None, unsigned long length=0):             # <<<<<<<<<<<<<<
 *         if buffer and length:
 *             raise ValueError("Only an empty buffer can be created with length")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":478
 *             self._alloc_buffer(length)
 * 
 *     cdef replace_ptr(SecBuffer self, PSecBuffer ptr):             # <<<<<<<<<<<<<<
 *         # Copy the existing C buffer to the new pointer and free the memory this object was managing.
 *         memcpy(ptr, self.c_value, sizeof(NativeSecBuffer))
 */

static PyObject *__pyx_f_6spnego_9_sspi_raw_4sspi_9SecBuffer_replace_ptr(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, PSecBuffer __pyx_v_ptr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("replace_ptr", 0);

  /* "spnego/_sspi_raw/sspi.pyx":480
 *     cdef replace_ptr(SecBuffer self, PSecBuffer ptr):
 *         # Copy the existing C buffer to the new pointer and free the memory this object was managing.
 *         memcpy(ptr, self.c_value, sizeof(NativeSecBuffer))             # <<<<<<<<<<<<<<
 *         free(self.c_value)
 * 
 */
  (void)(memcpy(__pyx_v_ptr, __pyx_v_self->c_value, (sizeof(SecBuffer))));

  /* "spnego/_sspi_raw/sspi.pyx":481
 *         # Copy the existing C buffer to the new pointer and free the memory this object was managing.
 *         memcpy(ptr, self.c_value, sizeof(NativeSecBuffer))
 *         free(self.c_value)             # <<<<<<<<<<<<<<
 * 
 *         # Make sure that this obj does not try and free the new pointer that was passed in (it is up to the caller to
 */
  free(__pyx_v_self->c_value);

  /* "spnego/_sspi_raw/sspi.pyx":485
 *         # Make sure that this obj does not try and free the new pointer that was passed in (it is up to the caller to
 *         # do that).
 *         self._is_owner = 0             # <<<<<<<<<<<<<<
 *         self.c_value = ptr
 * 
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GOTREF(__pyx_v_self->_is_owner);
  __Pyx_DECREF(__pyx_v_self->_is_owner);
  __pyx_v_self->_is_owner = __pyx_int_0;

  /* "spnego/_sspi_raw/sspi.pyx":486
 *         # do that).
 *         self._is_owner = 0
 *         self.c_value = ptr             # <<<<<<<<<<<<<<
 * 
 *     def __len__(SecBuffer self):
 */
  __pyx_v_self->c_value = __pyx_v_ptr;

  /* "spnego/_sspi_raw/sspi.pyx":478
 *             self._alloc_buffer(length)
 * 
 *     cdef replace_ptr(SecBuffer self, PSecBuffer ptr):             # <<<<<<<<<<<<<<
 *         # Copy the existing C buffer to the new pointer and free the memory this object was managing.
 *         memcpy(ptr, self.c_value, sizeof(NativeSecBuffer))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":488
 *         self.c_value = ptr
 * 
 *     def __len__(SecBuffer self):             # <<<<<<<<<<<<<<
 *         return self.c_value.cbBuffer if self.c_value else 0
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_3__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_2__len__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_2__len__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":489
 * 
 *     def __len__(SecBuffer self):
 *         return self.c_value.cbBuffer if self.c_value else 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(SecBuffer self):
 */
  if ((__pyx_v_self->c_value != 0)) {
    __pyx_t_1 = __pyx_v_self->c_value->cbBuffer;
  } else {
    __pyx_t_1 = 0;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":488
 *         self.c_value = ptr
 * 
 *     def __len__(SecBuffer self):             # <<<<<<<<<<<<<<
 *         return self.c_value.cbBuffer if self.c_value else 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":491
 *         return self.c_value.cbBuffer if self.c_value else 0
 * 
 *     def __repr__(SecBuffer self):             # <<<<<<<<<<<<<<
 *         return "<{0}.{1}(cbBuffer={2}, BufferType={3}, pvBuffer={4})>".format(
 *             type(self).__module__, type(self).__name__, self.c_value.cbBuffer, self.c_value.BufferType,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_4__repr__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_4__repr__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":492
 * 
 *     def __repr__(SecBuffer self):
 *         return "<{0}.{1}(cbBuffer={2}, BufferType={3}, pvBuffer={4})>".format(             # <<<<<<<<<<<<<<
 *             type(self).__module__, type(self).__name__, self.c_value.cbBuffer, self.c_value.BufferType,
 *             repr(self.buffer))
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_1_cbBuffer_2_BufferType_3_pvB, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "spnego/_sspi_raw/sspi.pyx":493
 *     def __repr__(SecBuffer self):
 *         return "<{0}.{1}(cbBuffer={2}, BufferType={3}, pvBuffer={4})>".format(
 *             type(self).__module__, type(self).__name__, self.c_value.cbBuffer, self.c_value.BufferType,             # <<<<<<<<<<<<<<
 *             repr(self.buffer))
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_module); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->c_value->cbBuffer); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->c_value->BufferType); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "spnego/_sspi_raw/sspi.pyx":494
 *         return "<{0}.{1}(cbBuffer={2}, BufferType={3}, pvBuffer={4})>".format(
 *             type(self).__module__, type(self).__name__, self.c_value.cbBuffer, self.c_value.BufferType,
 *             repr(self.buffer))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(SecBuffer self):
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_buffer); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyObject_Repr(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[6] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_8};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 492, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[6] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_8};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 492, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(5+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_10, 4+__pyx_t_9, __pyx_t_8);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_8 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":491
 *         return self.c_value.cbBuffer if self.c_value else 0
 * 
 *     def __repr__(SecBuffer self):             # <<<<<<<<<<<<<<
 *         return "<{0}.{1}(cbBuffer={2}, BufferType={3}, pvBuffer={4})>".format(
 *             type(self).__module__, type(self).__name__, self.c_value.cbBuffer, self.c_value.BufferType,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":496
 *             repr(self.buffer))
 * 
 *     def __str__(SecBuffer self):             # <<<<<<<<<<<<<<
 *         return "{0}(cbBuffer={1}, BufferType={2}, pvBuffer={3!r})".format(type(self).__name__, self.c_value.cbBuffer,
 *             self.c_value.BufferType, self.buffer)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_7__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_7__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_6__str__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_6__str__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":497
 * 
 *     def __str__(SecBuffer self):
 *         return "{0}(cbBuffer={1}, BufferType={2}, pvBuffer={3!r})".format(type(self).__name__, self.c_value.cbBuffer,             # <<<<<<<<<<<<<<
 *             self.c_value.BufferType, self.buffer)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_cbBuffer_1_BufferType_2_pvBuf, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->c_value->cbBuffer); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "spnego/_sspi_raw/sspi.pyx":498
 *     def __str__(SecBuffer self):
 *         return "{0}(cbBuffer={1}, BufferType={2}, pvBuffer={3!r})".format(type(self).__name__, self.c_value.cbBuffer,
 *             self.c_value.BufferType, self.buffer)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->c_value->BufferType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_buffer); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 497, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 497, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":496
 *             repr(self.buffer))
 * 
 *     def __str__(SecBuffer self):             # <<<<<<<<<<<<<<
 *         return "{0}(cbBuffer={1}, BufferType={2}, pvBuffer={3!r})".format(type(self).__name__, self.c_value.cbBuffer,
 *             self.c_value.BufferType, self.buffer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":501
 * 
 *     @property
 *     def buffer_type(SecBuffer self):             # <<<<<<<<<<<<<<
 *         return self.c_value.BufferType
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type___get__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":502
 *     @property
 *     def buffer_type(SecBuffer self):
 *         return self.c_value.BufferType             # <<<<<<<<<<<<<<
 * 
 *     @buffer_type.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->c_value->BufferType); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":501
 * 
 *     @property
 *     def buffer_type(SecBuffer self):             # <<<<<<<<<<<<<<
 *         return self.c_value.BufferType
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer.buffer_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":505
 * 
 *     @buffer_type.setter
 *     def buffer_type(SecBuffer self, unsigned long value):             # <<<<<<<<<<<<<<
 *         self.c_value.BufferType = value
 * 
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  unsigned long __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyInt_As_unsigned_long(__pyx_arg_value); if (unlikely((__pyx_v_value == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 505, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer.buffer_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type_2__set__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self), ((unsigned long)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, unsigned long __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":506
 *     @buffer_type.setter
 *     def buffer_type(SecBuffer self, unsigned long value):
 *         self.c_value.BufferType = value             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_v_self->c_value->BufferType = __pyx_v_value;

  /* "spnego/_sspi_raw/sspi.pyx":505
 * 
 *     @buffer_type.setter
 *     def buffer_type(SecBuffer self, unsigned long value):             # <<<<<<<<<<<<<<
 *         self.c_value.BufferType = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":509
 * 
 *     @property
 *     def buffer(SecBuffer self):             # <<<<<<<<<<<<<<
 *         if self.c_value.cbBuffer and self.c_value.pvBuffer:
 *             return (<char *>self.c_value.pvBuffer)[:self.c_value.cbBuffer]
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer___get__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":510
 *     @property
 *     def buffer(SecBuffer self):
 *         if self.c_value.cbBuffer and self.c_value.pvBuffer:             # <<<<<<<<<<<<<<
 *             return (<char *>self.c_value.pvBuffer)[:self.c_value.cbBuffer]
 *         elif self.c_value.pvBuffer:
 */
  __pyx_t_2 = (__pyx_v_self->c_value->cbBuffer != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->c_value->pvBuffer != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":511
 *     def buffer(SecBuffer self):
 *         if self.c_value.cbBuffer and self.c_value.pvBuffer:
 *             return (<char *>self.c_value.pvBuffer)[:self.c_value.cbBuffer]             # <<<<<<<<<<<<<<
 *         elif self.c_value.pvBuffer:
 *             return b""  # The size was 0 but pvBuffer was not a NULL pointer.
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyBytes_FromStringAndSize(((char *)__pyx_v_self->c_value->pvBuffer) + 0, __pyx_v_self->c_value->cbBuffer - 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "spnego/_sspi_raw/sspi.pyx":510
 *     @property
 *     def buffer(SecBuffer self):
 *         if self.c_value.cbBuffer and self.c_value.pvBuffer:             # <<<<<<<<<<<<<<
 *             return (<char *>self.c_value.pvBuffer)[:self.c_value.cbBuffer]
 *         elif self.c_value.pvBuffer:
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":512
 *         if self.c_value.cbBuffer and self.c_value.pvBuffer:
 *             return (<char *>self.c_value.pvBuffer)[:self.c_value.cbBuffer]
 *         elif self.c_value.pvBuffer:             # <<<<<<<<<<<<<<
 *             return b""  # The size was 0 but pvBuffer was not a NULL pointer.
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->c_value->pvBuffer != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":513
 *             return (<char *>self.c_value.pvBuffer)[:self.c_value.cbBuffer]
 *         elif self.c_value.pvBuffer:
 *             return b""  # The size was 0 but pvBuffer was not a NULL pointer.             # <<<<<<<<<<<<<<
 *         else:
 *             return
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_kp_b__11);
    __pyx_r = __pyx_kp_b__11;
    goto __pyx_L0;

    /* "spnego/_sspi_raw/sspi.pyx":512
 *         if self.c_value.cbBuffer and self.c_value.pvBuffer:
 *             return (<char *>self.c_value.pvBuffer)[:self.c_value.cbBuffer]
 *         elif self.c_value.pvBuffer:             # <<<<<<<<<<<<<<
 *             return b""  # The size was 0 but pvBuffer was not a NULL pointer.
 *         else:
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":515
 *             return b""  # The size was 0 but pvBuffer was not a NULL pointer.
 *         else:
 *             return             # <<<<<<<<<<<<<<
 * 
 *     @buffer.setter
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "spnego/_sspi_raw/sspi.pyx":509
 * 
 *     @property
 *     def buffer(SecBuffer self):             # <<<<<<<<<<<<<<
 *         if self.c_value.cbBuffer and self.c_value.pvBuffer:
 *             return (<char *>self.c_value.pvBuffer)[:self.c_value.cbBuffer]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer.buffer.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":518
 * 
 *     @buffer.setter
 *     def buffer(SecBuffer self, bytes value):             # <<<<<<<<<<<<<<
 *         value_len = len(value)
 *         self._alloc_buffer(value_len)
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyBytes_Type), 1, "value", 1))) __PYX_ERR(1, 518, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer_2__set__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, PyObject *__pyx_v_value) {
  Py_ssize_t __pyx_v_value_len;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":519
 *     @buffer.setter
 *     def buffer(SecBuffer self, bytes value):
 *         value_len = len(value)             # <<<<<<<<<<<<<<
 *         self._alloc_buffer(value_len)
 *         memcpy(self.c_value.pvBuffer, <char *>value, value_len)
 */
  if (unlikely(__pyx_v_value == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 519, __pyx_L1_error)
  }
  __pyx_t_1 = PyBytes_GET_SIZE(__pyx_v_value); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 519, __pyx_L1_error)
  __pyx_v_value_len = __pyx_t_1;

  /* "spnego/_sspi_raw/sspi.pyx":520
 *     def buffer(SecBuffer self, bytes value):
 *         value_len = len(value)
 *         self._alloc_buffer(value_len)             # <<<<<<<<<<<<<<
 *         memcpy(self.c_value.pvBuffer, <char *>value, value_len)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_alloc_buffer); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_value_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":521
 *         value_len = len(value)
 *         self._alloc_buffer(value_len)
 *         memcpy(self.c_value.pvBuffer, <char *>value, value_len)             # <<<<<<<<<<<<<<
 * 
 *     def _alloc_buffer(SecBuffer self, unsigned long length):
 */
  if (unlikely(__pyx_v_value == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 521, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(1, 521, __pyx_L1_error)
  (void)(memcpy(__pyx_v_self->c_value->pvBuffer, ((char *)__pyx_t_6), __pyx_v_value_len));

  /* "spnego/_sspi_raw/sspi.pyx":518
 * 
 *     @buffer.setter
 *     def buffer(SecBuffer self, bytes value):             # <<<<<<<<<<<<<<
 *         value_len = len(value)
 *         self._alloc_buffer(value_len)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer.buffer.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":523
 *         memcpy(self.c_value.pvBuffer, <char *>value, value_len)
 * 
 *     def _alloc_buffer(SecBuffer self, unsigned long length):             # <<<<<<<<<<<<<<
 *         self._dealloc_buffer()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_9_alloc_buffer(PyObject *__pyx_v_self, PyObject *__pyx_arg_length); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_9_alloc_buffer(PyObject *__pyx_v_self, PyObject *__pyx_arg_length) {
  unsigned long __pyx_v_length;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_alloc_buffer (wrapper)", 0);
  assert(__pyx_arg_length); {
    __pyx_v_length = __Pyx_PyInt_As_unsigned_long(__pyx_arg_length); if (unlikely((__pyx_v_length == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 523, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer._alloc_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_8_alloc_buffer(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self), ((unsigned long)__pyx_v_length));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_8_alloc_buffer(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, unsigned long __pyx_v_length) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  void *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_alloc_buffer", 0);

  /* "spnego/_sspi_raw/sspi.pyx":524
 * 
 *     def _alloc_buffer(SecBuffer self, unsigned long length):
 *         self._dealloc_buffer()             # <<<<<<<<<<<<<<
 * 
 *         # We store our allocated memory pointer so we know when we free we are freeing what we allocated.
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dealloc_buffer); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":527
 * 
 *         # We store our allocated memory pointer so we know when we free we are freeing what we allocated.
 *         self._p_buffer = malloc(length)             # <<<<<<<<<<<<<<
 *         if not self._p_buffer:
 *             raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover
 */
  __pyx_v_self->_p_buffer = malloc(__pyx_v_length);

  /* "spnego/_sspi_raw/sspi.pyx":528
 *         # We store our allocated memory pointer so we know when we free we are freeing what we allocated.
 *         self._p_buffer = malloc(length)
 *         if not self._p_buffer:             # <<<<<<<<<<<<<<
 *             raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover
 * 
 */
  __pyx_t_4 = ((!(__pyx_v_self->_p_buffer != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "spnego/_sspi_raw/sspi.pyx":529
 *         self._p_buffer = malloc(length)
 *         if not self._p_buffer:
 *             raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover             # <<<<<<<<<<<<<<
 * 
 *         self.c_value.pvBuffer = self._p_buffer
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 529, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":528
 *         # We store our allocated memory pointer so we know when we free we are freeing what we allocated.
 *         self._p_buffer = malloc(length)
 *         if not self._p_buffer:             # <<<<<<<<<<<<<<
 *             raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":531
 *             raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover
 * 
 *         self.c_value.pvBuffer = self._p_buffer             # <<<<<<<<<<<<<<
 *         self.c_value.cbBuffer = length
 * 
 */
  __pyx_t_5 = __pyx_v_self->_p_buffer;
  __pyx_v_self->c_value->pvBuffer = __pyx_t_5;

  /* "spnego/_sspi_raw/sspi.pyx":532
 * 
 *         self.c_value.pvBuffer = self._p_buffer
 *         self.c_value.cbBuffer = length             # <<<<<<<<<<<<<<
 * 
 *     def _dealloc_buffer(SecBuffer self):
 */
  __pyx_v_self->c_value->cbBuffer = __pyx_v_length;

  /* "spnego/_sspi_raw/sspi.pyx":523
 *         memcpy(self.c_value.pvBuffer, <char *>value, value_len)
 * 
 *     def _alloc_buffer(SecBuffer self, unsigned long length):             # <<<<<<<<<<<<<<
 *         self._dealloc_buffer()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer._alloc_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":534
 *         self.c_value.cbBuffer = length
 * 
 *     def _dealloc_buffer(SecBuffer self):             # <<<<<<<<<<<<<<
 *         # We do need to check if the actual C struct pvBuffer was allocated by Windows and call FreeContextBuffer on
 *         # that.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_11_dealloc_buffer(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_11_dealloc_buffer(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_dealloc_buffer (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_10_dealloc_buffer(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_10_dealloc_buffer(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_dealloc_buffer", 0);

  /* "spnego/_sspi_raw/sspi.pyx":537
 *         # We do need to check if the actual C struct pvBuffer was allocated by Windows and call FreeContextBuffer on
 *         # that.
 *         if self.c_value != NULL and self.c_value.pvBuffer != NULL and self.sys_alloc:             # <<<<<<<<<<<<<<
 *             FreeContextBuffer(self.c_value.pvBuffer)
 * 
 */
  __pyx_t_2 = ((__pyx_v_self->c_value != NULL) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_self->c_value->pvBuffer != NULL) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->sys_alloc); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 537, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":538
 *         # that.
 *         if self.c_value != NULL and self.c_value.pvBuffer != NULL and self.sys_alloc:
 *             FreeContextBuffer(self.c_value.pvBuffer)             # <<<<<<<<<<<<<<
 * 
 *         if self.c_value:
 */
    (void)(FreeContextBuffer(__pyx_v_self->c_value->pvBuffer));

    /* "spnego/_sspi_raw/sspi.pyx":537
 *         # We do need to check if the actual C struct pvBuffer was allocated by Windows and call FreeContextBuffer on
 *         # that.
 *         if self.c_value != NULL and self.c_value.pvBuffer != NULL and self.sys_alloc:             # <<<<<<<<<<<<<<
 *             FreeContextBuffer(self.c_value.pvBuffer)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":540
 *             FreeContextBuffer(self.c_value.pvBuffer)
 * 
 *         if self.c_value:             # <<<<<<<<<<<<<<
 *             self.c_value.pvBuffer = NULL
 * 
 */
  __pyx_t_1 = (__pyx_v_self->c_value != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":541
 * 
 *         if self.c_value:
 *             self.c_value.pvBuffer = NULL             # <<<<<<<<<<<<<<
 * 
 *         # Because the C struct pvBuffer may have a pointer set by Windows we track our allocated memory using an
 */
    __pyx_v_self->c_value->pvBuffer = NULL;

    /* "spnego/_sspi_raw/sspi.pyx":540
 *             FreeContextBuffer(self.c_value.pvBuffer)
 * 
 *         if self.c_value:             # <<<<<<<<<<<<<<
 *             self.c_value.pvBuffer = NULL
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":545
 *         # Because the C struct pvBuffer may have a pointer set by Windows we track our allocated memory using an
 *         # internal attribute which we free().
 *         if self._p_buffer:             # <<<<<<<<<<<<<<
 *             free(self._p_buffer)
 *         self._p_buffer = NULL
 */
  __pyx_t_1 = (__pyx_v_self->_p_buffer != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":546
 *         # internal attribute which we free().
 *         if self._p_buffer:
 *             free(self._p_buffer)             # <<<<<<<<<<<<<<
 *         self._p_buffer = NULL
 * 
 */
    free(__pyx_v_self->_p_buffer);

    /* "spnego/_sspi_raw/sspi.pyx":545
 *         # Because the C struct pvBuffer may have a pointer set by Windows we track our allocated memory using an
 *         # internal attribute which we free().
 *         if self._p_buffer:             # <<<<<<<<<<<<<<
 *             free(self._p_buffer)
 *         self._p_buffer = NULL
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":547
 *         if self._p_buffer:
 *             free(self._p_buffer)
 *         self._p_buffer = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(SecBuffer self):
 */
  __pyx_v_self->_p_buffer = NULL;

  /* "spnego/_sspi_raw/sspi.pyx":534
 *         self.c_value.cbBuffer = length
 * 
 *     def _dealloc_buffer(SecBuffer self):             # <<<<<<<<<<<<<<
 *         # We do need to check if the actual C struct pvBuffer was allocated by Windows and call FreeContextBuffer on
 *         # that.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer._dealloc_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":549
 *         self._p_buffer = NULL
 * 
 *     def __dealloc__(SecBuffer self):             # <<<<<<<<<<<<<<
 *         self._dealloc_buffer()
 * 
 */

/* Python wrapper */
static void __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_13__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_13__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_12__dealloc__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_12__dealloc__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":550
 * 
 *     def __dealloc__(SecBuffer self):
 *         self._dealloc_buffer()             # <<<<<<<<<<<<<<
 * 
 *         if self.c_value != NULL and self._is_owner:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dealloc_buffer); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":552
 *         self._dealloc_buffer()
 * 
 *         if self.c_value != NULL and self._is_owner:             # <<<<<<<<<<<<<<
 *             free(self.c_value)
 * 
 */
  __pyx_t_5 = ((__pyx_v_self->c_value != NULL) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_self->_is_owner); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 552, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "spnego/_sspi_raw/sspi.pyx":553
 * 
 *         if self.c_value != NULL and self._is_owner:
 *             free(self.c_value)             # <<<<<<<<<<<<<<
 * 
 *         self.c_value = NULL
 */
    free(__pyx_v_self->c_value);

    /* "spnego/_sspi_raw/sspi.pyx":552
 *         self._dealloc_buffer()
 * 
 *         if self.c_value != NULL and self._is_owner:             # <<<<<<<<<<<<<<
 *             free(self.c_value)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":555
 *             free(self.c_value)
 * 
 *         self.c_value = NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->c_value = NULL;

  /* "spnego/_sspi_raw/sspi.pyx":549
 *         self._p_buffer = NULL
 * 
 *     def __dealloc__(SecBuffer self):             # <<<<<<<<<<<<<<
 *         self._dealloc_buffer()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("spnego._sspi_raw.sspi.SecBuffer.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_14__reduce_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_16__setstate_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_9SecBuffer_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.SecBuffer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":560
 * cdef class _AuthIdentityBase:
 * 
 *     cdef void *__c_value__(_AuthIdentityBase self):             # <<<<<<<<<<<<<<
 *         return NULL
 * 
 */

static void *__pyx_f_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase___c_value__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_v_self) {
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__c_value__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":561
 * 
 *     cdef void *__c_value__(_AuthIdentityBase self):
 *         return NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":560
 * cdef class _AuthIdentityBase:
 * 
 *     cdef void *__c_value__(_AuthIdentityBase self):             # <<<<<<<<<<<<<<
 *         return NULL
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase___reduce_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase___reduce_cython__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__AuthIdentityBase, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__AuthIdentityBase, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle__AuthIdentityBase, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__AuthIdentityBase, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__AuthIdentityBase); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__AuthIdentityBase, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__AuthIdentityBase, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle__AuthIdentityBase, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__AuthIdentityBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle__AuthIdentityBase); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi._AuthIdentityBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__AuthIdentityBase, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__AuthIdentityBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase_2__setstate_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase_2__setstate_cython__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__AuthIdentityBase, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__AuthIdentityBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6spnego_9_sspi_raw_4sspi___pyx_unpickle__AuthIdentityBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__AuthIdentityBase, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__AuthIdentityBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi._AuthIdentityBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":566
 * cdef class WinNTAuthIdentity(_AuthIdentityBase):
 * 
 *     def __cinit__(WinNTAuthIdentity self, unicode username, unicode domain, unicode password):             # <<<<<<<<<<<<<<
 *         self.c_value.Flags = SEC_WINNT_AUTH_IDENTITY_UNICODE
 *         self.username = username
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_username = 0;
  PyObject *__pyx_v_domain = 0;
  PyObject *__pyx_v_password = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_username,&__pyx_n_s_domain,&__pyx_n_s_password,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_username)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_domain)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 566, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_password)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 566, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 566, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_username = ((PyObject*)values[0]);
    __pyx_v_domain = ((PyObject*)values[1]);
    __pyx_v_password = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 566, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_username), (&PyUnicode_Type), 1, "username", 1))) __PYX_ERR(1, 566, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_domain), (&PyUnicode_Type), 1, "domain", 1))) __PYX_ERR(1, 566, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_password), (&PyUnicode_Type), 1, "password", 1))) __PYX_ERR(1, 566, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity___cinit__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self), __pyx_v_username, __pyx_v_domain, __pyx_v_password);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity___cinit__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self, PyObject *__pyx_v_username, PyObject *__pyx_v_domain, PyObject *__pyx_v_password) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":567
 * 
 *     def __cinit__(WinNTAuthIdentity self, unicode username, unicode domain, unicode password):
 *         self.c_value.Flags = SEC_WINNT_AUTH_IDENTITY_UNICODE             # <<<<<<<<<<<<<<
 *         self.username = username
 *         self.domain = domain
 */
  __pyx_v_self->c_value.Flags = SEC_WINNT_AUTH_IDENTITY_UNICODE;

  /* "spnego/_sspi_raw/sspi.pyx":568
 *     def __cinit__(WinNTAuthIdentity self, unicode username, unicode domain, unicode password):
 *         self.c_value.Flags = SEC_WINNT_AUTH_IDENTITY_UNICODE
 *         self.username = username             # <<<<<<<<<<<<<<
 *         self.domain = domain
 *         self.password = password
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_username, __pyx_v_username) < 0) __PYX_ERR(1, 568, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":569
 *         self.c_value.Flags = SEC_WINNT_AUTH_IDENTITY_UNICODE
 *         self.username = username
 *         self.domain = domain             # <<<<<<<<<<<<<<
 *         self.password = password
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_domain, __pyx_v_domain) < 0) __PYX_ERR(1, 569, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":570
 *         self.username = username
 *         self.domain = domain
 *         self.password = password             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(WinNTAuthIdentity self):
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_password, __pyx_v_password) < 0) __PYX_ERR(1, 570, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":566
 * cdef class WinNTAuthIdentity(_AuthIdentityBase):
 * 
 *     def __cinit__(WinNTAuthIdentity self, unicode username, unicode domain, unicode password):             # <<<<<<<<<<<<<<
 *         self.c_value.Flags = SEC_WINNT_AUTH_IDENTITY_UNICODE
 *         self.username = username
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":572
 *         self.password = password
 * 
 *     def __repr__(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         domain = u"%s\\" % self.domain if self.domain else u""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_3__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_2__repr__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_2__repr__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self) {
  PyObject *__pyx_v_domain = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":573
 * 
 *     def __repr__(WinNTAuthIdentity self):
 *         domain = u"%s\\" % self.domain if self.domain else u""             # <<<<<<<<<<<<<<
 * 
 *         return to_native(u"<{0}.{1} {2}{3}>".format(type(self).__module__, type(self).__name__, domain,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_domain); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_domain); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_kp_u__11);
    __pyx_t_1 = __pyx_kp_u__11;
  }
  __pyx_v_domain = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":575
 *         domain = u"%s\\" % self.domain if self.domain else u""
 * 
 *         return to_native(u"<{0}.{1} {2}{3}>".format(type(self).__module__, type(self).__name__, domain,             # <<<<<<<<<<<<<<
 *             self.username or u""))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_to_native); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_1_2_3, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_module); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "spnego/_sspi_raw/sspi.pyx":576
 * 
 *         return to_native(u"<{0}.{1} {2}{3}>".format(type(self).__module__, type(self).__name__, domain,
 *             self.username or u""))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(WinNTAuthIdentity self):
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_username); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 576, __pyx_L1_error)
  if (!__pyx_t_3) {
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_9);
    __pyx_t_8 = __pyx_t_9;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_kp_u__11);
  __pyx_t_8 = __pyx_kp_u__11;
  __pyx_L3_bool_binop_done:;
  __pyx_t_9 = NULL;
  __pyx_t_10 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_10 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_t_6, __pyx_t_7, __pyx_v_domain, __pyx_t_8};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 575, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_t_6, __pyx_t_7, __pyx_v_domain, __pyx_t_8};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 575, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  {
    __pyx_t_11 = PyTuple_New(4+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__pyx_t_9) {
      __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_7);
    __Pyx_INCREF(__pyx_v_domain);
    __Pyx_GIVEREF(__pyx_v_domain);
    PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_v_domain);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_10, __pyx_t_8);
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":572
 *         self.password = password
 * 
 *     def __repr__(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         domain = u"%s\\" % self.domain if self.domain else u""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_domain);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":578
 *             self.username or u""))
 * 
 *     def __str__(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         domain = u"%s\\" % self.domain if self.domain else u""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_5__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_5__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_4__str__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_4__str__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self) {
  PyObject *__pyx_v_domain = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":579
 * 
 *     def __str__(WinNTAuthIdentity self):
 *         domain = u"%s\\" % self.domain if self.domain else u""             # <<<<<<<<<<<<<<
 * 
 *         return to_native(u"{0}{1}".format(domain, self.username or u""))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_domain); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_domain); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_kp_u__11);
    __pyx_t_1 = __pyx_kp_u__11;
  }
  __pyx_v_domain = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":581
 *         domain = u"%s\\" % self.domain if self.domain else u""
 * 
 *         return to_native(u"{0}{1}".format(domain, self.username or u""))             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_to_native); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_1, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_username); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 581, __pyx_L1_error)
  if (!__pyx_t_3) {
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_6 = __pyx_t_7;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_kp_u__11);
  __pyx_t_6 = __pyx_kp_u__11;
  __pyx_L3_bool_binop_done:;
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_domain, __pyx_t_6};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 581, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_domain, __pyx_t_6};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 581, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_domain);
    __Pyx_GIVEREF(__pyx_v_domain);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_domain);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":578
 *             self.username or u""))
 * 
 *     def __str__(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         domain = u"%s\\" % self.domain if self.domain else u""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_domain);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":584
 * 
 *     @property
 *     def username(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         return self._username.to_text(self.c_value.UserLength)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username___get__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":585
 *     @property
 *     def username(WinNTAuthIdentity self):
 *         return self._username.to_text(self.c_value.UserLength)             # <<<<<<<<<<<<<<
 * 
 *     @username.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_username), __pyx_n_s_to_text); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->c_value.UserLength); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":584
 * 
 *     @property
 *     def username(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         return self._username.to_text(self.c_value.UserLength)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.username.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":588
 * 
 *     @username.setter
 *     def username(WinNTAuthIdentity self, unicode value):             # <<<<<<<<<<<<<<
 *         self._username = WideChar.from_text(value)
 *         self.c_value.User = <void *>self._username.buffer
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyUnicode_Type), 1, "value", 1))) __PYX_ERR(1, 588, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username_2__set__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  long __pyx_t_4;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":589
 *     @username.setter
 *     def username(WinNTAuthIdentity self, unicode value):
 *         self._username = WideChar.from_text(value)             # <<<<<<<<<<<<<<
 *         self.c_value.User = <void *>self._username.buffer
 *         self.c_value.UserLength = max(self._username.length - 1, 0)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6spnego_9_sspi_raw_4text_WideChar), __pyx_n_s_from_text); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_value);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6spnego_9_sspi_raw_4text_WideChar))))) __PYX_ERR(1, 589, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_username);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_username));
  __pyx_v_self->_username = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":590
 *     def username(WinNTAuthIdentity self, unicode value):
 *         self._username = WideChar.from_text(value)
 *         self.c_value.User = <void *>self._username.buffer             # <<<<<<<<<<<<<<
 *         self.c_value.UserLength = max(self._username.length - 1, 0)
 * 
 */
  __pyx_v_self->c_value.User = ((void *)__pyx_v_self->_username->buffer);

  /* "spnego/_sspi_raw/sspi.pyx":591
 *         self._username = WideChar.from_text(value)
 *         self.c_value.User = <void *>self._username.buffer
 *         self.c_value.UserLength = max(self._username.length - 1, 0)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = (__pyx_v_self->_username->length - 1);
  if (((__pyx_t_4 > __pyx_t_5) != 0)) {
    __pyx_t_6 = __pyx_t_4;
  } else {
    __pyx_t_6 = __pyx_t_5;
  }
  __pyx_v_self->c_value.UserLength = __pyx_t_6;

  /* "spnego/_sspi_raw/sspi.pyx":588
 * 
 *     @username.setter
 *     def username(WinNTAuthIdentity self, unicode value):             # <<<<<<<<<<<<<<
 *         self._username = WideChar.from_text(value)
 *         self.c_value.User = <void *>self._username.buffer
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.username.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":594
 * 
 *     @property
 *     def domain(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         return self._domain.to_text(self.c_value.DomainLength)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain___get__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":595
 *     @property
 *     def domain(WinNTAuthIdentity self):
 *         return self._domain.to_text(self.c_value.DomainLength)             # <<<<<<<<<<<<<<
 * 
 *     @domain.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_domain), __pyx_n_s_to_text); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->c_value.DomainLength); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":594
 * 
 *     @property
 *     def domain(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         return self._domain.to_text(self.c_value.DomainLength)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.domain.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":598
 * 
 *     @domain.setter
 *     def domain(WinNTAuthIdentity self, unicode value):             # <<<<<<<<<<<<<<
 *         self._domain = WideChar.from_text(value)
 *         self.c_value.Domain = <void *>self._domain.buffer
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyUnicode_Type), 1, "value", 1))) __PYX_ERR(1, 598, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain_2__set__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  long __pyx_t_4;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":599
 *     @domain.setter
 *     def domain(WinNTAuthIdentity self, unicode value):
 *         self._domain = WideChar.from_text(value)             # <<<<<<<<<<<<<<
 *         self.c_value.Domain = <void *>self._domain.buffer
 *         self.c_value.DomainLength = max(self._domain.length - 1, 0)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6spnego_9_sspi_raw_4text_WideChar), __pyx_n_s_from_text); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_value);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6spnego_9_sspi_raw_4text_WideChar))))) __PYX_ERR(1, 599, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_domain);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_domain));
  __pyx_v_self->_domain = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":600
 *     def domain(WinNTAuthIdentity self, unicode value):
 *         self._domain = WideChar.from_text(value)
 *         self.c_value.Domain = <void *>self._domain.buffer             # <<<<<<<<<<<<<<
 *         self.c_value.DomainLength = max(self._domain.length - 1, 0)
 * 
 */
  __pyx_v_self->c_value.Domain = ((void *)__pyx_v_self->_domain->buffer);

  /* "spnego/_sspi_raw/sspi.pyx":601
 *         self._domain = WideChar.from_text(value)
 *         self.c_value.Domain = <void *>self._domain.buffer
 *         self.c_value.DomainLength = max(self._domain.length - 1, 0)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = (__pyx_v_self->_domain->length - 1);
  if (((__pyx_t_4 > __pyx_t_5) != 0)) {
    __pyx_t_6 = __pyx_t_4;
  } else {
    __pyx_t_6 = __pyx_t_5;
  }
  __pyx_v_self->c_value.DomainLength = __pyx_t_6;

  /* "spnego/_sspi_raw/sspi.pyx":598
 * 
 *     @domain.setter
 *     def domain(WinNTAuthIdentity self, unicode value):             # <<<<<<<<<<<<<<
 *         self._domain = WideChar.from_text(value)
 *         self.c_value.Domain = <void *>self._domain.buffer
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.domain.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":604
 * 
 *     @property
 *     def password(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         return self._password.to_text(self.c_value.PasswordLength)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password___get__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password___get__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":605
 *     @property
 *     def password(WinNTAuthIdentity self):
 *         return self._password.to_text(self.c_value.PasswordLength)             # <<<<<<<<<<<<<<
 * 
 *     @password.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_password), __pyx_n_s_to_text); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->c_value.PasswordLength); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":604
 * 
 *     @property
 *     def password(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         return self._password.to_text(self.c_value.PasswordLength)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.password.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":608
 * 
 *     @password.setter
 *     def password(WinNTAuthIdentity self, unicode value):             # <<<<<<<<<<<<<<
 *         self._password = WideChar.from_text(value)
 *         self.c_value.Password = <void *>self._password.buffer
 */

/* Python wrapper */
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyUnicode_Type), 1, "value", 1))) __PYX_ERR(1, 608, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password_2__set__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password_2__set__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  long __pyx_t_4;
  long __pyx_t_5;
  long __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":609
 *     @password.setter
 *     def password(WinNTAuthIdentity self, unicode value):
 *         self._password = WideChar.from_text(value)             # <<<<<<<<<<<<<<
 *         self.c_value.Password = <void *>self._password.buffer
 *         self.c_value.PasswordLength = max(self._password.length - 1, 0)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6spnego_9_sspi_raw_4text_WideChar), __pyx_n_s_from_text); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_value);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6spnego_9_sspi_raw_4text_WideChar))))) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_password);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_password));
  __pyx_v_self->_password = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":610
 *     def password(WinNTAuthIdentity self, unicode value):
 *         self._password = WideChar.from_text(value)
 *         self.c_value.Password = <void *>self._password.buffer             # <<<<<<<<<<<<<<
 *         self.c_value.PasswordLength = max(self._password.length - 1, 0)
 * 
 */
  __pyx_v_self->c_value.Password = ((void *)__pyx_v_self->_password->buffer);

  /* "spnego/_sspi_raw/sspi.pyx":611
 *         self._password = WideChar.from_text(value)
 *         self.c_value.Password = <void *>self._password.buffer
 *         self.c_value.PasswordLength = max(self._password.length - 1, 0)             # <<<<<<<<<<<<<<
 * 
 *     cdef void *__c_value__(WinNTAuthIdentity self):
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = (__pyx_v_self->_password->length - 1);
  if (((__pyx_t_4 > __pyx_t_5) != 0)) {
    __pyx_t_6 = __pyx_t_4;
  } else {
    __pyx_t_6 = __pyx_t_5;
  }
  __pyx_v_self->c_value.PasswordLength = __pyx_t_6;

  /* "spnego/_sspi_raw/sspi.pyx":608
 * 
 *     @password.setter
 *     def password(WinNTAuthIdentity self, unicode value):             # <<<<<<<<<<<<<<
 *         self._password = WideChar.from_text(value)
 *         self.c_value.Password = <void *>self._password.buffer
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.password.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":613
 *         self.c_value.PasswordLength = max(self._password.length - 1, 0)
 * 
 *     cdef void *__c_value__(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         return &self.c_value
 * 
 */

static void *__pyx_f_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity___c_value__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self) {
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__c_value__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":614
 * 
 *     cdef void *__c_value__(WinNTAuthIdentity self):
 *         return &self.c_value             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(WinNTAuthIdentity self):
 */
  __pyx_r = (&__pyx_v_self->c_value);
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":613
 *         self.c_value.PasswordLength = max(self._password.length - 1, 0)
 * 
 *     cdef void *__c_value__(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         return &self.c_value
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":616
 *         return &self.c_value
 * 
 *     def __dealloc__(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         self.username = None
 *         self.domain = None
 */

/* Python wrapper */
static void __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_7__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_7__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6__dealloc__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6__dealloc__(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "spnego/_sspi_raw/sspi.pyx":617
 * 
 *     def __dealloc__(WinNTAuthIdentity self):
 *         self.username = None             # <<<<<<<<<<<<<<
 *         self.domain = None
 *         self.password = None
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_username, Py_None) < 0) __PYX_ERR(1, 617, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":618
 *     def __dealloc__(WinNTAuthIdentity self):
 *         self.username = None
 *         self.domain = None             # <<<<<<<<<<<<<<
 *         self.password = None
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_domain, Py_None) < 0) __PYX_ERR(1, 618, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":619
 *         self.username = None
 *         self.domain = None
 *         self.password = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_password, Py_None) < 0) __PYX_ERR(1, 619, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":616
 *         return &self.c_value
 * 
 *     def __dealloc__(WinNTAuthIdentity self):             # <<<<<<<<<<<<<<
 *         self.username = None
 *         self.domain = None
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("spnego._sspi_raw.sspi.WinNTAuthIdentity.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8__reduce_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_10__setstate_cython__(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.WinNTAuthIdentity.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":622
 * 
 * 
 * def accept_security_context(Credential credential not None, SecurityContext context not None,             # <<<<<<<<<<<<<<
 *     SecBufferDesc input_buffer=None, unsigned long context_req=0, unsigned long target_data_rep=SECURITY_NATIVE_DREP,
 *     SecBufferDesc output_buffer=None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_1accept_security_context(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_1accept_security_context = {"accept_security_context", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6spnego_9_sspi_raw_4sspi_1accept_security_context, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_1accept_security_context(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_credential = 0;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context = 0;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_input_buffer = 0;
  unsigned long __pyx_v_context_req;
  unsigned long __pyx_v_target_data_rep;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_output_buffer = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept_security_context (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_credential,&__pyx_n_s_context,&__pyx_n_s_input_buffer,&__pyx_n_s_context_req,&__pyx_n_s_target_data_rep,&__pyx_n_s_output_buffer,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "spnego/_sspi_raw/sspi.pyx":623
 * 
 * def accept_security_context(Credential credential not None, SecurityContext context not None,
 *     SecBufferDesc input_buffer=None, unsigned long context_req=0, unsigned long target_data_rep=SECURITY_NATIVE_DREP,             # <<<<<<<<<<<<<<
 *     SecBufferDesc output_buffer=None):
 * 
 */
    values[2] = (PyObject *)((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)Py_None);

    /* "spnego/_sspi_raw/sspi.pyx":624
 * def accept_security_context(Credential credential not None, SecurityContext context not None,
 *     SecBufferDesc input_buffer=None, unsigned long context_req=0, unsigned long target_data_rep=SECURITY_NATIVE_DREP,
 *     SecBufferDesc output_buffer=None):             # <<<<<<<<<<<<<<
 * 
 *     cdef PCtxtHandle input_context = &context.handle if context.init else NULL
 */
    values[5] = (PyObject *)((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_credential)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept_security_context", 0, 2, 6, 1); __PYX_ERR(1, 622, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_input_buffer);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context_req);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target_data_rep);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_buffer);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept_security_context") < 0)) __PYX_ERR(1, 622, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_credential = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *)values[0]);
    __pyx_v_context = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)values[1]);
    __pyx_v_input_buffer = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)values[2]);
    if (values[3]) {
      __pyx_v_context_req = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_context_req == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 623, __pyx_L3_error)
    } else {
      __pyx_v_context_req = ((unsigned long)0);
    }
    if (values[4]) {
      __pyx_v_target_data_rep = __Pyx_PyInt_As_unsigned_long(values[4]); if (unlikely((__pyx_v_target_data_rep == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 623, __pyx_L3_error)
    } else {
      __pyx_v_target_data_rep = __pyx_k__16;
    }
    __pyx_v_output_buffer = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)values[5]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept_security_context", 0, 2, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 622, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.accept_security_context", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_credential), __pyx_ptype_6spnego_9_sspi_raw_4sspi_Credential, 0, "credential", 0))) __PYX_ERR(1, 622, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext, 0, "context", 0))) __PYX_ERR(1, 622, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_input_buffer), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBufferDesc, 1, "input_buffer", 0))) __PYX_ERR(1, 623, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_output_buffer), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBufferDesc, 1, "output_buffer", 0))) __PYX_ERR(1, 624, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_accept_security_context(__pyx_self, __pyx_v_credential, __pyx_v_context, __pyx_v_input_buffer, __pyx_v_context_req, __pyx_v_target_data_rep, __pyx_v_output_buffer);

  /* "spnego/_sspi_raw/sspi.pyx":622
 * 
 * 
 * def accept_security_context(Credential credential not None, SecurityContext context not None,             # <<<<<<<<<<<<<<
 *     SecBufferDesc input_buffer=None, unsigned long context_req=0, unsigned long target_data_rep=SECURITY_NATIVE_DREP,
 *     SecBufferDesc output_buffer=None):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_accept_security_context(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_credential, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_input_buffer, unsigned long __pyx_v_context_req, unsigned long __pyx_v_target_data_rep, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_output_buffer) {
  PCtxtHandle __pyx_v_input_context;
  PSecBufferDesc __pyx_v_input;
  PSecBufferDesc __pyx_v_output;
  SECURITY_INTEGER __pyx_v_expiry;
  PyObject *__pyx_v_buffer = NULL;
  SECURITY_STATUS __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  CtxtHandle *__pyx_t_1;
  PSecBufferDesc __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  Py_ssize_t __pyx_t_8;
  SECURITY_STATUS __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("accept_security_context", 0);

  /* "spnego/_sspi_raw/sspi.pyx":626
 *     SecBufferDesc output_buffer=None):
 * 
 *     cdef PCtxtHandle input_context = &context.handle if context.init else NULL             # <<<<<<<<<<<<<<
 *     cdef PSecBufferDesc input = input_buffer.__c_value__() if input_buffer else NULL
 *     cdef PSecBufferDesc output = output_buffer.__c_value__() if output_buffer else NULL
 */
  if ((__pyx_v_context->init != 0)) {
    __pyx_t_1 = (&__pyx_v_context->handle);
  } else {
    __pyx_t_1 = NULL;
  }
  __pyx_v_input_context = __pyx_t_1;

  /* "spnego/_sspi_raw/sspi.pyx":627
 * 
 *     cdef PCtxtHandle input_context = &context.handle if context.init else NULL
 *     cdef PSecBufferDesc input = input_buffer.__c_value__() if input_buffer else NULL             # <<<<<<<<<<<<<<
 *     cdef PSecBufferDesc output = output_buffer.__c_value__() if output_buffer else NULL
 *     cdef SECURITY_INTEGER expiry
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_input_buffer)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 627, __pyx_L1_error)
  if (__pyx_t_3) {
    __pyx_t_2 = ((struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_input_buffer->__pyx_vtab)->__pyx___c_value__(__pyx_v_input_buffer);
  } else {
    __pyx_t_2 = NULL;
  }
  __pyx_v_input = __pyx_t_2;

  /* "spnego/_sspi_raw/sspi.pyx":628
 *     cdef PCtxtHandle input_context = &context.handle if context.init else NULL
 *     cdef PSecBufferDesc input = input_buffer.__c_value__() if input_buffer else NULL
 *     cdef PSecBufferDesc output = output_buffer.__c_value__() if output_buffer else NULL             # <<<<<<<<<<<<<<
 *     cdef SECURITY_INTEGER expiry
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_output_buffer)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 628, __pyx_L1_error)
  if (__pyx_t_3) {
    __pyx_t_2 = ((struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_output_buffer->__pyx_vtab)->__pyx___c_value__(__pyx_v_output_buffer);
  } else {
    __pyx_t_2 = NULL;
  }
  __pyx_v_output = __pyx_t_2;

  /* "spnego/_sspi_raw/sspi.pyx":631
 *     cdef SECURITY_INTEGER expiry
 * 
 *     for buffer in (output_buffer or []):             # <<<<<<<<<<<<<<
 *         if len((<SecBuffer>buffer)) == 0:
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_output_buffer)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 631, __pyx_L1_error)
  if (!__pyx_t_3) {
  } else {
    __Pyx_INCREF(((PyObject *)__pyx_v_output_buffer));
    __pyx_t_4 = ((PyObject *)__pyx_v_output_buffer);
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_4 = __pyx_t_5;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_L5_bool_binop_done:;
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 631, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(1, 631, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 631, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(1, 631, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 631, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_7(__pyx_t_5);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 631, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_buffer, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "spnego/_sspi_raw/sspi.pyx":632
 * 
 *     for buffer in (output_buffer or []):
 *         if len((<SecBuffer>buffer)) == 0:             # <<<<<<<<<<<<<<
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY
 *             (<SecBuffer>buffer).sys_alloc = 1
 */
    __pyx_t_8 = PyObject_Length(__pyx_v_buffer); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(1, 632, __pyx_L1_error)
    __pyx_t_3 = ((__pyx_t_8 == 0) != 0);
    if (__pyx_t_3) {

      /* "spnego/_sspi_raw/sspi.pyx":633
 *     for buffer in (output_buffer or []):
 *         if len((<SecBuffer>buffer)) == 0:
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY             # <<<<<<<<<<<<<<
 *             (<SecBuffer>buffer).sys_alloc = 1
 * 
 */
      __pyx_v_context_req = (__pyx_v_context_req | ISC_REQ_ALLOCATE_MEMORY);

      /* "spnego/_sspi_raw/sspi.pyx":634
 *         if len((<SecBuffer>buffer)) == 0:
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY
 *             (<SecBuffer>buffer).sys_alloc = 1             # <<<<<<<<<<<<<<
 * 
 *     res = AcceptSecurityContext(&credential.handle, input_context, input, context_req, target_data_rep,
 */
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      __Pyx_GOTREF(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_buffer)->sys_alloc);
      __Pyx_DECREF(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_buffer)->sys_alloc);
      ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_buffer)->sys_alloc = __pyx_int_1;

      /* "spnego/_sspi_raw/sspi.pyx":632
 * 
 *     for buffer in (output_buffer or []):
 *         if len((<SecBuffer>buffer)) == 0:             # <<<<<<<<<<<<<<
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY
 *             (<SecBuffer>buffer).sys_alloc = 1
 */
    }

    /* "spnego/_sspi_raw/sspi.pyx":631
 *     cdef SECURITY_INTEGER expiry
 * 
 *     for buffer in (output_buffer or []):             # <<<<<<<<<<<<<<
 *         if len((<SecBuffer>buffer)) == 0:
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":636
 *             (<SecBuffer>buffer).sys_alloc = 1
 * 
 *     res = AcceptSecurityContext(&credential.handle, input_context, input, context_req, target_data_rep,             # <<<<<<<<<<<<<<
 *         &context.handle, output, &context.context_attr, &expiry)
 * 
 */
  __pyx_v_res = AcceptSecurityContext((&__pyx_v_credential->handle), __pyx_v_input_context, __pyx_v_input, __pyx_v_context_req, __pyx_v_target_data_rep, (&__pyx_v_context->handle), __pyx_v_output, (&__pyx_v_context->context_attr), (&__pyx_v_expiry));

  /* "spnego/_sspi_raw/sspi.pyx":639
 *         &context.handle, output, &context.context_attr, &expiry)
 * 
 *     if res in [_SEC_I_COMPLETE_AND_CONTINUE, _SEC_I_COMPLETE_NEEDED]:             # <<<<<<<<<<<<<<
 *         res = CompleteAuthToken(&context.handle, output)
 * 
 */
  __pyx_t_9 = __pyx_v_res;
  __pyx_t_10 = ((__pyx_t_9 == SEC_I_COMPLETE_AND_CONTINUE) != 0);
  if (!__pyx_t_10) {
  } else {
    __pyx_t_3 = __pyx_t_10;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_10 = ((__pyx_t_9 == SEC_I_COMPLETE_NEEDED) != 0);
  __pyx_t_3 = __pyx_t_10;
  __pyx_L9_bool_binop_done:;
  __pyx_t_10 = (__pyx_t_3 != 0);
  if (__pyx_t_10) {

    /* "spnego/_sspi_raw/sspi.pyx":640
 * 
 *     if res in [_SEC_I_COMPLETE_AND_CONTINUE, _SEC_I_COMPLETE_NEEDED]:
 *         res = CompleteAuthToken(&context.handle, output)             # <<<<<<<<<<<<<<
 * 
 *     if res not in [_SEC_I_CONTINUE_NEEDED, _SEC_E_OK]:
 */
    __pyx_v_res = CompleteAuthToken((&__pyx_v_context->handle), __pyx_v_output);

    /* "spnego/_sspi_raw/sspi.pyx":639
 *         &context.handle, output, &context.context_attr, &expiry)
 * 
 *     if res in [_SEC_I_COMPLETE_AND_CONTINUE, _SEC_I_COMPLETE_NEEDED]:             # <<<<<<<<<<<<<<
 *         res = CompleteAuthToken(&context.handle, output)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":642
 *         res = CompleteAuthToken(&context.handle, output)
 * 
 *     if res not in [_SEC_I_CONTINUE_NEEDED, _SEC_E_OK]:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  __pyx_t_9 = __pyx_v_res;
  __pyx_t_3 = ((__pyx_t_9 != SEC_I_CONTINUE_NEEDED) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_10 = __pyx_t_3;
    goto __pyx_L12_bool_binop_done;
  }
  __pyx_t_3 = ((__pyx_t_9 != SEC_E_OK) != 0);
  __pyx_t_10 = __pyx_t_3;
  __pyx_L12_bool_binop_done:;
  __pyx_t_3 = (__pyx_t_10 != 0);
  if (__pyx_t_3) {

    /* "spnego/_sspi_raw/sspi.pyx":643
 * 
 *     if res not in [_SEC_I_CONTINUE_NEEDED, _SEC_E_OK]:
 *         PyErr_SetFromWindowsErr(res)             # <<<<<<<<<<<<<<
 * 
 *     context.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart
 */
    __pyx_t_11 = PyErr_SetFromWindowsErr(__pyx_v_res); if (unlikely(__pyx_t_11 == ((PyObject *)NULL))) __PYX_ERR(1, 643, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":642
 *         res = CompleteAuthToken(&context.handle, output)
 * 
 *     if res not in [_SEC_I_CONTINUE_NEEDED, _SEC_E_OK]:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":645
 *         PyErr_SetFromWindowsErr(res)
 * 
 *     context.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart             # <<<<<<<<<<<<<<
 *     context.init = 1
 * 
 */
  __pyx_v_context->expiry = ((((unsigned PY_LONG_LONG)__pyx_v_expiry.HighPart) << 32) | __pyx_v_expiry.LowPart);

  /* "spnego/_sspi_raw/sspi.pyx":646
 * 
 *     context.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart
 *     context.init = 1             # <<<<<<<<<<<<<<
 * 
 *     return res
 */
  __pyx_v_context->init = 1;

  /* "spnego/_sspi_raw/sspi.pyx":648
 *     context.init = 1
 * 
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyInt_From_LONG(__pyx_v_res); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":622
 * 
 * 
 * def accept_security_context(Credential credential not None, SecurityContext context not None,             # <<<<<<<<<<<<<<
 *     SecBufferDesc input_buffer=None, unsigned long context_req=0, unsigned long target_data_rep=SECURITY_NATIVE_DREP,
 *     SecBufferDesc output_buffer=None):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.accept_security_context", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":651
 * 
 * 
 * def acquire_credentials_handle(unicode principal, unicode package not None,             # <<<<<<<<<<<<<<
 *     unsigned long credential_use=SECPKG_CRED_OUTBOUND, _AuthIdentityBase auth_data=None):
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_3acquire_credentials_handle(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_3acquire_credentials_handle = {"acquire_credentials_handle", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6spnego_9_sspi_raw_4sspi_3acquire_credentials_handle, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_3acquire_credentials_handle(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_principal = 0;
  PyObject *__pyx_v_package = 0;
  unsigned long __pyx_v_credential_use;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_v_auth_data = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("acquire_credentials_handle (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_principal,&__pyx_n_s_package,&__pyx_n_s_credential_use,&__pyx_n_s_auth_data,0};
    PyObject* values[4] = {0,0,0,0};

    /* "spnego/_sspi_raw/sspi.pyx":652
 * 
 * def acquire_credentials_handle(unicode principal, unicode package not None,
 *     unsigned long credential_use=SECPKG_CRED_OUTBOUND, _AuthIdentityBase auth_data=None):             # <<<<<<<<<<<<<<
 * 
 *     cdef WideChar w_principal = WideChar.from_text(principal)
 */
    values[3] = (PyObject *)((struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_principal)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_package)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("acquire_credentials_handle", 0, 2, 4, 1); __PYX_ERR(1, 651, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_credential_use);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_auth_data);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "acquire_credentials_handle") < 0)) __PYX_ERR(1, 651, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_principal = ((PyObject*)values[0]);
    __pyx_v_package = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_credential_use = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_credential_use == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 652, __pyx_L3_error)
    } else {
      __pyx_v_credential_use = __pyx_k__17;
    }
    __pyx_v_auth_data = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *)values[3]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("acquire_credentials_handle", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 651, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.acquire_credentials_handle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_principal), (&PyUnicode_Type), 1, "principal", 1))) __PYX_ERR(1, 651, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_package), (&PyUnicode_Type), 0, "package", 1))) __PYX_ERR(1, 651, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_auth_data), __pyx_ptype_6spnego_9_sspi_raw_4sspi__AuthIdentityBase, 1, "auth_data", 0))) __PYX_ERR(1, 652, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_2acquire_credentials_handle(__pyx_self, __pyx_v_principal, __pyx_v_package, __pyx_v_credential_use, __pyx_v_auth_data);

  /* "spnego/_sspi_raw/sspi.pyx":651
 * 
 * 
 * def acquire_credentials_handle(unicode principal, unicode package not None,             # <<<<<<<<<<<<<<
 *     unsigned long credential_use=SECPKG_CRED_OUTBOUND, _AuthIdentityBase auth_data=None):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_2acquire_credentials_handle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_principal, PyObject *__pyx_v_package, unsigned long __pyx_v_credential_use, struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_v_auth_data) {
  struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *__pyx_v_w_principal = 0;
  struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *__pyx_v_w_package = 0;
  void *__pyx_v_p_auth_data;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_cred = 0;
  SECURITY_INTEGER __pyx_v_expiry;
  SECURITY_STATUS __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  void *__pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("acquire_credentials_handle", 0);

  /* "spnego/_sspi_raw/sspi.pyx":654
 *     unsigned long credential_use=SECPKG_CRED_OUTBOUND, _AuthIdentityBase auth_data=None):
 * 
 *     cdef WideChar w_principal = WideChar.from_text(principal)             # <<<<<<<<<<<<<<
 *     cdef WideChar w_package = WideChar.from_text(package)
 *     cdef void *p_auth_data = auth_data.__c_value__() if auth_data else NULL
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6spnego_9_sspi_raw_4text_WideChar), __pyx_n_s_from_text); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_principal) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_principal);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6spnego_9_sspi_raw_4text_WideChar))))) __PYX_ERR(1, 654, __pyx_L1_error)
  __pyx_v_w_principal = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":655
 * 
 *     cdef WideChar w_principal = WideChar.from_text(principal)
 *     cdef WideChar w_package = WideChar.from_text(package)             # <<<<<<<<<<<<<<
 *     cdef void *p_auth_data = auth_data.__c_value__() if auth_data else NULL
 *     cdef Credential cred = Credential()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6spnego_9_sspi_raw_4text_WideChar), __pyx_n_s_from_text); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_package) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_package);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6spnego_9_sspi_raw_4text_WideChar))))) __PYX_ERR(1, 655, __pyx_L1_error)
  __pyx_v_w_package = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":656
 *     cdef WideChar w_principal = WideChar.from_text(principal)
 *     cdef WideChar w_package = WideChar.from_text(package)
 *     cdef void *p_auth_data = auth_data.__c_value__() if auth_data else NULL             # <<<<<<<<<<<<<<
 *     cdef Credential cred = Credential()
 *     cdef SECURITY_INTEGER expiry
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_auth_data)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 656, __pyx_L1_error)
  if (__pyx_t_5) {
    __pyx_t_4 = ((struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *)__pyx_v_auth_data->__pyx_vtab)->__pyx___c_value__(__pyx_v_auth_data);
  } else {
    __pyx_t_4 = NULL;
  }
  __pyx_v_p_auth_data = __pyx_t_4;

  /* "spnego/_sspi_raw/sspi.pyx":657
 *     cdef WideChar w_package = WideChar.from_text(package)
 *     cdef void *p_auth_data = auth_data.__c_value__() if auth_data else NULL
 *     cdef Credential cred = Credential()             # <<<<<<<<<<<<<<
 *     cdef SECURITY_INTEGER expiry
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6spnego_9_sspi_raw_4sspi_Credential)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cred = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":660
 *     cdef SECURITY_INTEGER expiry
 * 
 *     res = AcquireCredentialsHandleW(w_principal.buffer, w_package.buffer, credential_use, NULL, p_auth_data, NULL,             # <<<<<<<<<<<<<<
 *         NULL, &cred.handle, &expiry)
 * 
 */
  __pyx_v_res = AcquireCredentialsHandleW(__pyx_v_w_principal->buffer, __pyx_v_w_package->buffer, __pyx_v_credential_use, NULL, __pyx_v_p_auth_data, NULL, NULL, (&__pyx_v_cred->handle), (&__pyx_v_expiry));

  /* "spnego/_sspi_raw/sspi.pyx":663
 *         NULL, &cred.handle, &expiry)
 * 
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  __pyx_t_5 = ((__pyx_v_res != SEC_E_OK) != 0);
  if (__pyx_t_5) {

    /* "spnego/_sspi_raw/sspi.pyx":664
 * 
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)             # <<<<<<<<<<<<<<
 * 
 *     cred.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart
 */
    __pyx_t_6 = PyErr_SetFromWindowsErr(__pyx_v_res); if (unlikely(__pyx_t_6 == ((PyObject *)NULL))) __PYX_ERR(1, 664, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":663
 *         NULL, &cred.handle, &expiry)
 * 
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":666
 *         PyErr_SetFromWindowsErr(res)
 * 
 *     cred.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart             # <<<<<<<<<<<<<<
 * 
 *     return cred
 */
  __pyx_v_cred->expiry = ((((unsigned PY_LONG_LONG)__pyx_v_expiry.HighPart) << 32) | __pyx_v_expiry.LowPart);

  /* "spnego/_sspi_raw/sspi.pyx":668
 *     cred.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart
 * 
 *     return cred             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_cred));
  __pyx_r = ((PyObject *)__pyx_v_cred);
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":651
 * 
 * 
 * def acquire_credentials_handle(unicode principal, unicode package not None,             # <<<<<<<<<<<<<<
 *     unsigned long credential_use=SECPKG_CRED_OUTBOUND, _AuthIdentityBase auth_data=None):
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.acquire_credentials_handle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_w_principal);
  __Pyx_XDECREF((PyObject *)__pyx_v_w_package);
  __Pyx_XDECREF((PyObject *)__pyx_v_cred);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":671
 * 
 * 
 * def decrypt_message(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):             # <<<<<<<<<<<<<<
 *     cdef unsigned long qop = 0;
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_5decrypt_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_5decrypt_message = {"decrypt_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6spnego_9_sspi_raw_4sspi_5decrypt_message, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_5decrypt_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context = 0;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message = 0;
  unsigned long __pyx_v_seq_no;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("decrypt_message (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,&__pyx_n_s_message,&__pyx_n_s_seq_no,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("decrypt_message", 0, 2, 3, 1); __PYX_ERR(1, 671, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seq_no);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "decrypt_message") < 0)) __PYX_ERR(1, 671, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_context = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)values[0]);
    __pyx_v_message = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)values[1]);
    if (values[2]) {
      __pyx_v_seq_no = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_seq_no == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 671, __pyx_L3_error)
    } else {
      __pyx_v_seq_no = ((unsigned long)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("decrypt_message", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 671, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.decrypt_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext, 0, "context", 0))) __PYX_ERR(1, 671, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBufferDesc, 0, "message", 0))) __PYX_ERR(1, 671, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_4decrypt_message(__pyx_self, __pyx_v_context, __pyx_v_message, __pyx_v_seq_no);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_4decrypt_message(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message, unsigned long __pyx_v_seq_no) {
  unsigned long __pyx_v_qop;
  SECURITY_STATUS __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decrypt_message", 0);

  /* "spnego/_sspi_raw/sspi.pyx":672
 * 
 * def decrypt_message(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):
 *     cdef unsigned long qop = 0;             # <<<<<<<<<<<<<<
 * 
 *     res = DecryptMessage(&context.handle, message.__c_value__(), seq_no, &qop)
 */
  __pyx_v_qop = 0;

  /* "spnego/_sspi_raw/sspi.pyx":674
 *     cdef unsigned long qop = 0;
 * 
 *     res = DecryptMessage(&context.handle, message.__c_value__(), seq_no, &qop)             # <<<<<<<<<<<<<<
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)
 */
  __pyx_v_res = DecryptMessage((&__pyx_v_context->handle), ((struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_message->__pyx_vtab)->__pyx___c_value__(__pyx_v_message), __pyx_v_seq_no, (&__pyx_v_qop));

  /* "spnego/_sspi_raw/sspi.pyx":675
 * 
 *     res = DecryptMessage(&context.handle, message.__c_value__(), seq_no, &qop)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  __pyx_t_1 = ((__pyx_v_res != SEC_E_OK) != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":676
 *     res = DecryptMessage(&context.handle, message.__c_value__(), seq_no, &qop)
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)             # <<<<<<<<<<<<<<
 * 
 *     return qop
 */
    __pyx_t_2 = PyErr_SetFromWindowsErr(__pyx_v_res); if (unlikely(__pyx_t_2 == ((PyObject *)NULL))) __PYX_ERR(1, 676, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":675
 * 
 *     res = DecryptMessage(&context.handle, message.__c_value__(), seq_no, &qop)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":678
 *         PyErr_SetFromWindowsErr(res)
 * 
 *     return qop             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_v_qop); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":671
 * 
 * 
 * def decrypt_message(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):             # <<<<<<<<<<<<<<
 *     cdef unsigned long qop = 0;
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.decrypt_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":681
 * 
 * 
 * def encrypt_message(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0,             # <<<<<<<<<<<<<<
 *     unsigned long qop=0):
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_7encrypt_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_7encrypt_message = {"encrypt_message", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6spnego_9_sspi_raw_4sspi_7encrypt_message, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_7encrypt_message(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context = 0;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message = 0;
  unsigned long __pyx_v_seq_no;
  unsigned long __pyx_v_qop;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("encrypt_message (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,&__pyx_n_s_message,&__pyx_n_s_seq_no,&__pyx_n_s_qop,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("encrypt_message", 0, 2, 4, 1); __PYX_ERR(1, 681, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seq_no);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_qop);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "encrypt_message") < 0)) __PYX_ERR(1, 681, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_context = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)values[0]);
    __pyx_v_message = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)values[1]);
    if (values[2]) {
      __pyx_v_seq_no = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_seq_no == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 681, __pyx_L3_error)
    } else {
      __pyx_v_seq_no = ((unsigned long)0);
    }
    if (values[3]) {
      __pyx_v_qop = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_qop == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 682, __pyx_L3_error)
    } else {
      __pyx_v_qop = ((unsigned long)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("encrypt_message", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 681, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.encrypt_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext, 0, "context", 0))) __PYX_ERR(1, 681, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBufferDesc, 0, "message", 0))) __PYX_ERR(1, 681, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_6encrypt_message(__pyx_self, __pyx_v_context, __pyx_v_message, __pyx_v_seq_no, __pyx_v_qop);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_6encrypt_message(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message, unsigned long __pyx_v_seq_no, unsigned long __pyx_v_qop) {
  SECURITY_STATUS __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encrypt_message", 0);

  /* "spnego/_sspi_raw/sspi.pyx":684
 *     unsigned long qop=0):
 * 
 *     res = EncryptMessage(&context.handle, qop, message.__c_value__(), seq_no)             # <<<<<<<<<<<<<<
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)
 */
  __pyx_v_res = EncryptMessage((&__pyx_v_context->handle), __pyx_v_qop, ((struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_message->__pyx_vtab)->__pyx___c_value__(__pyx_v_message), __pyx_v_seq_no);

  /* "spnego/_sspi_raw/sspi.pyx":685
 * 
 *     res = EncryptMessage(&context.handle, qop, message.__c_value__(), seq_no)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  __pyx_t_1 = ((__pyx_v_res != SEC_E_OK) != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":686
 *     res = EncryptMessage(&context.handle, qop, message.__c_value__(), seq_no)
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = PyErr_SetFromWindowsErr(__pyx_v_res); if (unlikely(__pyx_t_2 == ((PyObject *)NULL))) __PYX_ERR(1, 686, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":685
 * 
 *     res = EncryptMessage(&context.handle, qop, message.__c_value__(), seq_no)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":681
 * 
 * 
 * def encrypt_message(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0,             # <<<<<<<<<<<<<<
 *     unsigned long qop=0):
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.encrypt_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":689
 * 
 * 
 * def initialize_security_context(Credential credential not None, SecurityContext context not None,             # <<<<<<<<<<<<<<
 *     unicode target_name not None, unsigned long context_req=0, SecBufferDesc input_buffer=None,
 *     unsigned long target_data_rep=SECURITY_NATIVE_DREP, SecBufferDesc output_buffer=None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9initialize_security_context(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_9initialize_security_context = {"initialize_security_context", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6spnego_9_sspi_raw_4sspi_9initialize_security_context, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_9initialize_security_context(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_credential = 0;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context = 0;
  PyObject *__pyx_v_target_name = 0;
  unsigned long __pyx_v_context_req;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_input_buffer = 0;
  unsigned long __pyx_v_target_data_rep;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_output_buffer = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("initialize_security_context (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_credential,&__pyx_n_s_context,&__pyx_n_s_target_name,&__pyx_n_s_context_req,&__pyx_n_s_input_buffer,&__pyx_n_s_target_data_rep,&__pyx_n_s_output_buffer,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};

    /* "spnego/_sspi_raw/sspi.pyx":690
 * 
 * def initialize_security_context(Credential credential not None, SecurityContext context not None,
 *     unicode target_name not None, unsigned long context_req=0, SecBufferDesc input_buffer=None,             # <<<<<<<<<<<<<<
 *     unsigned long target_data_rep=SECURITY_NATIVE_DREP, SecBufferDesc output_buffer=None):
 * 
 */
    values[4] = (PyObject *)((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)Py_None);

    /* "spnego/_sspi_raw/sspi.pyx":691
 * def initialize_security_context(Credential credential not None, SecurityContext context not None,
 *     unicode target_name not None, unsigned long context_req=0, SecBufferDesc input_buffer=None,
 *     unsigned long target_data_rep=SECURITY_NATIVE_DREP, SecBufferDesc output_buffer=None):             # <<<<<<<<<<<<<<
 * 
 *     cdef PCtxtHandle input_context = &context.handle if context.init else NULL
 */
    values[6] = (PyObject *)((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_credential)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("initialize_security_context", 0, 3, 7, 1); __PYX_ERR(1, 689, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("initialize_security_context", 0, 3, 7, 2); __PYX_ERR(1, 689, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context_req);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_input_buffer);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target_data_rep);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output_buffer);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "initialize_security_context") < 0)) __PYX_ERR(1, 689, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_credential = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *)values[0]);
    __pyx_v_context = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)values[1]);
    __pyx_v_target_name = ((PyObject*)values[2]);
    if (values[3]) {
      __pyx_v_context_req = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_context_req == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 690, __pyx_L3_error)
    } else {
      __pyx_v_context_req = ((unsigned long)0);
    }
    __pyx_v_input_buffer = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)values[4]);
    if (values[5]) {
      __pyx_v_target_data_rep = __Pyx_PyInt_As_unsigned_long(values[5]); if (unlikely((__pyx_v_target_data_rep == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 691, __pyx_L3_error)
    } else {
      __pyx_v_target_data_rep = __pyx_k__18;
    }
    __pyx_v_output_buffer = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)values[6]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("initialize_security_context", 0, 3, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 689, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.initialize_security_context", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_credential), __pyx_ptype_6spnego_9_sspi_raw_4sspi_Credential, 0, "credential", 0))) __PYX_ERR(1, 689, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext, 0, "context", 0))) __PYX_ERR(1, 689, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_target_name), (&PyUnicode_Type), 0, "target_name", 1))) __PYX_ERR(1, 690, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_input_buffer), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBufferDesc, 1, "input_buffer", 0))) __PYX_ERR(1, 690, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_output_buffer), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBufferDesc, 1, "output_buffer", 0))) __PYX_ERR(1, 691, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_8initialize_security_context(__pyx_self, __pyx_v_credential, __pyx_v_context, __pyx_v_target_name, __pyx_v_context_req, __pyx_v_input_buffer, __pyx_v_target_data_rep, __pyx_v_output_buffer);

  /* "spnego/_sspi_raw/sspi.pyx":689
 * 
 * 
 * def initialize_security_context(Credential credential not None, SecurityContext context not None,             # <<<<<<<<<<<<<<
 *     unicode target_name not None, unsigned long context_req=0, SecBufferDesc input_buffer=None,
 *     unsigned long target_data_rep=SECURITY_NATIVE_DREP, SecBufferDesc output_buffer=None):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_8initialize_security_context(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential *__pyx_v_credential, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, PyObject *__pyx_v_target_name, unsigned long __pyx_v_context_req, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_input_buffer, unsigned long __pyx_v_target_data_rep, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_output_buffer) {
  PCtxtHandle __pyx_v_input_context;
  struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *__pyx_v_w_target_name = 0;
  PSecBufferDesc __pyx_v_input;
  PSecBufferDesc __pyx_v_output;
  SECURITY_INTEGER __pyx_v_expiry;
  PyObject *__pyx_v_buffer = NULL;
  SECURITY_STATUS __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  CtxtHandle *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PSecBufferDesc __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  SECURITY_STATUS __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("initialize_security_context", 0);

  /* "spnego/_sspi_raw/sspi.pyx":693
 *     unsigned long target_data_rep=SECURITY_NATIVE_DREP, SecBufferDesc output_buffer=None):
 * 
 *     cdef PCtxtHandle input_context = &context.handle if context.init else NULL             # <<<<<<<<<<<<<<
 *     cdef WideChar w_target_name = WideChar.from_text(target_name)
 *     cdef PSecBufferDesc input = input_buffer.__c_value__() if input_buffer else NULL
 */
  if ((__pyx_v_context->init != 0)) {
    __pyx_t_1 = (&__pyx_v_context->handle);
  } else {
    __pyx_t_1 = NULL;
  }
  __pyx_v_input_context = __pyx_t_1;

  /* "spnego/_sspi_raw/sspi.pyx":694
 * 
 *     cdef PCtxtHandle input_context = &context.handle if context.init else NULL
 *     cdef WideChar w_target_name = WideChar.from_text(target_name)             # <<<<<<<<<<<<<<
 *     cdef PSecBufferDesc input = input_buffer.__c_value__() if input_buffer else NULL
 *     cdef PSecBufferDesc output = output_buffer.__c_value__() if output_buffer else NULL
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6spnego_9_sspi_raw_4text_WideChar), __pyx_n_s_from_text); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_target_name) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_target_name);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6spnego_9_sspi_raw_4text_WideChar))))) __PYX_ERR(1, 694, __pyx_L1_error)
  __pyx_v_w_target_name = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":695
 *     cdef PCtxtHandle input_context = &context.handle if context.init else NULL
 *     cdef WideChar w_target_name = WideChar.from_text(target_name)
 *     cdef PSecBufferDesc input = input_buffer.__c_value__() if input_buffer else NULL             # <<<<<<<<<<<<<<
 *     cdef PSecBufferDesc output = output_buffer.__c_value__() if output_buffer else NULL
 *     cdef SECURITY_INTEGER expiry
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_input_buffer)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 695, __pyx_L1_error)
  if (__pyx_t_6) {
    __pyx_t_5 = ((struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_input_buffer->__pyx_vtab)->__pyx___c_value__(__pyx_v_input_buffer);
  } else {
    __pyx_t_5 = NULL;
  }
  __pyx_v_input = __pyx_t_5;

  /* "spnego/_sspi_raw/sspi.pyx":696
 *     cdef WideChar w_target_name = WideChar.from_text(target_name)
 *     cdef PSecBufferDesc input = input_buffer.__c_value__() if input_buffer else NULL
 *     cdef PSecBufferDesc output = output_buffer.__c_value__() if output_buffer else NULL             # <<<<<<<<<<<<<<
 *     cdef SECURITY_INTEGER expiry
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_output_buffer)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 696, __pyx_L1_error)
  if (__pyx_t_6) {
    __pyx_t_5 = ((struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_output_buffer->__pyx_vtab)->__pyx___c_value__(__pyx_v_output_buffer);
  } else {
    __pyx_t_5 = NULL;
  }
  __pyx_v_output = __pyx_t_5;

  /* "spnego/_sspi_raw/sspi.pyx":699
 *     cdef SECURITY_INTEGER expiry
 * 
 *     for buffer in (output_buffer or []):             # <<<<<<<<<<<<<<
 *         if len((<SecBuffer>buffer)) == 0:
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_output_buffer)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 699, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __Pyx_INCREF(((PyObject *)__pyx_v_output_buffer));
    __pyx_t_2 = ((PyObject *)__pyx_v_output_buffer);
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L5_bool_binop_done:;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 699, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 699, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 699, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 699, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 699, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 699, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_buffer, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "spnego/_sspi_raw/sspi.pyx":700
 * 
 *     for buffer in (output_buffer or []):
 *         if len((<SecBuffer>buffer)) == 0:             # <<<<<<<<<<<<<<
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY
 *             (<SecBuffer>buffer).sys_alloc = 1
 */
    __pyx_t_9 = PyObject_Length(__pyx_v_buffer); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(1, 700, __pyx_L1_error)
    __pyx_t_6 = ((__pyx_t_9 == 0) != 0);
    if (__pyx_t_6) {

      /* "spnego/_sspi_raw/sspi.pyx":701
 *     for buffer in (output_buffer or []):
 *         if len((<SecBuffer>buffer)) == 0:
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY             # <<<<<<<<<<<<<<
 *             (<SecBuffer>buffer).sys_alloc = 1
 * 
 */
      __pyx_v_context_req = (__pyx_v_context_req | ISC_REQ_ALLOCATE_MEMORY);

      /* "spnego/_sspi_raw/sspi.pyx":702
 *         if len((<SecBuffer>buffer)) == 0:
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY
 *             (<SecBuffer>buffer).sys_alloc = 1             # <<<<<<<<<<<<<<
 * 
 *     res = InitializeSecurityContextW(&credential.handle, input_context, w_target_name.buffer, context_req, 0,
 */
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      __Pyx_GOTREF(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_buffer)->sys_alloc);
      __Pyx_DECREF(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_buffer)->sys_alloc);
      ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)__pyx_v_buffer)->sys_alloc = __pyx_int_1;

      /* "spnego/_sspi_raw/sspi.pyx":700
 * 
 *     for buffer in (output_buffer or []):
 *         if len((<SecBuffer>buffer)) == 0:             # <<<<<<<<<<<<<<
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY
 *             (<SecBuffer>buffer).sys_alloc = 1
 */
    }

    /* "spnego/_sspi_raw/sspi.pyx":699
 *     cdef SECURITY_INTEGER expiry
 * 
 *     for buffer in (output_buffer or []):             # <<<<<<<<<<<<<<
 *         if len((<SecBuffer>buffer)) == 0:
 *             context_req |= ISC_REQ_ALLOCATE_MEMORY
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":704
 *             (<SecBuffer>buffer).sys_alloc = 1
 * 
 *     res = InitializeSecurityContextW(&credential.handle, input_context, w_target_name.buffer, context_req, 0,             # <<<<<<<<<<<<<<
 *         target_data_rep, input, 0, &context.handle, output, &context.context_attr, &expiry)
 * 
 */
  __pyx_v_res = InitializeSecurityContextW((&__pyx_v_credential->handle), __pyx_v_input_context, __pyx_v_w_target_name->buffer, __pyx_v_context_req, 0, __pyx_v_target_data_rep, __pyx_v_input, 0, (&__pyx_v_context->handle), __pyx_v_output, (&__pyx_v_context->context_attr), (&__pyx_v_expiry));

  /* "spnego/_sspi_raw/sspi.pyx":707
 *         target_data_rep, input, 0, &context.handle, output, &context.context_attr, &expiry)
 * 
 *     if res in [_SEC_I_COMPLETE_AND_CONTINUE, _SEC_I_COMPLETE_NEEDED]:             # <<<<<<<<<<<<<<
 *         res = CompleteAuthToken(&context.handle, output)
 * 
 */
  __pyx_t_10 = __pyx_v_res;
  __pyx_t_11 = ((__pyx_t_10 == SEC_I_COMPLETE_AND_CONTINUE) != 0);
  if (!__pyx_t_11) {
  } else {
    __pyx_t_6 = __pyx_t_11;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_11 = ((__pyx_t_10 == SEC_I_COMPLETE_NEEDED) != 0);
  __pyx_t_6 = __pyx_t_11;
  __pyx_L9_bool_binop_done:;
  __pyx_t_11 = (__pyx_t_6 != 0);
  if (__pyx_t_11) {

    /* "spnego/_sspi_raw/sspi.pyx":708
 * 
 *     if res in [_SEC_I_COMPLETE_AND_CONTINUE, _SEC_I_COMPLETE_NEEDED]:
 *         res = CompleteAuthToken(&context.handle, output)             # <<<<<<<<<<<<<<
 * 
 *     if res not in [_SEC_I_CONTINUE_NEEDED, _SEC_E_OK]:
 */
    __pyx_v_res = CompleteAuthToken((&__pyx_v_context->handle), __pyx_v_output);

    /* "spnego/_sspi_raw/sspi.pyx":707
 *         target_data_rep, input, 0, &context.handle, output, &context.context_attr, &expiry)
 * 
 *     if res in [_SEC_I_COMPLETE_AND_CONTINUE, _SEC_I_COMPLETE_NEEDED]:             # <<<<<<<<<<<<<<
 *         res = CompleteAuthToken(&context.handle, output)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":710
 *         res = CompleteAuthToken(&context.handle, output)
 * 
 *     if res not in [_SEC_I_CONTINUE_NEEDED, _SEC_E_OK]:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  __pyx_t_10 = __pyx_v_res;
  __pyx_t_6 = ((__pyx_t_10 != SEC_I_CONTINUE_NEEDED) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_11 = __pyx_t_6;
    goto __pyx_L12_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_t_10 != SEC_E_OK) != 0);
  __pyx_t_11 = __pyx_t_6;
  __pyx_L12_bool_binop_done:;
  __pyx_t_6 = (__pyx_t_11 != 0);
  if (__pyx_t_6) {

    /* "spnego/_sspi_raw/sspi.pyx":711
 * 
 *     if res not in [_SEC_I_CONTINUE_NEEDED, _SEC_E_OK]:
 *         PyErr_SetFromWindowsErr(res)             # <<<<<<<<<<<<<<
 * 
 *     context.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart
 */
    __pyx_t_12 = PyErr_SetFromWindowsErr(__pyx_v_res); if (unlikely(__pyx_t_12 == ((PyObject *)NULL))) __PYX_ERR(1, 711, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":710
 *         res = CompleteAuthToken(&context.handle, output)
 * 
 *     if res not in [_SEC_I_CONTINUE_NEEDED, _SEC_E_OK]:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":713
 *         PyErr_SetFromWindowsErr(res)
 * 
 *     context.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart             # <<<<<<<<<<<<<<
 *     context.init = 1
 * 
 */
  __pyx_v_context->expiry = ((((unsigned PY_LONG_LONG)__pyx_v_expiry.HighPart) << 32) | __pyx_v_expiry.LowPart);

  /* "spnego/_sspi_raw/sspi.pyx":714
 * 
 *     context.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart
 *     context.init = 1             # <<<<<<<<<<<<<<
 * 
 *     return res
 */
  __pyx_v_context->init = 1;

  /* "spnego/_sspi_raw/sspi.pyx":716
 *     context.init = 1
 * 
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_LONG(__pyx_v_res); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":689
 * 
 * 
 * def initialize_security_context(Credential credential not None, SecurityContext context not None,             # <<<<<<<<<<<<<<
 *     unicode target_name not None, unsigned long context_req=0, SecBufferDesc input_buffer=None,
 *     unsigned long target_data_rep=SECURITY_NATIVE_DREP, SecBufferDesc output_buffer=None):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.initialize_security_context", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_w_target_name);
  __Pyx_XDECREF(__pyx_v_buffer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":719
 * 
 * 
 * def make_signature(SecurityContext context not None, unsigned long qop, SecBufferDesc message not None,             # <<<<<<<<<<<<<<
 *     unsigned long seq_no=0):
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_11make_signature(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_11make_signature = {"make_signature", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6spnego_9_sspi_raw_4sspi_11make_signature, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_11make_signature(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context = 0;
  unsigned long __pyx_v_qop;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message = 0;
  unsigned long __pyx_v_seq_no;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_signature (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,&__pyx_n_s_qop,&__pyx_n_s_message,&__pyx_n_s_seq_no,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_qop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("make_signature", 0, 3, 4, 1); __PYX_ERR(1, 719, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("make_signature", 0, 3, 4, 2); __PYX_ERR(1, 719, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seq_no);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "make_signature") < 0)) __PYX_ERR(1, 719, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_context = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)values[0]);
    __pyx_v_qop = __Pyx_PyInt_As_unsigned_long(values[1]); if (unlikely((__pyx_v_qop == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 719, __pyx_L3_error)
    __pyx_v_message = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)values[2]);
    if (values[3]) {
      __pyx_v_seq_no = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_seq_no == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 720, __pyx_L3_error)
    } else {
      __pyx_v_seq_no = ((unsigned long)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("make_signature", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 719, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.make_signature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext, 0, "context", 0))) __PYX_ERR(1, 719, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBufferDesc, 0, "message", 0))) __PYX_ERR(1, 719, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_10make_signature(__pyx_self, __pyx_v_context, __pyx_v_qop, __pyx_v_message, __pyx_v_seq_no);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_10make_signature(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, unsigned long __pyx_v_qop, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message, unsigned long __pyx_v_seq_no) {
  SECURITY_STATUS __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("make_signature", 0);

  /* "spnego/_sspi_raw/sspi.pyx":722
 *     unsigned long seq_no=0):
 * 
 *     res = MakeSignature(&context.handle, qop, message.__c_value__(), seq_no)             # <<<<<<<<<<<<<<
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)
 */
  __pyx_v_res = MakeSignature((&__pyx_v_context->handle), __pyx_v_qop, ((struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_message->__pyx_vtab)->__pyx___c_value__(__pyx_v_message), __pyx_v_seq_no);

  /* "spnego/_sspi_raw/sspi.pyx":723
 * 
 *     res = MakeSignature(&context.handle, qop, message.__c_value__(), seq_no)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  __pyx_t_1 = ((__pyx_v_res != SEC_E_OK) != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":724
 *     res = MakeSignature(&context.handle, qop, message.__c_value__(), seq_no)
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = PyErr_SetFromWindowsErr(__pyx_v_res); if (unlikely(__pyx_t_2 == ((PyObject *)NULL))) __PYX_ERR(1, 724, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":723
 * 
 *     res = MakeSignature(&context.handle, qop, message.__c_value__(), seq_no)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":719
 * 
 * 
 * def make_signature(SecurityContext context not None, unsigned long qop, SecBufferDesc message not None,             # <<<<<<<<<<<<<<
 *     unsigned long seq_no=0):
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.make_signature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":727
 * 
 * 
 * def query_context_attributes(SecurityContext context not None, unsigned long attribute):             # <<<<<<<<<<<<<<
 *     if attribute not in [SecPkgAttr.names, SecPkgAttr.package_info, SecPkgAttr.session_key, SecPkgAttr.sizes]:
 *         raise NotImplementedError("Only names, package_info, session_key, or sizes is implemented")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13query_context_attributes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_13query_context_attributes = {"query_context_attributes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6spnego_9_sspi_raw_4sspi_13query_context_attributes, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_13query_context_attributes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context = 0;
  unsigned long __pyx_v_attribute;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("query_context_attributes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,&__pyx_n_s_attribute,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_attribute)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("query_context_attributes", 1, 2, 2, 1); __PYX_ERR(1, 727, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "query_context_attributes") < 0)) __PYX_ERR(1, 727, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_context = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)values[0]);
    __pyx_v_attribute = __Pyx_PyInt_As_unsigned_long(values[1]); if (unlikely((__pyx_v_attribute == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 727, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("query_context_attributes", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 727, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.query_context_attributes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext, 0, "context", 0))) __PYX_ERR(1, 727, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_12query_context_attributes(__pyx_self, __pyx_v_context, __pyx_v_attribute);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_12query_context_attributes(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, unsigned long __pyx_v_attribute) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned long __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query_context_attributes", 0);

  /* "spnego/_sspi_raw/sspi.pyx":728
 * 
 * def query_context_attributes(SecurityContext context not None, unsigned long attribute):
 *     if attribute not in [SecPkgAttr.names, SecPkgAttr.package_info, SecPkgAttr.session_key, SecPkgAttr.sizes]:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("Only names, package_info, session_key, or sizes is implemented")
 * 
 */
  __pyx_t_1 = __pyx_v_attribute;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SecPkgAttr); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_names); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_long(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_SecPkgAttr); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_package_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SecPkgAttr); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_session_key); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SecPkgAttr); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sizes); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  __pyx_t_6 = (__pyx_t_2 != 0);
  if (unlikely(__pyx_t_6)) {

    /* "spnego/_sspi_raw/sspi.pyx":729
 * def query_context_attributes(SecurityContext context not None, unsigned long attribute):
 *     if attribute not in [SecPkgAttr.names, SecPkgAttr.package_info, SecPkgAttr.session_key, SecPkgAttr.sizes]:
 *         raise NotImplementedError("Only names, package_info, session_key, or sizes is implemented")             # <<<<<<<<<<<<<<
 * 
 *     if attribute == SecPkgAttr.names:
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 729, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 729, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":728
 * 
 * def query_context_attributes(SecurityContext context not None, unsigned long attribute):
 *     if attribute not in [SecPkgAttr.names, SecPkgAttr.package_info, SecPkgAttr.session_key, SecPkgAttr.sizes]:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("Only names, package_info, session_key, or sizes is implemented")
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":731
 *         raise NotImplementedError("Only names, package_info, session_key, or sizes is implemented")
 * 
 *     if attribute == SecPkgAttr.names:             # <<<<<<<<<<<<<<
 *         return _query_context_names(context)
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_long(__pyx_v_attribute); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_SecPkgAttr); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_names); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 731, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 731, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_6) {

    /* "spnego/_sspi_raw/sspi.pyx":732
 * 
 *     if attribute == SecPkgAttr.names:
 *         return _query_context_names(context)             # <<<<<<<<<<<<<<
 * 
 *     elif attribute == SecPkgAttr.package_info:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_query_context_names); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 732, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_context)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_context));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 732, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "spnego/_sspi_raw/sspi.pyx":731
 *         raise NotImplementedError("Only names, package_info, session_key, or sizes is implemented")
 * 
 *     if attribute == SecPkgAttr.names:             # <<<<<<<<<<<<<<
 *         return _query_context_names(context)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":734
 *         return _query_context_names(context)
 * 
 *     elif attribute == SecPkgAttr.package_info:             # <<<<<<<<<<<<<<
 *         return _query_context_package_info(context)
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_attribute); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SecPkgAttr); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_package_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_6) {

    /* "spnego/_sspi_raw/sspi.pyx":735
 * 
 *     elif attribute == SecPkgAttr.package_info:
 *         return _query_context_package_info(context)             # <<<<<<<<<<<<<<
 * 
 *     elif attribute == SecPkgAttr.session_key:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_query_context_package_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, ((PyObject *)__pyx_v_context)) : __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_v_context));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "spnego/_sspi_raw/sspi.pyx":734
 *         return _query_context_names(context)
 * 
 *     elif attribute == SecPkgAttr.package_info:             # <<<<<<<<<<<<<<
 *         return _query_context_package_info(context)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":737
 *         return _query_context_package_info(context)
 * 
 *     elif attribute == SecPkgAttr.session_key:             # <<<<<<<<<<<<<<
 *         return _query_context_session_key(context)
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_v_attribute); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SecPkgAttr); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_session_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 737, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 737, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {

    /* "spnego/_sspi_raw/sspi.pyx":738
 * 
 *     elif attribute == SecPkgAttr.session_key:
 *         return _query_context_session_key(context)             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_query_context_session_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, ((PyObject *)__pyx_v_context)) : __Pyx_PyObject_CallOneArg(__pyx_t_5, ((PyObject *)__pyx_v_context));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "spnego/_sspi_raw/sspi.pyx":737
 *         return _query_context_package_info(context)
 * 
 *     elif attribute == SecPkgAttr.session_key:             # <<<<<<<<<<<<<<
 *         return _query_context_session_key(context)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":741
 * 
 *     else:
 *         return _query_context_sizes(context)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_query_context_sizes); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, ((PyObject *)__pyx_v_context)) : __Pyx_PyObject_CallOneArg(__pyx_t_5, ((PyObject *)__pyx_v_context));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "spnego/_sspi_raw/sspi.pyx":727
 * 
 * 
 * def query_context_attributes(SecurityContext context not None, unsigned long attribute):             # <<<<<<<<<<<<<<
 *     if attribute not in [SecPkgAttr.names, SecPkgAttr.package_info, SecPkgAttr.session_key, SecPkgAttr.sizes]:
 *         raise NotImplementedError("Only names, package_info, session_key, or sizes is implemented")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.query_context_attributes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":744
 * 
 * 
 * def _query_context_names(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_Names info
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_15_query_context_names(PyObject *__pyx_self, PyObject *__pyx_v_context); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_15_query_context_names = {"_query_context_names", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_15_query_context_names, METH_O, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_15_query_context_names(PyObject *__pyx_self, PyObject *__pyx_v_context) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_query_context_names (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext, 0, "context", 0))) __PYX_ERR(1, 744, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_14_query_context_names(__pyx_self, ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)__pyx_v_context));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_14_query_context_names(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context) {
  SecPkgContext_Names __pyx_v_info;
  SECURITY_STATUS __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  unsigned long __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_query_context_names", 0);

  /* "spnego/_sspi_raw/sspi.pyx":747
 *     cdef SecPkgContext_Names info
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.names, &info)             # <<<<<<<<<<<<<<
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SecPkgAttr); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_names); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_long(__pyx_t_2); if (unlikely((__pyx_t_3 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 747, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_res = QueryContextAttributesW((&__pyx_v_context->handle), __pyx_t_3, (&__pyx_v_info));

  /* "spnego/_sspi_raw/sspi.pyx":748
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.names, &info)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  __pyx_t_4 = ((__pyx_v_res != SEC_E_OK) != 0);
  if (__pyx_t_4) {

    /* "spnego/_sspi_raw/sspi.pyx":749
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.names, &info)
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_5 = PyErr_SetFromWindowsErr(__pyx_v_res); if (unlikely(__pyx_t_5 == ((PyObject *)NULL))) __PYX_ERR(1, 749, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":748
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.names, &info)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":751
 *         PyErr_SetFromWindowsErr(res)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         return u16_to_text(info.sUserName, -1)
 *     finally:
 */
  /*try:*/ {

    /* "spnego/_sspi_raw/sspi.pyx":752
 * 
 *     try:
 *         return u16_to_text(info.sUserName, -1)             # <<<<<<<<<<<<<<
 *     finally:
 *         FreeContextBuffer(<void*>info.sUserName)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_6spnego_9_sspi_raw_4text_u16_to_text(__pyx_v_info.sUserName, -1L); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 752, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L4_return;
  }

  /* "spnego/_sspi_raw/sspi.pyx":754
 *         return u16_to_text(info.sUserName, -1)
 *     finally:
 *         FreeContextBuffer(<void*>info.sUserName)             # <<<<<<<<<<<<<<
 * 
 * def _query_context_package_info(SecurityContext context not None):
 */
  /*finally:*/ {
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_6 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        (void)(FreeContextBuffer(((void *)__pyx_v_info.sUserName)));
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
    }
    __pyx_L4_return: {
      __pyx_t_14 = __pyx_r;
      __pyx_r = 0;
      (void)(FreeContextBuffer(((void *)__pyx_v_info.sUserName)));
      __pyx_r = __pyx_t_14;
      __pyx_t_14 = 0;
      goto __pyx_L0;
    }
  }

  /* "spnego/_sspi_raw/sspi.pyx":744
 * 
 * 
 * def _query_context_names(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_Names info
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi._query_context_names", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":756
 *         FreeContextBuffer(<void*>info.sUserName)
 * 
 * def _query_context_package_info(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_PackageInfoW raw_info
 *     cdef PSecPkgInfoW info
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17_query_context_package_info(PyObject *__pyx_self, PyObject *__pyx_v_context); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_17_query_context_package_info = {"_query_context_package_info", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_17_query_context_package_info, METH_O, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_17_query_context_package_info(PyObject *__pyx_self, PyObject *__pyx_v_context) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_query_context_package_info (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext, 0, "context", 0))) __PYX_ERR(1, 756, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_16_query_context_package_info(__pyx_self, ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)__pyx_v_context));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_16_query_context_package_info(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context) {
  SecPkgContext_PackageInfoW __pyx_v_raw_info;
  PSecPkgInfoW __pyx_v_info;
  SECURITY_STATUS __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  unsigned long __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_query_context_package_info", 0);

  /* "spnego/_sspi_raw/sspi.pyx":760
 *     cdef PSecPkgInfoW info
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.package_info, &raw_info)             # <<<<<<<<<<<<<<
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SecPkgAttr); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_package_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_long(__pyx_t_2); if (unlikely((__pyx_t_3 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 760, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_res = QueryContextAttributesW((&__pyx_v_context->handle), __pyx_t_3, (&__pyx_v_raw_info));

  /* "spnego/_sspi_raw/sspi.pyx":761
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.package_info, &raw_info)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  __pyx_t_4 = ((__pyx_v_res != SEC_E_OK) != 0);
  if (__pyx_t_4) {

    /* "spnego/_sspi_raw/sspi.pyx":762
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.package_info, &raw_info)
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_5 = PyErr_SetFromWindowsErr(__pyx_v_res); if (unlikely(__pyx_t_5 == ((PyObject *)NULL))) __PYX_ERR(1, 762, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":761
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.package_info, &raw_info)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":764
 *         PyErr_SetFromWindowsErr(res)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         info = <PSecPkgInfoW>raw_info.PackageInfo
 *         return SecPkgInfo(info.fCapabilities, info.wVersion, info.wRPCID, info.cbMaxToken, u16_to_text(info.Name, -1),
 */
  /*try:*/ {

    /* "spnego/_sspi_raw/sspi.pyx":765
 * 
 *     try:
 *         info = <PSecPkgInfoW>raw_info.PackageInfo             # <<<<<<<<<<<<<<
 *         return SecPkgInfo(info.fCapabilities, info.wVersion, info.wRPCID, info.cbMaxToken, u16_to_text(info.Name, -1),
 *             u16_to_text(info.Comment, -1))
 */
    __pyx_v_info = ((PSecPkgInfoW)__pyx_v_raw_info.PackageInfo);

    /* "spnego/_sspi_raw/sspi.pyx":766
 *     try:
 *         info = <PSecPkgInfoW>raw_info.PackageInfo
 *         return SecPkgInfo(info.fCapabilities, info.wVersion, info.wRPCID, info.cbMaxToken, u16_to_text(info.Name, -1),             # <<<<<<<<<<<<<<
 *             u16_to_text(info.Comment, -1))
 *     finally:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SecPkgInfo); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 766, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyInt_From_unsigned_long(__pyx_v_info->fCapabilities); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 766, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_unsigned_short(__pyx_v_info->wVersion); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 766, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyInt_From_unsigned_short(__pyx_v_info->wRPCID); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 766, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyInt_From_unsigned_long(__pyx_v_info->cbMaxToken); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 766, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __pyx_f_6spnego_9_sspi_raw_4text_u16_to_text(__pyx_v_info->Name, -1L); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 766, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_10);

    /* "spnego/_sspi_raw/sspi.pyx":767
 *         info = <PSecPkgInfoW>raw_info.PackageInfo
 *         return SecPkgInfo(info.fCapabilities, info.wVersion, info.wRPCID, info.cbMaxToken, u16_to_text(info.Name, -1),
 *             u16_to_text(info.Comment, -1))             # <<<<<<<<<<<<<<
 *     finally:
 *         FreeContextBuffer(<void*>raw_info.PackageInfo)
 */
    __pyx_t_11 = __pyx_f_6spnego_9_sspi_raw_4text_u16_to_text(__pyx_v_info->Comment, -1L); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 767, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    __pyx_t_13 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_13 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[7] = {__pyx_t_12, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_13, 6+__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 766, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[7] = {__pyx_t_12, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_13, 6+__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 766, __pyx_L5_error)
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    } else
    #endif
    {
      __pyx_t_14 = PyTuple_New(6+__pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 766, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_14);
      if (__pyx_t_12) {
        __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_12); __pyx_t_12 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_13, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_13, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_13, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_14, 3+__pyx_t_13, __pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_14, 4+__pyx_t_13, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_14, 5+__pyx_t_13, __pyx_t_11);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_8 = 0;
      __pyx_t_9 = 0;
      __pyx_t_10 = 0;
      __pyx_t_11 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 766, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L4_return;
  }

  /* "spnego/_sspi_raw/sspi.pyx":769
 *             u16_to_text(info.Comment, -1))
 *     finally:
 *         FreeContextBuffer(<void*>raw_info.PackageInfo)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*finally:*/ {
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_22);
      __pyx_t_13 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
      {
        (void)(FreeContextBuffer(((void *)__pyx_v_raw_info.PackageInfo)));
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_22);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
      }
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
      goto __pyx_L1_error;
    }
    __pyx_L4_return: {
      __pyx_t_22 = __pyx_r;
      __pyx_r = 0;
      (void)(FreeContextBuffer(((void *)__pyx_v_raw_info.PackageInfo)));
      __pyx_r = __pyx_t_22;
      __pyx_t_22 = 0;
      goto __pyx_L0;
    }
  }

  /* "spnego/_sspi_raw/sspi.pyx":756
 *         FreeContextBuffer(<void*>info.sUserName)
 * 
 * def _query_context_package_info(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_PackageInfoW raw_info
 *     cdef PSecPkgInfoW info
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi._query_context_package_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":772
 * 
 * 
 * def _query_context_session_key(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_SessionKey info
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_19_query_context_session_key(PyObject *__pyx_self, PyObject *__pyx_v_context); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_19_query_context_session_key = {"_query_context_session_key", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_19_query_context_session_key, METH_O, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_19_query_context_session_key(PyObject *__pyx_self, PyObject *__pyx_v_context) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_query_context_session_key (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext, 0, "context", 0))) __PYX_ERR(1, 772, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_18_query_context_session_key(__pyx_self, ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)__pyx_v_context));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_18_query_context_session_key(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context) {
  SecPkgContext_SessionKey __pyx_v_info;
  SECURITY_STATUS __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  unsigned long __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_query_context_session_key", 0);

  /* "spnego/_sspi_raw/sspi.pyx":775
 *     cdef SecPkgContext_SessionKey info
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.session_key, &info)             # <<<<<<<<<<<<<<
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SecPkgAttr); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_session_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_long(__pyx_t_2); if (unlikely((__pyx_t_3 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 775, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_res = QueryContextAttributesW((&__pyx_v_context->handle), __pyx_t_3, (&__pyx_v_info));

  /* "spnego/_sspi_raw/sspi.pyx":776
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.session_key, &info)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  __pyx_t_4 = ((__pyx_v_res != SEC_E_OK) != 0);
  if (__pyx_t_4) {

    /* "spnego/_sspi_raw/sspi.pyx":777
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.session_key, &info)
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_5 = PyErr_SetFromWindowsErr(__pyx_v_res); if (unlikely(__pyx_t_5 == ((PyObject *)NULL))) __PYX_ERR(1, 777, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":776
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.session_key, &info)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":779
 *         PyErr_SetFromWindowsErr(res)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         return (<char *>info.SessionKey)[:info.SessionKeyLength]
 *     finally:
 */
  /*try:*/ {

    /* "spnego/_sspi_raw/sspi.pyx":780
 * 
 *     try:
 *         return (<char *>info.SessionKey)[:info.SessionKeyLength]             # <<<<<<<<<<<<<<
 *     finally:
 *         FreeContextBuffer(info.SessionKey)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(((char *)__pyx_v_info.SessionKey) + 0, __pyx_v_info.SessionKeyLength - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 780, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L4_return;
  }

  /* "spnego/_sspi_raw/sspi.pyx":782
 *         return (<char *>info.SessionKey)[:info.SessionKeyLength]
 *     finally:
 *         FreeContextBuffer(info.SessionKey)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*finally:*/ {
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_6 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        (void)(FreeContextBuffer(__pyx_v_info.SessionKey));
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
    }
    __pyx_L4_return: {
      __pyx_t_14 = __pyx_r;
      __pyx_r = 0;
      (void)(FreeContextBuffer(__pyx_v_info.SessionKey));
      __pyx_r = __pyx_t_14;
      __pyx_t_14 = 0;
      goto __pyx_L0;
    }
  }

  /* "spnego/_sspi_raw/sspi.pyx":772
 * 
 * 
 * def _query_context_session_key(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_SessionKey info
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi._query_context_session_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":785
 * 
 * 
 * def _query_context_sizes(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_Sizes sizes
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_21_query_context_sizes(PyObject *__pyx_self, PyObject *__pyx_v_context); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_21_query_context_sizes = {"_query_context_sizes", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_21_query_context_sizes, METH_O, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_21_query_context_sizes(PyObject *__pyx_self, PyObject *__pyx_v_context) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_query_context_sizes (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext, 0, "context", 0))) __PYX_ERR(1, 785, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_20_query_context_sizes(__pyx_self, ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)__pyx_v_context));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_20_query_context_sizes(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context) {
  SecPkgContext_Sizes __pyx_v_sizes;
  SECURITY_STATUS __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  unsigned long __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_query_context_sizes", 0);

  /* "spnego/_sspi_raw/sspi.pyx":788
 *     cdef SecPkgContext_Sizes sizes
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.sizes, &sizes)             # <<<<<<<<<<<<<<
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SecPkgAttr); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_sizes); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_long(__pyx_t_2); if (unlikely((__pyx_t_3 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 788, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_res = QueryContextAttributesW((&__pyx_v_context->handle), __pyx_t_3, (&__pyx_v_sizes));

  /* "spnego/_sspi_raw/sspi.pyx":789
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.sizes, &sizes)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  __pyx_t_4 = ((__pyx_v_res != SEC_E_OK) != 0);
  if (__pyx_t_4) {

    /* "spnego/_sspi_raw/sspi.pyx":790
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.sizes, &sizes)
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)             # <<<<<<<<<<<<<<
 * 
 *     return SecPkgAttrSizes(sizes.cbMaxToken, sizes.cbMaxSignature, sizes.cbBlockSize, sizes.cbSecurityTrailer)
 */
    __pyx_t_5 = PyErr_SetFromWindowsErr(__pyx_v_res); if (unlikely(__pyx_t_5 == ((PyObject *)NULL))) __PYX_ERR(1, 790, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":789
 * 
 *     res = QueryContextAttributesW(&context.handle, SecPkgAttr.sizes, &sizes)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":792
 *         PyErr_SetFromWindowsErr(res)
 * 
 *     return SecPkgAttrSizes(sizes.cbMaxToken, sizes.cbMaxSignature, sizes.cbBlockSize, sizes.cbSecurityTrailer)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SecPkgAttrSizes); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_long(__pyx_v_sizes.cbMaxToken); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_long(__pyx_v_sizes.cbMaxSignature); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_long(__pyx_v_sizes.cbBlockSize); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_unsigned_long(__pyx_v_sizes.cbSecurityTrailer); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = NULL;
  __pyx_t_11 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_11 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_10, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_11, 4+__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 792, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_10, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_11, 4+__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 792, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  } else
  #endif
  {
    __pyx_t_12 = PyTuple_New(4+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__pyx_t_10) {
      __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_11, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_12, 3+__pyx_t_11, __pyx_t_9);
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":785
 * 
 * 
 * def _query_context_sizes(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_Sizes sizes
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi._query_context_sizes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "spnego/_sspi_raw/sspi.pyx":795
 * 
 * 
 * def verify_signature(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):             # <<<<<<<<<<<<<<
 *     cdef unsigned long qop = 0
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_23verify_signature(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_23verify_signature = {"verify_signature", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6spnego_9_sspi_raw_4sspi_23verify_signature, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_23verify_signature(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context = 0;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message = 0;
  unsigned long __pyx_v_seq_no;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("verify_signature (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,&__pyx_n_s_message,&__pyx_n_s_seq_no,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_context)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_message)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("verify_signature", 0, 2, 3, 1); __PYX_ERR(1, 795, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seq_no);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "verify_signature") < 0)) __PYX_ERR(1, 795, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_context = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *)values[0]);
    __pyx_v_message = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)values[1]);
    if (values[2]) {
      __pyx_v_seq_no = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_seq_no == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 795, __pyx_L3_error)
    } else {
      __pyx_v_seq_no = ((unsigned long)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("verify_signature", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 795, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.verify_signature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext, 0, "context", 0))) __PYX_ERR(1, 795, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_message), __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBufferDesc, 0, "message", 0))) __PYX_ERR(1, 795, __pyx_L1_error)
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_22verify_signature(__pyx_self, __pyx_v_context, __pyx_v_message, __pyx_v_seq_no);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_22verify_signature(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext *__pyx_v_context, struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *__pyx_v_message, unsigned long __pyx_v_seq_no) {
  unsigned long __pyx_v_qop;
  SECURITY_STATUS __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("verify_signature", 0);

  /* "spnego/_sspi_raw/sspi.pyx":796
 * 
 * def verify_signature(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):
 *     cdef unsigned long qop = 0             # <<<<<<<<<<<<<<
 * 
 *     res = VerifySignature(&context.handle, message.__c_value__(), seq_no, &qop)
 */
  __pyx_v_qop = 0;

  /* "spnego/_sspi_raw/sspi.pyx":798
 *     cdef unsigned long qop = 0
 * 
 *     res = VerifySignature(&context.handle, message.__c_value__(), seq_no, &qop)             # <<<<<<<<<<<<<<
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)
 */
  __pyx_v_res = VerifySignature((&__pyx_v_context->handle), ((struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)__pyx_v_message->__pyx_vtab)->__pyx___c_value__(__pyx_v_message), __pyx_v_seq_no, (&__pyx_v_qop));

  /* "spnego/_sspi_raw/sspi.pyx":799
 * 
 *     res = VerifySignature(&context.handle, message.__c_value__(), seq_no, &qop)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  __pyx_t_1 = ((__pyx_v_res != SEC_E_OK) != 0);
  if (__pyx_t_1) {

    /* "spnego/_sspi_raw/sspi.pyx":800
 *     res = VerifySignature(&context.handle, message.__c_value__(), seq_no, &qop)
 *     if res != _SEC_E_OK:
 *         PyErr_SetFromWindowsErr(res)             # <<<<<<<<<<<<<<
 * 
 *     return qop
 */
    __pyx_t_2 = PyErr_SetFromWindowsErr(__pyx_v_res); if (unlikely(__pyx_t_2 == ((PyObject *)NULL))) __PYX_ERR(1, 800, __pyx_L1_error)

    /* "spnego/_sspi_raw/sspi.pyx":799
 * 
 *     res = VerifySignature(&context.handle, message.__c_value__(), seq_no, &qop)
 *     if res != _SEC_E_OK:             # <<<<<<<<<<<<<<
 *         PyErr_SetFromWindowsErr(res)
 * 
 */
  }

  /* "spnego/_sspi_raw/sspi.pyx":802
 *         PyErr_SetFromWindowsErr(res)
 * 
 *     return qop             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_v_qop); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "spnego/_sspi_raw/sspi.pyx":795
 * 
 * 
 * def verify_signature(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):             # <<<<<<<<<<<<<<
 *     cdef unsigned long qop = 0
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.verify_signature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__AuthIdentityBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_25__pyx_unpickle__AuthIdentityBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6spnego_9_sspi_raw_4sspi_25__pyx_unpickle__AuthIdentityBase = {"__pyx_unpickle__AuthIdentityBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6spnego_9_sspi_raw_4sspi_25__pyx_unpickle__AuthIdentityBase, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6spnego_9_sspi_raw_4sspi_25__pyx_unpickle__AuthIdentityBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__AuthIdentityBase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__AuthIdentityBase", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__AuthIdentityBase", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle__AuthIdentityBase") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__AuthIdentityBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.__pyx_unpickle__AuthIdentityBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6spnego_9_sspi_raw_4sspi_24__pyx_unpickle__AuthIdentityBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6spnego_9_sspi_raw_4sspi_24__pyx_unpickle__AuthIdentityBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__AuthIdentityBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = _AuthIdentityBase.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = _AuthIdentityBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = _AuthIdentityBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__AuthIdentityBase__set_state(<_AuthIdentityBase> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6spnego_9_sspi_raw_4sspi__AuthIdentityBase), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = _AuthIdentityBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__AuthIdentityBase__set_state(<_AuthIdentityBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = _AuthIdentityBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__AuthIdentityBase__set_state(<_AuthIdentityBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__AuthIdentityBase__set_state(_AuthIdentityBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6spnego_9_sspi_raw_4sspi___pyx_unpickle__AuthIdentityBase__set_state(((struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = _AuthIdentityBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__AuthIdentityBase__set_state(<_AuthIdentityBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__AuthIdentityBase__set_state(<_AuthIdentityBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__AuthIdentityBase__set_state(_AuthIdentityBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__AuthIdentityBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.__pyx_unpickle__AuthIdentityBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__AuthIdentityBase__set_state(<_AuthIdentityBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__AuthIdentityBase__set_state(_AuthIdentityBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_6spnego_9_sspi_raw_4sspi___pyx_unpickle__AuthIdentityBase__set_state(struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__AuthIdentityBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__AuthIdentityBase__set_state(_AuthIdentityBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__AuthIdentityBase__set_state(_AuthIdentityBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__AuthIdentityBase__set_state(_AuthIdentityBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__AuthIdentityBase__set_state(<_AuthIdentityBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__AuthIdentityBase__set_state(_AuthIdentityBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("spnego._sspi_raw.sspi.__pyx_unpickle__AuthIdentityBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi_Credential(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi_Credential(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6spnego_9_sspi_raw_4sspi_10Credential_expiry(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_6expiry_1__get__(o);
}

static PyMethodDef __pyx_methods_6spnego_9_sspi_raw_4sspi_Credential[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_10Credential_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6spnego_9_sspi_raw_4sspi_Credential[] = {
  {(char *)"expiry", __pyx_getprop_6spnego_9_sspi_raw_4sspi_10Credential_expiry, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6spnego_9_sspi_raw_4sspi_Credential = {
  PyVarObject_HEAD_INIT(0, 0)
  "spnego._sspi_raw.sspi.Credential", /*tp_name*/
  sizeof(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_Credential), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi_Credential, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6spnego_9_sspi_raw_4sspi_Credential, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6spnego_9_sspi_raw_4sspi_Credential, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6spnego_9_sspi_raw_4sspi_Credential, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi_SecurityContext(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi_SecurityContext(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6spnego_9_sspi_raw_4sspi_15SecurityContext_context_attr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_12context_attr_1__get__(o);
}

static PyObject *__pyx_getprop_6spnego_9_sspi_raw_4sspi_15SecurityContext_expiry(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_6expiry_1__get__(o);
}

static PyMethodDef __pyx_methods_6spnego_9_sspi_raw_4sspi_SecurityContext[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_15SecurityContext_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6spnego_9_sspi_raw_4sspi_SecurityContext[] = {
  {(char *)"context_attr", __pyx_getprop_6spnego_9_sspi_raw_4sspi_15SecurityContext_context_attr, 0, (char *)0, 0},
  {(char *)"expiry", __pyx_getprop_6spnego_9_sspi_raw_4sspi_15SecurityContext_expiry, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6spnego_9_sspi_raw_4sspi_SecurityContext = {
  PyVarObject_HEAD_INIT(0, 0)
  "spnego._sspi_raw.sspi.SecurityContext", /*tp_name*/
  sizeof(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecurityContext), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi_SecurityContext, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6spnego_9_sspi_raw_4sspi_SecurityContext, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6spnego_9_sspi_raw_4sspi_SecurityContext, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6spnego_9_sspi_raw_4sspi_SecurityContext, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBufferDesc __pyx_vtable_6spnego_9_sspi_raw_4sspi_SecBufferDesc;

static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi_SecBufferDesc(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)o);
  p->__pyx_vtab = __pyx_vtabptr_6spnego_9_sspi_raw_4sspi_SecBufferDesc;
  p->_buffers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi_SecBufferDesc(PyObject *o) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *p = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_16__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_buffers);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6spnego_9_sspi_raw_4sspi_SecBufferDesc(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *p = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)o;
  if (p->_buffers) {
    e = (*v)(p->_buffers, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6spnego_9_sspi_raw_4sspi_SecBufferDesc(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *p = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *)o;
  tmp = ((PyObject*)p->_buffers);
  p->_buffers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6spnego_9_sspi_raw_4sspi_SecBufferDesc(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_6spnego_9_sspi_raw_4sspi_SecBufferDesc(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7__setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_version(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version_1__get__(o);
}

static int __pyx_setprop_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_version(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_7version_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6spnego_9_sspi_raw_4sspi_SecBufferDesc[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_18__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_20__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6spnego_9_sspi_raw_4sspi_SecBufferDesc[] = {
  {(char *)"version", __pyx_getprop_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_version, __pyx_setprop_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_version, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_SecBufferDesc = {
  __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6spnego_9_sspi_raw_4sspi_SecBufferDesc, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_SecBufferDesc = {
  __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_3__len__, /*mp_length*/
  __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_5__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_6spnego_9_sspi_raw_4sspi_SecBufferDesc, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6spnego_9_sspi_raw_4sspi_SecBufferDesc = {
  PyVarObject_HEAD_INIT(0, 0)
  "spnego._sspi_raw.sspi.SecBufferDesc", /*tp_name*/
  sizeof(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi_SecBufferDesc, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_12__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_SecBufferDesc, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_SecBufferDesc, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_14__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6spnego_9_sspi_raw_4sspi_SecBufferDesc, /*tp_traverse*/
  __pyx_tp_clear_6spnego_9_sspi_raw_4sspi_SecBufferDesc, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_6spnego_9_sspi_raw_4sspi_13SecBufferDesc_9__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6spnego_9_sspi_raw_4sspi_SecBufferDesc, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6spnego_9_sspi_raw_4sspi_SecBufferDesc, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6spnego_9_sspi_raw_4sspi_SecBufferDesc, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_SecBuffer __pyx_vtable_6spnego_9_sspi_raw_4sspi_SecBuffer;

static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi_SecBuffer(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)o);
  p->__pyx_vtab = __pyx_vtabptr_6spnego_9_sspi_raw_4sspi_SecBuffer;
  p->sys_alloc = Py_None; Py_INCREF(Py_None);
  p->_is_owner = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi_SecBuffer(PyObject *o) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *p = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_13__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->sys_alloc);
  Py_CLEAR(p->_is_owner);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6spnego_9_sspi_raw_4sspi_SecBuffer(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *p = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)o;
  if (p->sys_alloc) {
    e = (*v)(p->sys_alloc, a); if (e) return e;
  }
  if (p->_is_owner) {
    e = (*v)(p->_is_owner, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6spnego_9_sspi_raw_4sspi_SecBuffer(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *p = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *)o;
  tmp = ((PyObject*)p->sys_alloc);
  p->sys_alloc = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_is_owner);
  p->_is_owner = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6spnego_9_sspi_raw_4sspi_9SecBuffer_buffer_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type_1__get__(o);
}

static int __pyx_setprop_6spnego_9_sspi_raw_4sspi_9SecBuffer_buffer_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_11buffer_type_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6spnego_9_sspi_raw_4sspi_9SecBuffer_buffer(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer_1__get__(o);
}

static int __pyx_setprop_6spnego_9_sspi_raw_4sspi_9SecBuffer_buffer(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_6buffer_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6spnego_9_sspi_raw_4sspi_SecBuffer[] = {
  {"_alloc_buffer", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_9_alloc_buffer, METH_O, 0},
  {"_dealloc_buffer", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_11_dealloc_buffer, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_15__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_17__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6spnego_9_sspi_raw_4sspi_SecBuffer[] = {
  {(char *)"buffer_type", __pyx_getprop_6spnego_9_sspi_raw_4sspi_9SecBuffer_buffer_type, __pyx_setprop_6spnego_9_sspi_raw_4sspi_9SecBuffer_buffer_type, (char *)0, 0},
  {(char *)"buffer", __pyx_getprop_6spnego_9_sspi_raw_4sspi_9SecBuffer_buffer, __pyx_setprop_6spnego_9_sspi_raw_4sspi_9SecBuffer_buffer, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_SecBuffer = {
  __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_SecBuffer = {
  __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_3__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6spnego_9_sspi_raw_4sspi_SecBuffer = {
  PyVarObject_HEAD_INIT(0, 0)
  "spnego._sspi_raw.sspi.SecBuffer", /*tp_name*/
  sizeof(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi_SecBuffer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_SecBuffer, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_SecBuffer, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_6spnego_9_sspi_raw_4sspi_9SecBuffer_7__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6spnego_9_sspi_raw_4sspi_SecBuffer, /*tp_traverse*/
  __pyx_tp_clear_6spnego_9_sspi_raw_4sspi_SecBuffer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6spnego_9_sspi_raw_4sspi_SecBuffer, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6spnego_9_sspi_raw_4sspi_SecBuffer, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6spnego_9_sspi_raw_4sspi_SecBuffer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi__AuthIdentityBase __pyx_vtable_6spnego_9_sspi_raw_4sspi__AuthIdentityBase;

static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi__AuthIdentityBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_6spnego_9_sspi_raw_4sspi__AuthIdentityBase;
  return o;
}

static void __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi__AuthIdentityBase(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_6spnego_9_sspi_raw_4sspi__AuthIdentityBase[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6spnego_9_sspi_raw_4sspi__AuthIdentityBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "spnego._sspi_raw.sspi._AuthIdentityBase", /*tp_name*/
  sizeof(struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi__AuthIdentityBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6spnego_9_sspi_raw_4sspi__AuthIdentityBase, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6spnego_9_sspi_raw_4sspi__AuthIdentityBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity __pyx_vtable_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity;

static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *p;
  PyObject *o = __pyx_tp_new_6spnego_9_sspi_raw_4sspi__AuthIdentityBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6spnego_9_sspi_raw_4sspi__AuthIdentityBase*)__pyx_vtabptr_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity;
  p->_username = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)Py_None); Py_INCREF(Py_None);
  p->_domain = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)Py_None); Py_INCREF(Py_None);
  p->_password = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity(PyObject *o) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *p = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_7__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_username);
  Py_CLEAR(p->_domain);
  Py_CLEAR(p->_password);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi__AuthIdentityBase(o);
}

static int __pyx_tp_traverse_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *p = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)o;
  e = ((likely(__pyx_ptype_6spnego_9_sspi_raw_4sspi__AuthIdentityBase)) ? ((__pyx_ptype_6spnego_9_sspi_raw_4sspi__AuthIdentityBase->tp_traverse) ? __pyx_ptype_6spnego_9_sspi_raw_4sspi__AuthIdentityBase->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity)); if (e) return e;
  if (p->_username) {
    e = (*v)(((PyObject *)p->_username), a); if (e) return e;
  }
  if (p->_domain) {
    e = (*v)(((PyObject *)p->_domain), a); if (e) return e;
  }
  if (p->_password) {
    e = (*v)(((PyObject *)p->_password), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *p = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity *)o;
  if (likely(__pyx_ptype_6spnego_9_sspi_raw_4sspi__AuthIdentityBase)) { if (__pyx_ptype_6spnego_9_sspi_raw_4sspi__AuthIdentityBase->tp_clear) __pyx_ptype_6spnego_9_sspi_raw_4sspi__AuthIdentityBase->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity);
  tmp = ((PyObject*)p->_username);
  p->_username = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_domain);
  p->_domain = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_password);
  p->_password = ((struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_username(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username_1__get__(o);
}

static int __pyx_setprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_username(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8username_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_domain(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain_1__get__(o);
}

static int __pyx_setprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_domain(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_6domain_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_password(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password_1__get__(o);
}

static int __pyx_setprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_password(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_8password_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity[] = {
  {(char *)"username", __pyx_getprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_username, __pyx_setprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_username, (char *)0, 0},
  {(char *)"domain", __pyx_getprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_domain, __pyx_setprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_domain, (char *)0, 0},
  {(char *)"password", __pyx_getprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_password, __pyx_setprop_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_password, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity = {
  PyVarObject_HEAD_INIT(0, 0)
  "spnego._sspi_raw.sspi.WinNTAuthIdentity", /*tp_name*/
  sizeof(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity_5__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity, /*tp_traverse*/
  __pyx_tp_clear_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ *__pyx_freelist_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__[8];
static int __pyx_freecount_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ = 0;

static PyObject *__pyx_tp_new_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__)))) {
    o = (PyObject*)__pyx_freelist_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__[--__pyx_freecount_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__];
    memset(o, 0, sizeof(struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__(PyObject *o) {
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ *p = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_val);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__)))) {
    __pyx_freelist_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__[__pyx_freecount_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__++] = ((struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ *p = (struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_val) {
    e = (*v)(p->__pyx_v_val, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "spnego._sspi_raw.sspi.__pyx_scope_struct____iter__", /*tp_name*/
  sizeof(struct __pyx_obj_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_sspi(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_sspi},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "sspi",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_0_1, __pyx_k_0_1, sizeof(__pyx_k_0_1), 0, 1, 0, 0},
  {&__pyx_kp_u_0_1_2_3, __pyx_k_0_1_2_3, sizeof(__pyx_k_0_1_2_3), 0, 1, 0, 0},
  {&__pyx_kp_u_0_1_cbBuffer_2_BufferType_3_pvB, __pyx_k_0_1_cbBuffer_2_BufferType_3_pvB, sizeof(__pyx_k_0_1_cbBuffer_2_BufferType_3_pvB), 0, 1, 0, 0},
  {&__pyx_kp_u_0_1_ulVersion_2_cBuffers_3, __pyx_k_0_1_ulVersion_2_cBuffers_3, sizeof(__pyx_k_0_1_ulVersion_2_cBuffers_3), 0, 1, 0, 0},
  {&__pyx_kp_u_0_cbBuffer_1_BufferType_2_pvBuf, __pyx_k_0_cbBuffer_1_BufferType_2_pvBuf, sizeof(__pyx_k_0_cbBuffer_1_BufferType_2_pvBuf), 0, 1, 0, 0},
  {&__pyx_kp_u_0_ulVersion_1_cBuffers_2, __pyx_k_0_ulVersion_1_cBuffers_2, sizeof(__pyx_k_0_ulVersion_1_cBuffers_2), 0, 1, 0, 0},
  {&__pyx_n_s_AuthIdentityBase, __pyx_k_AuthIdentityBase, sizeof(__pyx_k_AuthIdentityBase), 0, 0, 1, 1},
  {&__pyx_kp_u_Cannot_malloc_SecBufferDesc_buff, __pyx_k_Cannot_malloc_SecBufferDesc_buff, sizeof(__pyx_k_Cannot_malloc_SecBufferDesc_buff), 0, 1, 0, 0},
  {&__pyx_kp_u_Cannot_malloc_SecBuffer_buffer, __pyx_k_Cannot_malloc_SecBuffer_buffer, sizeof(__pyx_k_Cannot_malloc_SecBuffer_buffer), 0, 1, 0, 0},
  {&__pyx_n_s_ClientContextAttr, __pyx_k_ClientContextAttr, sizeof(__pyx_k_ClientContextAttr), 0, 0, 1, 1},
  {&__pyx_n_s_ClientContextReq, __pyx_k_ClientContextReq, sizeof(__pyx_k_ClientContextReq), 0, 0, 1, 1},
  {&__pyx_n_s_Credential, __pyx_k_Credential, sizeof(__pyx_k_Credential), 0, 0, 1, 1},
  {&__pyx_n_s_CredentialUse, __pyx_k_CredentialUse, sizeof(__pyx_k_CredentialUse), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_k_Incompatible_checksums_s_vs_0xd4, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xd4), 0, 0, 1, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_kp_u_Only_an_empty_buffer_can_be_crea, __pyx_k_Only_an_empty_buffer_can_be_crea, sizeof(__pyx_k_Only_an_empty_buffer_can_be_crea), 0, 1, 0, 0},
  {&__pyx_kp_u_Only_names_package_info_session, __pyx_k_Only_names_package_info_session, sizeof(__pyx_k_Only_names_package_info_session), 0, 1, 0, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_BUFFER_TOO_SMALL, __pyx_k_SEC_E_BUFFER_TOO_SMALL, sizeof(__pyx_k_SEC_E_BUFFER_TOO_SMALL), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_CONTEXT_EXPIRED, __pyx_k_SEC_E_CONTEXT_EXPIRED, sizeof(__pyx_k_SEC_E_CONTEXT_EXPIRED), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_CRYPTO_SYSTEM_INVALID, __pyx_k_SEC_E_CRYPTO_SYSTEM_INVALID, sizeof(__pyx_k_SEC_E_CRYPTO_SYSTEM_INVALID), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_INCOMPLETE_MESSAGE, __pyx_k_SEC_E_INCOMPLETE_MESSAGE, sizeof(__pyx_k_SEC_E_INCOMPLETE_MESSAGE), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_INSUFFICIENT_MEMORY, __pyx_k_SEC_E_INSUFFICIENT_MEMORY, sizeof(__pyx_k_SEC_E_INSUFFICIENT_MEMORY), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_INTERNAL_ERROR, __pyx_k_SEC_E_INTERNAL_ERROR, sizeof(__pyx_k_SEC_E_INTERNAL_ERROR), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_INVALID_HANDLE, __pyx_k_SEC_E_INVALID_HANDLE, sizeof(__pyx_k_SEC_E_INVALID_HANDLE), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_INVALID_TOKEN, __pyx_k_SEC_E_INVALID_TOKEN, sizeof(__pyx_k_SEC_E_INVALID_TOKEN), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_LOGON_DENIED, __pyx_k_SEC_E_LOGON_DENIED, sizeof(__pyx_k_SEC_E_LOGON_DENIED), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_MESSAGE_ALTERED, __pyx_k_SEC_E_MESSAGE_ALTERED, sizeof(__pyx_k_SEC_E_MESSAGE_ALTERED), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_NOT_OWNER, __pyx_k_SEC_E_NOT_OWNER, sizeof(__pyx_k_SEC_E_NOT_OWNER), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_NO_AUTHENTICATING_AUTHORIT, __pyx_k_SEC_E_NO_AUTHENTICATING_AUTHORIT, sizeof(__pyx_k_SEC_E_NO_AUTHENTICATING_AUTHORIT), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_NO_CREDENTIALS, __pyx_k_SEC_E_NO_CREDENTIALS, sizeof(__pyx_k_SEC_E_NO_CREDENTIALS), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_OK, __pyx_k_SEC_E_OK, sizeof(__pyx_k_SEC_E_OK), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_OUT_OF_SEQUENCE, __pyx_k_SEC_E_OUT_OF_SEQUENCE, sizeof(__pyx_k_SEC_E_OUT_OF_SEQUENCE), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_QOP_NOT_SUPPORTED, __pyx_k_SEC_E_QOP_NOT_SUPPORTED, sizeof(__pyx_k_SEC_E_QOP_NOT_SUPPORTED), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_SECPKG_NOT_FOUND, __pyx_k_SEC_E_SECPKG_NOT_FOUND, sizeof(__pyx_k_SEC_E_SECPKG_NOT_FOUND), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_TARGET_UNKNOWN, __pyx_k_SEC_E_TARGET_UNKNOWN, sizeof(__pyx_k_SEC_E_TARGET_UNKNOWN), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_UNKNOWN_CREDENTIALS, __pyx_k_SEC_E_UNKNOWN_CREDENTIALS, sizeof(__pyx_k_SEC_E_UNKNOWN_CREDENTIALS), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_UNSUPPORTED_FUNCTION, __pyx_k_SEC_E_UNSUPPORTED_FUNCTION, sizeof(__pyx_k_SEC_E_UNSUPPORTED_FUNCTION), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_E_WRONG_PRINCIPAL, __pyx_k_SEC_E_WRONG_PRINCIPAL, sizeof(__pyx_k_SEC_E_WRONG_PRINCIPAL), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_I_COMPLETE_AND_CONTINUE, __pyx_k_SEC_I_COMPLETE_AND_CONTINUE, sizeof(__pyx_k_SEC_I_COMPLETE_AND_CONTINUE), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_I_COMPLETE_NEEDED, __pyx_k_SEC_I_COMPLETE_NEEDED, sizeof(__pyx_k_SEC_I_COMPLETE_NEEDED), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_I_CONTINUE_NEEDED, __pyx_k_SEC_I_CONTINUE_NEEDED, sizeof(__pyx_k_SEC_I_CONTINUE_NEEDED), 0, 0, 1, 1},
  {&__pyx_n_s_SEC_I_INCOMPLETE_CREDENTIALS, __pyx_k_SEC_I_INCOMPLETE_CREDENTIALS, sizeof(__pyx_k_SEC_I_INCOMPLETE_CREDENTIALS), 0, 0, 1, 1},
  {&__pyx_n_s_SSPIQoP, __pyx_k_SSPIQoP, sizeof(__pyx_k_SSPIQoP), 0, 0, 1, 1},
  {&__pyx_n_s_SecBuffer, __pyx_k_SecBuffer, sizeof(__pyx_k_SecBuffer), 0, 0, 1, 1},
  {&__pyx_n_s_SecBufferDesc, __pyx_k_SecBufferDesc, sizeof(__pyx_k_SecBufferDesc), 0, 0, 1, 1},
  {&__pyx_n_s_SecBufferDesc___iter, __pyx_k_SecBufferDesc___iter, sizeof(__pyx_k_SecBufferDesc___iter), 0, 0, 1, 1},
  {&__pyx_n_s_SecBufferType, __pyx_k_SecBufferType, sizeof(__pyx_k_SecBufferType), 0, 0, 1, 1},
  {&__pyx_n_s_SecPkgAttr, __pyx_k_SecPkgAttr, sizeof(__pyx_k_SecPkgAttr), 0, 0, 1, 1},
  {&__pyx_n_s_SecPkgAttrSizes, __pyx_k_SecPkgAttrSizes, sizeof(__pyx_k_SecPkgAttrSizes), 0, 0, 1, 1},
  {&__pyx_n_u_SecPkgAttrSizes, __pyx_k_SecPkgAttrSizes, sizeof(__pyx_k_SecPkgAttrSizes), 0, 1, 0, 1},
  {&__pyx_n_s_SecPkgInfo, __pyx_k_SecPkgInfo, sizeof(__pyx_k_SecPkgInfo), 0, 0, 1, 1},
  {&__pyx_n_u_SecPkgInfo, __pyx_k_SecPkgInfo, sizeof(__pyx_k_SecPkgInfo), 0, 1, 0, 1},
  {&__pyx_n_s_SecStatus, __pyx_k_SecStatus, sizeof(__pyx_k_SecStatus), 0, 0, 1, 1},
  {&__pyx_n_s_SecurityContext, __pyx_k_SecurityContext, sizeof(__pyx_k_SecurityContext), 0, 0, 1, 1},
  {&__pyx_n_s_ServerContextAttr, __pyx_k_ServerContextAttr, sizeof(__pyx_k_ServerContextAttr), 0, 0, 1, 1},
  {&__pyx_n_s_ServerContextReq, __pyx_k_ServerContextReq, sizeof(__pyx_k_ServerContextReq), 0, 0, 1, 1},
  {&__pyx_n_s_TargetDataRep, __pyx_k_TargetDataRep, sizeof(__pyx_k_TargetDataRep), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_WinNTAuthIdentity, __pyx_k_WinNTAuthIdentity, sizeof(__pyx_k_WinNTAuthIdentity), 0, 0, 1, 1},
  {&__pyx_kp_b__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 0, 0, 0},
  {&__pyx_kp_u__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 1, 0, 0},
  {&__pyx_n_s_accept_security_context, __pyx_k_accept_security_context, sizeof(__pyx_k_accept_security_context), 0, 0, 1, 1},
  {&__pyx_n_s_acquire_credentials_handle, __pyx_k_acquire_credentials_handle, sizeof(__pyx_k_acquire_credentials_handle), 0, 0, 1, 1},
  {&__pyx_n_s_alloc_buffer, __pyx_k_alloc_buffer, sizeof(__pyx_k_alloc_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_allocate_memory, __pyx_k_allocate_memory, sizeof(__pyx_k_allocate_memory), 0, 0, 1, 1},
  {&__pyx_n_s_allocated_memory, __pyx_k_allocated_memory, sizeof(__pyx_k_allocated_memory), 0, 0, 1, 1},
  {&__pyx_n_s_allow_context_replay, __pyx_k_allow_context_replay, sizeof(__pyx_k_allow_context_replay), 0, 0, 1, 1},
  {&__pyx_n_s_allow_missing_binding, __pyx_k_allow_missing_binding, sizeof(__pyx_k_allow_missing_binding), 0, 0, 1, 1},
  {&__pyx_n_s_allow_non_user_logons, __pyx_k_allow_non_user_logons, sizeof(__pyx_k_allow_non_user_logons), 0, 0, 1, 1},
  {&__pyx_n_s_allow_null_session, __pyx_k_allow_null_session, sizeof(__pyx_k_allow_null_session), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_attribute, __pyx_k_attribute, sizeof(__pyx_k_attribute), 0, 0, 1, 1},
  {&__pyx_n_s_attrmask, __pyx_k_attrmask, sizeof(__pyx_k_attrmask), 0, 0, 1, 1},
  {&__pyx_n_s_auth_data, __pyx_k_auth_data, sizeof(__pyx_k_auth_data), 0, 0, 1, 1},
  {&__pyx_n_u_block_size, __pyx_k_block_size, sizeof(__pyx_k_block_size), 0, 1, 0, 1},
  {&__pyx_n_s_both, __pyx_k_both, sizeof(__pyx_k_both), 0, 0, 1, 1},
  {&__pyx_n_s_buffer, __pyx_k_buffer, sizeof(__pyx_k_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_buffer_type, __pyx_k_buffer_type, sizeof(__pyx_k_buffer_type), 0, 0, 1, 1},
  {&__pyx_n_s_buffers, __pyx_k_buffers, sizeof(__pyx_k_buffers), 0, 0, 1, 1},
  {&__pyx_n_s_cBuffers, __pyx_k_cBuffers, sizeof(__pyx_k_cBuffers), 0, 0, 1, 1},
  {&__pyx_n_s_call_level, __pyx_k_call_level, sizeof(__pyx_k_call_level), 0, 0, 1, 1},
  {&__pyx_n_u_capabilities, __pyx_k_capabilities, sizeof(__pyx_k_capabilities), 0, 1, 0, 1},
  {&__pyx_n_s_change_pass_response, __pyx_k_change_pass_response, sizeof(__pyx_k_change_pass_response), 0, 0, 1, 1},
  {&__pyx_n_s_channel_bindings, __pyx_k_channel_bindings, sizeof(__pyx_k_channel_bindings), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_u_comment, __pyx_k_comment, sizeof(__pyx_k_comment), 0, 1, 0, 1},
  {&__pyx_n_s_confidentiality, __pyx_k_confidentiality, sizeof(__pyx_k_confidentiality), 0, 0, 1, 1},
  {&__pyx_n_s_connection, __pyx_k_connection, sizeof(__pyx_k_connection), 0, 0, 1, 1},
  {&__pyx_n_s_context, __pyx_k_context, sizeof(__pyx_k_context), 0, 0, 1, 1},
  {&__pyx_n_s_context_req, __pyx_k_context_req, sizeof(__pyx_k_context_req), 0, 0, 1, 1},
  {&__pyx_n_s_cred, __pyx_k_cred, sizeof(__pyx_k_cred), 0, 0, 1, 1},
  {&__pyx_n_s_credential, __pyx_k_credential, sizeof(__pyx_k_credential), 0, 0, 1, 1},
  {&__pyx_n_s_credential_use, __pyx_k_credential_use, sizeof(__pyx_k_credential_use), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_datagram, __pyx_k_datagram, sizeof(__pyx_k_datagram), 0, 0, 1, 1},
  {&__pyx_n_s_dealloc_buffer, __pyx_k_dealloc_buffer, sizeof(__pyx_k_dealloc_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_decrypt_message, __pyx_k_decrypt_message, sizeof(__pyx_k_decrypt_message), 0, 0, 1, 1},
  {&__pyx_n_s_delegate, __pyx_k_delegate, sizeof(__pyx_k_delegate), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_domain, __pyx_k_domain, sizeof(__pyx_k_domain), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encrypt_message, __pyx_k_encrypt_message, sizeof(__pyx_k_encrypt_message), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_expiry, __pyx_k_expiry, sizeof(__pyx_k_expiry), 0, 0, 1, 1},
  {&__pyx_n_s_extended_error, __pyx_k_extended_error, sizeof(__pyx_k_extended_error), 0, 0, 1, 1},
  {&__pyx_n_s_extra, __pyx_k_extra, sizeof(__pyx_k_extra), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_forward_credentials, __pyx_k_forward_credentials, sizeof(__pyx_k_forward_credentials), 0, 0, 1, 1},
  {&__pyx_n_s_fragment_only, __pyx_k_fragment_only, sizeof(__pyx_k_fragment_only), 0, 0, 1, 1},
  {&__pyx_n_s_fragment_supplied, __pyx_k_fragment_supplied, sizeof(__pyx_k_fragment_supplied), 0, 0, 1, 1},
  {&__pyx_n_s_fragment_to_fit, __pyx_k_fragment_to_fit, sizeof(__pyx_k_fragment_to_fit), 0, 0, 1, 1},
  {&__pyx_n_s_from_text, __pyx_k_from_text, sizeof(__pyx_k_from_text), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_http, __pyx_k_http, sizeof(__pyx_k_http), 0, 0, 1, 1},
  {&__pyx_n_s_identify, __pyx_k_identify, sizeof(__pyx_k_identify), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inbound, __pyx_k_inbound, sizeof(__pyx_k_inbound), 0, 0, 1, 1},
  {&__pyx_n_s_info, __pyx_k_info, sizeof(__pyx_k_info), 0, 0, 1, 1},
  {&__pyx_n_s_initialize_security_context, __pyx_k_initialize_security_context, sizeof(__pyx_k_initialize_security_context), 0, 0, 1, 1},
  {&__pyx_n_s_input, __pyx_k_input, sizeof(__pyx_k_input), 0, 0, 1, 1},
  {&__pyx_n_s_input_buffer, __pyx_k_input_buffer, sizeof(__pyx_k_input_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_input_context, __pyx_k_input_context, sizeof(__pyx_k_input_context), 0, 0, 1, 1},
  {&__pyx_n_s_integrity, __pyx_k_integrity, sizeof(__pyx_k_integrity), 0, 0, 1, 1},
  {&__pyx_n_s_intermediate_return, __pyx_k_intermediate_return, sizeof(__pyx_k_intermediate_return), 0, 0, 1, 1},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 0, 1, 1},
  {&__pyx_n_s_licensing, __pyx_k_licensing, sizeof(__pyx_k_licensing), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_make_signature, __pyx_k_make_signature, sizeof(__pyx_k_make_signature), 0, 0, 1, 1},
  {&__pyx_n_s_manual_cred_validation, __pyx_k_manual_cred_validation, sizeof(__pyx_k_manual_cred_validation), 0, 0, 1, 1},
  {&__pyx_n_u_max_signature, __pyx_k_max_signature, sizeof(__pyx_k_max_signature), 0, 1, 0, 1},
  {&__pyx_n_u_max_token, __pyx_k_max_token, sizeof(__pyx_k_max_token), 0, 1, 0, 1},
  {&__pyx_n_s_mechlist, __pyx_k_mechlist, sizeof(__pyx_k_mechlist), 0, 0, 1, 1},
  {&__pyx_n_s_mechlist_signature, __pyx_k_mechlist_signature, sizeof(__pyx_k_mechlist_signature), 0, 0, 1, 1},
  {&__pyx_n_s_message, __pyx_k_message, sizeof(__pyx_k_message), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_missing, __pyx_k_missing, sizeof(__pyx_k_missing), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_mutual_auth, __pyx_k_mutual_auth, sizeof(__pyx_k_mutual_auth), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_u_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 1, 0, 1},
  {&__pyx_n_s_namedtuple, __pyx_k_namedtuple, sizeof(__pyx_k_namedtuple), 0, 0, 1, 1},
  {&__pyx_n_s_names, __pyx_k_names, sizeof(__pyx_k_names), 0, 0, 1, 1},
  {&__pyx_n_s_native, __pyx_k_native, sizeof(__pyx_k_native), 0, 0, 1, 1},
  {&__pyx_n_s_negotiation_info, __pyx_k_negotiation_info, sizeof(__pyx_k_negotiation_info), 0, 0, 1, 1},
  {&__pyx_n_s_network, __pyx_k_network, sizeof(__pyx_k_network), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_no_additional_token, __pyx_k_no_additional_token, sizeof(__pyx_k_no_additional_token), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_no_integrity, __pyx_k_no_integrity, sizeof(__pyx_k_no_integrity), 0, 0, 1, 1},
  {&__pyx_n_s_no_token, __pyx_k_no_token, sizeof(__pyx_k_no_token), 0, 0, 1, 1},
  {&__pyx_n_s_null_session, __pyx_k_null_session, sizeof(__pyx_k_null_session), 0, 0, 1, 1},
  {&__pyx_n_s_outbound, __pyx_k_outbound, sizeof(__pyx_k_outbound), 0, 0, 1, 1},
  {&__pyx_n_s_output, __pyx_k_output, sizeof(__pyx_k_output), 0, 0, 1, 1},
  {&__pyx_n_s_output_buffer, __pyx_k_output_buffer, sizeof(__pyx_k_output_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_pBuffers, __pyx_k_pBuffers, sizeof(__pyx_k_pBuffers), 0, 0, 1, 1},
  {&__pyx_n_s_p_auth_data, __pyx_k_p_auth_data, sizeof(__pyx_k_p_auth_data), 0, 0, 1, 1},
  {&__pyx_n_s_package, __pyx_k_package, sizeof(__pyx_k_package), 0, 0, 1, 1},
  {&__pyx_n_s_package_info, __pyx_k_package_info, sizeof(__pyx_k_package_info), 0, 0, 1, 1},
  {&__pyx_n_s_padding, __pyx_k_padding, sizeof(__pyx_k_padding), 0, 0, 1, 1},
  {&__pyx_n_s_password, __pyx_k_password, sizeof(__pyx_k_password), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pkg_params, __pyx_k_pkg_params, sizeof(__pyx_k_pkg_params), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_principal, __pyx_k_principal, sizeof(__pyx_k_principal), 0, 0, 1, 1},
  {&__pyx_n_s_proxy_bindings, __pyx_k_proxy_bindings, sizeof(__pyx_k_proxy_bindings), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle__AuthIdentityBase, __pyx_k_pyx_unpickle__AuthIdentityBase, sizeof(__pyx_k_pyx_unpickle__AuthIdentityBase), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qop, __pyx_k_qop, sizeof(__pyx_k_qop), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_query_context_attributes, __pyx_k_query_context_attributes, sizeof(__pyx_k_query_context_attributes), 0, 0, 1, 1},
  {&__pyx_n_s_query_context_names, __pyx_k_query_context_names, sizeof(__pyx_k_query_context_names), 0, 0, 1, 1},
  {&__pyx_n_s_query_context_package_info, __pyx_k_query_context_package_info, sizeof(__pyx_k_query_context_package_info), 0, 0, 1, 1},
  {&__pyx_n_s_query_context_session_key, __pyx_k_query_context_session_key, sizeof(__pyx_k_query_context_session_key), 0, 0, 1, 1},
  {&__pyx_n_s_query_context_sizes, __pyx_k_query_context_sizes, sizeof(__pyx_k_query_context_sizes), 0, 0, 1, 1},
  {&__pyx_n_s_raw_info, __pyx_k_raw_info, sizeof(__pyx_k_raw_info), 0, 0, 1, 1},
  {&__pyx_n_s_readonly, __pyx_k_readonly, sizeof(__pyx_k_readonly), 0, 0, 1, 1},
  {&__pyx_n_s_readonly_with_checksum, __pyx_k_readonly_with_checksum, sizeof(__pyx_k_readonly_with_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_reauthentication, __pyx_k_reauthentication, sizeof(__pyx_k_reauthentication), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_replay_detect, __pyx_k_replay_detect, sizeof(__pyx_k_replay_detect), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_reserved, __pyx_k_reserved, sizeof(__pyx_k_reserved), 0, 0, 1, 1},
  {&__pyx_n_u_rpcid, __pyx_k_rpcid, sizeof(__pyx_k_rpcid), 0, 1, 0, 1},
  {&__pyx_kp_u_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 1, 0, 0},
  {&__pyx_n_u_security_trailer, __pyx_k_security_trailer, sizeof(__pyx_k_security_trailer), 0, 1, 0, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_seq_no, __pyx_k_seq_no, sizeof(__pyx_k_seq_no), 0, 0, 1, 1},
  {&__pyx_n_s_sequence_detect, __pyx_k_sequence_detect, sizeof(__pyx_k_sequence_detect), 0, 0, 1, 1},
  {&__pyx_n_s_session_key, __pyx_k_session_key, sizeof(__pyx_k_session_key), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_sizes, __pyx_k_sizes, sizeof(__pyx_k_sizes), 0, 0, 1, 1},
  {&__pyx_n_s_spnego__sspi_raw_sspi, __pyx_k_spnego__sspi_raw_sspi, sizeof(__pyx_k_spnego__sspi_raw_sspi), 0, 0, 1, 1},
  {&__pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_k_spnego__sspi_raw_sspi_pyx, sizeof(__pyx_k_spnego__sspi_raw_sspi_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_spnego__text, __pyx_k_spnego__text, sizeof(__pyx_k_spnego__text), 0, 0, 1, 1},
  {&__pyx_n_s_stream, __pyx_k_stream, sizeof(__pyx_k_stream), 0, 0, 1, 1},
  {&__pyx_n_s_stream_header, __pyx_k_stream_header, sizeof(__pyx_k_stream_header), 0, 0, 1, 1},
  {&__pyx_n_s_stream_trailer, __pyx_k_stream_trailer, sizeof(__pyx_k_stream_trailer), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_target, __pyx_k_target, sizeof(__pyx_k_target), 0, 0, 1, 1},
  {&__pyx_n_s_target_data_rep, __pyx_k_target_data_rep, sizeof(__pyx_k_target_data_rep), 0, 0, 1, 1},
  {&__pyx_n_s_target_host, __pyx_k_target_host, sizeof(__pyx_k_target_host), 0, 0, 1, 1},
  {&__pyx_n_s_target_name, __pyx_k_target_name, sizeof(__pyx_k_target_name), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_third_leg_failed, __pyx_k_third_leg_failed, sizeof(__pyx_k_third_leg_failed), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_to_native, __pyx_k_to_native, sizeof(__pyx_k_to_native), 0, 0, 1, 1},
  {&__pyx_n_s_to_text, __pyx_k_to_text, sizeof(__pyx_k_to_text), 0, 0, 1, 1},
  {&__pyx_n_s_token, __pyx_k_token, sizeof(__pyx_k_token), 0, 0, 1, 1},
  {&__pyx_n_s_ulVersion, __pyx_k_ulVersion, sizeof(__pyx_k_ulVersion), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_use_dce_style, __pyx_k_use_dce_style, sizeof(__pyx_k_use_dce_style), 0, 0, 1, 1},
  {&__pyx_n_s_use_session_key, __pyx_k_use_session_key, sizeof(__pyx_k_use_session_key), 0, 0, 1, 1},
  {&__pyx_n_s_use_supplied_creds, __pyx_k_use_supplied_creds, sizeof(__pyx_k_use_supplied_creds), 0, 0, 1, 1},
  {&__pyx_n_s_used_collected_creds, __pyx_k_used_collected_creds, sizeof(__pyx_k_used_collected_creds), 0, 0, 1, 1},
  {&__pyx_n_s_used_dce_style, __pyx_k_used_dce_style, sizeof(__pyx_k_used_dce_style), 0, 0, 1, 1},
  {&__pyx_n_s_used_http_style, __pyx_k_used_http_style, sizeof(__pyx_k_used_http_style), 0, 0, 1, 1},
  {&__pyx_n_s_used_supplied_creds, __pyx_k_used_supplied_creds, sizeof(__pyx_k_used_supplied_creds), 0, 0, 1, 1},
  {&__pyx_n_s_username, __pyx_k_username, sizeof(__pyx_k_username), 0, 0, 1, 1},
  {&__pyx_n_s_verify_signature, __pyx_k_verify_signature, sizeof(__pyx_k_verify_signature), 0, 0, 1, 1},
  {&__pyx_n_s_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 0, 1, 1},
  {&__pyx_n_u_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 1, 0, 1},
  {&__pyx_n_s_w_package, __pyx_k_w_package, sizeof(__pyx_k_w_package), 0, 0, 1, 1},
  {&__pyx_n_s_w_principal, __pyx_k_w_principal, sizeof(__pyx_k_w_principal), 0, 0, 1, 1},
  {&__pyx_n_s_w_target_name, __pyx_k_w_target_name, sizeof(__pyx_k_w_target_name), 0, 0, 1, 1},
  {&__pyx_n_s_wrap_no_encrypt, __pyx_k_wrap_no_encrypt, sizeof(__pyx_k_wrap_no_encrypt), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(1, 416, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(1, 418, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(1, 463, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(1, 729, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "spnego/_sspi_raw/sspi.pyx":416
 *             self._c_value.pBuffers = <PSecBuffer>malloc(sizeof(NativeSecBuffer) * len(buffers))
 *             if not self._c_value.pBuffers:
 *                 raise MemoryError("Cannot malloc SecBufferDesc buffer")             # <<<<<<<<<<<<<<
 * 
 *             for idx, buffer in enumerate(buffers):
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_u_Cannot_malloc_SecBufferDesc_buff); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "spnego/_sspi_raw/sspi.pyx":463
 *     def __cinit__(SecBuffer self, unsigned long buffer_type, bytes buffer=None, unsigned long length=0):
 *         if buffer and length:
 *             raise ValueError("Only an empty buffer can be created with length")             # <<<<<<<<<<<<<<
 * 
 *         self.sys_alloc = 0
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_Only_an_empty_buffer_can_be_crea); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(1, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "spnego/_sspi_raw/sspi.pyx":470
 *         self.c_value = <PSecBuffer>calloc(1, sizeof(NativeSecBuffer))
 *         if not self.c_value:
 *             raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover             # <<<<<<<<<<<<<<
 * 
 *         self.buffer_type = buffer_type
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_u_Cannot_malloc_SecBuffer_buffer); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "spnego/_sspi_raw/sspi.pyx":729
 * def query_context_attributes(SecurityContext context not None, unsigned long attribute):
 *     if attribute not in [SecPkgAttr.names, SecPkgAttr.package_info, SecPkgAttr.session_key, SecPkgAttr.sizes]:
 *         raise NotImplementedError("Only names, package_info, session_key, or sizes is implemented")             # <<<<<<<<<<<<<<
 * 
 *     if attribute == SecPkgAttr.names:
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u_Only_names_package_info_session); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "spnego/_sspi_raw/sspi.pyx":622
 * 
 * 
 * def accept_security_context(Credential credential not None, SecurityContext context not None,             # <<<<<<<<<<<<<<
 *     SecBufferDesc input_buffer=None, unsigned long context_req=0, unsigned long target_data_rep=SECURITY_NATIVE_DREP,
 *     SecBufferDesc output_buffer=None):
 */
  __pyx_tuple__20 = PyTuple_Pack(12, __pyx_n_s_credential, __pyx_n_s_context, __pyx_n_s_input_buffer, __pyx_n_s_context_req, __pyx_n_s_target_data_rep, __pyx_n_s_output_buffer, __pyx_n_s_input_context, __pyx_n_s_input, __pyx_n_s_output, __pyx_n_s_expiry, __pyx_n_s_buffer, __pyx_n_s_res); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(1, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(6, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_accept_security_context, 622, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(1, 622, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":651
 * 
 * 
 * def acquire_credentials_handle(unicode principal, unicode package not None,             # <<<<<<<<<<<<<<
 *     unsigned long credential_use=SECPKG_CRED_OUTBOUND, _AuthIdentityBase auth_data=None):
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(10, __pyx_n_s_principal, __pyx_n_s_package, __pyx_n_s_credential_use, __pyx_n_s_auth_data, __pyx_n_s_w_principal, __pyx_n_s_w_package, __pyx_n_s_p_auth_data, __pyx_n_s_cred, __pyx_n_s_expiry, __pyx_n_s_res); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(1, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(4, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_acquire_credentials_handle, 651, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(1, 651, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":671
 * 
 * 
 * def decrypt_message(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):             # <<<<<<<<<<<<<<
 *     cdef unsigned long qop = 0;
 * 
 */
  __pyx_tuple__24 = PyTuple_Pack(5, __pyx_n_s_context, __pyx_n_s_message, __pyx_n_s_seq_no, __pyx_n_s_qop, __pyx_n_s_res); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(1, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_decrypt_message, 671, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(1, 671, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":681
 * 
 * 
 * def encrypt_message(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0,             # <<<<<<<<<<<<<<
 *     unsigned long qop=0):
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(5, __pyx_n_s_context, __pyx_n_s_message, __pyx_n_s_seq_no, __pyx_n_s_qop, __pyx_n_s_res); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(1, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_encrypt_message, 681, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(1, 681, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":689
 * 
 * 
 * def initialize_security_context(Credential credential not None, SecurityContext context not None,             # <<<<<<<<<<<<<<
 *     unicode target_name not None, unsigned long context_req=0, SecBufferDesc input_buffer=None,
 *     unsigned long target_data_rep=SECURITY_NATIVE_DREP, SecBufferDesc output_buffer=None):
 */
  __pyx_tuple__28 = PyTuple_Pack(14, __pyx_n_s_credential, __pyx_n_s_context, __pyx_n_s_target_name, __pyx_n_s_context_req, __pyx_n_s_input_buffer, __pyx_n_s_target_data_rep, __pyx_n_s_output_buffer, __pyx_n_s_input_context, __pyx_n_s_w_target_name, __pyx_n_s_input, __pyx_n_s_output, __pyx_n_s_expiry, __pyx_n_s_buffer, __pyx_n_s_res); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(1, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(7, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_initialize_security_context, 689, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(1, 689, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":719
 * 
 * 
 * def make_signature(SecurityContext context not None, unsigned long qop, SecBufferDesc message not None,             # <<<<<<<<<<<<<<
 *     unsigned long seq_no=0):
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(5, __pyx_n_s_context, __pyx_n_s_qop, __pyx_n_s_message, __pyx_n_s_seq_no, __pyx_n_s_res); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(1, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_make_signature, 719, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(1, 719, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":727
 * 
 * 
 * def query_context_attributes(SecurityContext context not None, unsigned long attribute):             # <<<<<<<<<<<<<<
 *     if attribute not in [SecPkgAttr.names, SecPkgAttr.package_info, SecPkgAttr.session_key, SecPkgAttr.sizes]:
 *         raise NotImplementedError("Only names, package_info, session_key, or sizes is implemented")
 */
  __pyx_tuple__32 = PyTuple_Pack(2, __pyx_n_s_context, __pyx_n_s_attribute); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(1, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_query_context_attributes, 727, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(1, 727, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":744
 * 
 * 
 * def _query_context_names(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_Names info
 * 
 */
  __pyx_tuple__34 = PyTuple_Pack(3, __pyx_n_s_context, __pyx_n_s_info, __pyx_n_s_res); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_query_context_names, 744, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(1, 744, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":756
 *         FreeContextBuffer(<void*>info.sUserName)
 * 
 * def _query_context_package_info(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_PackageInfoW raw_info
 *     cdef PSecPkgInfoW info
 */
  __pyx_tuple__36 = PyTuple_Pack(4, __pyx_n_s_context, __pyx_n_s_raw_info, __pyx_n_s_info, __pyx_n_s_res); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(1, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_query_context_package_info, 756, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(1, 756, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":772
 * 
 * 
 * def _query_context_session_key(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_SessionKey info
 * 
 */
  __pyx_tuple__38 = PyTuple_Pack(3, __pyx_n_s_context, __pyx_n_s_info, __pyx_n_s_res); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(1, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_query_context_session_key, 772, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(1, 772, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":785
 * 
 * 
 * def _query_context_sizes(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_Sizes sizes
 * 
 */
  __pyx_tuple__40 = PyTuple_Pack(3, __pyx_n_s_context, __pyx_n_s_sizes, __pyx_n_s_res); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(1, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_query_context_sizes, 785, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(1, 785, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":795
 * 
 * 
 * def verify_signature(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):             # <<<<<<<<<<<<<<
 *     cdef unsigned long qop = 0
 * 
 */
  __pyx_tuple__42 = PyTuple_Pack(5, __pyx_n_s_context, __pyx_n_s_message, __pyx_n_s_seq_no, __pyx_n_s_qop, __pyx_n_s_res); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(1, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_spnego__sspi_raw_sspi_pyx, __pyx_n_s_verify_signature, 795, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(1, 795, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle__AuthIdentityBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__44 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__44, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__AuthIdentityBase, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_268435456 = PyInt_FromLong(268435456L); if (unlikely(!__pyx_int_268435456)) __PYX_ERR(1, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_6spnego_9_sspi_raw_4sspi_Credential) < 0) __PYX_ERR(1, 385, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6spnego_9_sspi_raw_4sspi_Credential.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6spnego_9_sspi_raw_4sspi_Credential.tp_dictoffset && __pyx_type_6spnego_9_sspi_raw_4sspi_Credential.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6spnego_9_sspi_raw_4sspi_Credential.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Credential, (PyObject *)&__pyx_type_6spnego_9_sspi_raw_4sspi_Credential) < 0) __PYX_ERR(1, 385, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6spnego_9_sspi_raw_4sspi_Credential) < 0) __PYX_ERR(1, 385, __pyx_L1_error)
  __pyx_ptype_6spnego_9_sspi_raw_4sspi_Credential = &__pyx_type_6spnego_9_sspi_raw_4sspi_Credential;
  if (PyType_Ready(&__pyx_type_6spnego_9_sspi_raw_4sspi_SecurityContext) < 0) __PYX_ERR(1, 395, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6spnego_9_sspi_raw_4sspi_SecurityContext.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6spnego_9_sspi_raw_4sspi_SecurityContext.tp_dictoffset && __pyx_type_6spnego_9_sspi_raw_4sspi_SecurityContext.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6spnego_9_sspi_raw_4sspi_SecurityContext.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SecurityContext, (PyObject *)&__pyx_type_6spnego_9_sspi_raw_4sspi_SecurityContext) < 0) __PYX_ERR(1, 395, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6spnego_9_sspi_raw_4sspi_SecurityContext) < 0) __PYX_ERR(1, 395, __pyx_L1_error)
  __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecurityContext = &__pyx_type_6spnego_9_sspi_raw_4sspi_SecurityContext;
  __pyx_vtabptr_6spnego_9_sspi_raw_4sspi_SecBufferDesc = &__pyx_vtable_6spnego_9_sspi_raw_4sspi_SecBufferDesc;
  __pyx_vtable_6spnego_9_sspi_raw_4sspi_SecBufferDesc.__pyx___c_value__ = (PSecBufferDesc (*)(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBufferDesc *))__pyx_f_6spnego_9_sspi_raw_4sspi_13SecBufferDesc___c_value__;
  if (PyType_Ready(&__pyx_type_6spnego_9_sspi_raw_4sspi_SecBufferDesc) < 0) __PYX_ERR(1, 407, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6spnego_9_sspi_raw_4sspi_SecBufferDesc.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6spnego_9_sspi_raw_4sspi_SecBufferDesc.tp_dictoffset && __pyx_type_6spnego_9_sspi_raw_4sspi_SecBufferDesc.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6spnego_9_sspi_raw_4sspi_SecBufferDesc.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6spnego_9_sspi_raw_4sspi_SecBufferDesc.tp_dict, __pyx_vtabptr_6spnego_9_sspi_raw_4sspi_SecBufferDesc) < 0) __PYX_ERR(1, 407, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SecBufferDesc, (PyObject *)&__pyx_type_6spnego_9_sspi_raw_4sspi_SecBufferDesc) < 0) __PYX_ERR(1, 407, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6spnego_9_sspi_raw_4sspi_SecBufferDesc) < 0) __PYX_ERR(1, 407, __pyx_L1_error)
  __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBufferDesc = &__pyx_type_6spnego_9_sspi_raw_4sspi_SecBufferDesc;
  __pyx_vtabptr_6spnego_9_sspi_raw_4sspi_SecBuffer = &__pyx_vtable_6spnego_9_sspi_raw_4sspi_SecBuffer;
  __pyx_vtable_6spnego_9_sspi_raw_4sspi_SecBuffer.replace_ptr = (PyObject *(*)(struct __pyx_obj_6spnego_9_sspi_raw_4sspi_SecBuffer *, PSecBuffer))__pyx_f_6spnego_9_sspi_raw_4sspi_9SecBuffer_replace_ptr;
  if (PyType_Ready(&__pyx_type_6spnego_9_sspi_raw_4sspi_SecBuffer) < 0) __PYX_ERR(1, 459, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6spnego_9_sspi_raw_4sspi_SecBuffer.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6spnego_9_sspi_raw_4sspi_SecBuffer.tp_dictoffset && __pyx_type_6spnego_9_sspi_raw_4sspi_SecBuffer.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6spnego_9_sspi_raw_4sspi_SecBuffer.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6spnego_9_sspi_raw_4sspi_SecBuffer.tp_dict, __pyx_vtabptr_6spnego_9_sspi_raw_4sspi_SecBuffer) < 0) __PYX_ERR(1, 459, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SecBuffer, (PyObject *)&__pyx_type_6spnego_9_sspi_raw_4sspi_SecBuffer) < 0) __PYX_ERR(1, 459, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6spnego_9_sspi_raw_4sspi_SecBuffer) < 0) __PYX_ERR(1, 459, __pyx_L1_error)
  __pyx_ptype_6spnego_9_sspi_raw_4sspi_SecBuffer = &__pyx_type_6spnego_9_sspi_raw_4sspi_SecBuffer;
  __pyx_vtabptr_6spnego_9_sspi_raw_4sspi__AuthIdentityBase = &__pyx_vtable_6spnego_9_sspi_raw_4sspi__AuthIdentityBase;
  __pyx_vtable_6spnego_9_sspi_raw_4sspi__AuthIdentityBase.__pyx___c_value__ = (void *(*)(struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *))__pyx_f_6spnego_9_sspi_raw_4sspi_17_AuthIdentityBase___c_value__;
  if (PyType_Ready(&__pyx_type_6spnego_9_sspi_raw_4sspi__AuthIdentityBase) < 0) __PYX_ERR(1, 558, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6spnego_9_sspi_raw_4sspi__AuthIdentityBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6spnego_9_sspi_raw_4sspi__AuthIdentityBase.tp_dictoffset && __pyx_type_6spnego_9_sspi_raw_4sspi__AuthIdentityBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6spnego_9_sspi_raw_4sspi__AuthIdentityBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6spnego_9_sspi_raw_4sspi__AuthIdentityBase.tp_dict, __pyx_vtabptr_6spnego_9_sspi_raw_4sspi__AuthIdentityBase) < 0) __PYX_ERR(1, 558, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AuthIdentityBase, (PyObject *)&__pyx_type_6spnego_9_sspi_raw_4sspi__AuthIdentityBase) < 0) __PYX_ERR(1, 558, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6spnego_9_sspi_raw_4sspi__AuthIdentityBase) < 0) __PYX_ERR(1, 558, __pyx_L1_error)
  __pyx_ptype_6spnego_9_sspi_raw_4sspi__AuthIdentityBase = &__pyx_type_6spnego_9_sspi_raw_4sspi__AuthIdentityBase;
  __pyx_vtabptr_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity = &__pyx_vtable_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity;
  __pyx_vtable_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity.__pyx_base = *__pyx_vtabptr_6spnego_9_sspi_raw_4sspi__AuthIdentityBase;
  __pyx_vtable_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity.__pyx_base.__pyx___c_value__ = (void *(*)(struct __pyx_obj_6spnego_9_sspi_raw_4sspi__AuthIdentityBase *))__pyx_f_6spnego_9_sspi_raw_4sspi_17WinNTAuthIdentity___c_value__;
  __pyx_type_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity.tp_base = __pyx_ptype_6spnego_9_sspi_raw_4sspi__AuthIdentityBase;
  if (PyType_Ready(&__pyx_type_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity) < 0) __PYX_ERR(1, 564, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity.tp_dictoffset && __pyx_type_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity.tp_dict, __pyx_vtabptr_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity) < 0) __PYX_ERR(1, 564, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_WinNTAuthIdentity, (PyObject *)&__pyx_type_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity) < 0) __PYX_ERR(1, 564, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity) < 0) __PYX_ERR(1, 564, __pyx_L1_error)
  __pyx_ptype_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity = &__pyx_type_6spnego_9_sspi_raw_4sspi_WinNTAuthIdentity;
  if (PyType_Ready(&__pyx_type_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__) < 0) __PYX_ERR(1, 433, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__.tp_dictoffset && __pyx_type_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__ = &__pyx_type_6spnego_9_sspi_raw_4sspi___pyx_scope_struct____iter__;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("spnego._sspi_raw.text"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6spnego_9_sspi_raw_4text_WideChar = __Pyx_ImportType(__pyx_t_1, "spnego._sspi_raw.text", "WideChar", sizeof(struct __pyx_obj_6spnego_9_sspi_raw_4text_WideChar), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6spnego_9_sspi_raw_4text_WideChar) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("spnego._sspi_raw.text"); if (!__pyx_t_1) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "u16_to_text", (void (**)(void))&__pyx_f_6spnego_9_sspi_raw_4text_u16_to_text, "PyObject *(LPWSTR, size_t)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initsspi(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initsspi(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_sspi(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_sspi(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_sspi(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'sspi' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_sspi(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("sspi", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_spnego___sspi_raw__sspi) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(1, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "spnego._sspi_raw.sspi")) {
      if (unlikely(PyDict_SetItemString(modules, "spnego._sspi_raw.sspi", __pyx_m) < 0)) __PYX_ERR(1, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(1, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(1, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif

  /* "spnego/_sspi_raw/sspi.pyx":5
 * 
 * from collections import (
 *     namedtuple,             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_namedtuple);
  __Pyx_GIVEREF(__pyx_n_s_namedtuple);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_namedtuple);

  /* "spnego/_sspi_raw/sspi.pyx":4
 * # MIT License (see LICENSE or https://opensource.org/licenses/MIT)
 * 
 * from collections import (             # <<<<<<<<<<<<<<
 *     namedtuple,
 * )
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_namedtuple, __pyx_t_1) < 0) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":9
 * 
 * from spnego._text import (
 *     to_native,             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_to_native);
  __Pyx_GIVEREF(__pyx_n_s_to_native);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_to_native);

  /* "spnego/_sspi_raw/sspi.pyx":8
 * )
 * 
 * from spnego._text import (             # <<<<<<<<<<<<<<
 *     to_native,
 * )
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_spnego__text, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_to_native); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_to_native, __pyx_t_2) < 0) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":207
 * 
 * 
 * class ClientContextAttr:             # <<<<<<<<<<<<<<
 *     delegate = ISC_RET_DELEGATE
 *     mutual_auth = ISC_RET_MUTUAL_AUTH
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_ClientContextAttr, __pyx_n_s_ClientContextAttr, (PyObject *) NULL, __pyx_n_s_spnego__sspi_raw_sspi, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "spnego/_sspi_raw/sspi.pyx":208
 * 
 * class ClientContextAttr:
 *     delegate = ISC_RET_DELEGATE             # <<<<<<<<<<<<<<
 *     mutual_auth = ISC_RET_MUTUAL_AUTH
 *     replay_detect = ISC_RET_REPLAY_DETECT
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_DELEGATE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_delegate, __pyx_t_2) < 0) __PYX_ERR(1, 208, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":209
 * class ClientContextAttr:
 *     delegate = ISC_RET_DELEGATE
 *     mutual_auth = ISC_RET_MUTUAL_AUTH             # <<<<<<<<<<<<<<
 *     replay_detect = ISC_RET_REPLAY_DETECT
 *     sequence_detect = ISC_RET_SEQUENCE_DETECT
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_MUTUAL_AUTH); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_mutual_auth, __pyx_t_2) < 0) __PYX_ERR(1, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":210
 *     delegate = ISC_RET_DELEGATE
 *     mutual_auth = ISC_RET_MUTUAL_AUTH
 *     replay_detect = ISC_RET_REPLAY_DETECT             # <<<<<<<<<<<<<<
 *     sequence_detect = ISC_RET_SEQUENCE_DETECT
 *     confidentiality = ISC_RET_CONFIDENTIALITY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_REPLAY_DETECT); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_replay_detect, __pyx_t_2) < 0) __PYX_ERR(1, 210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":211
 *     mutual_auth = ISC_RET_MUTUAL_AUTH
 *     replay_detect = ISC_RET_REPLAY_DETECT
 *     sequence_detect = ISC_RET_SEQUENCE_DETECT             # <<<<<<<<<<<<<<
 *     confidentiality = ISC_RET_CONFIDENTIALITY
 *     use_session_key = ISC_RET_USE_SESSION_KEY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_SEQUENCE_DETECT); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_sequence_detect, __pyx_t_2) < 0) __PYX_ERR(1, 211, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":212
 *     replay_detect = ISC_RET_REPLAY_DETECT
 *     sequence_detect = ISC_RET_SEQUENCE_DETECT
 *     confidentiality = ISC_RET_CONFIDENTIALITY             # <<<<<<<<<<<<<<
 *     use_session_key = ISC_RET_USE_SESSION_KEY
 *     used_collected_creds = ISC_RET_USED_COLLECTED_CREDS
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_CONFIDENTIALITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_confidentiality, __pyx_t_2) < 0) __PYX_ERR(1, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":213
 *     sequence_detect = ISC_RET_SEQUENCE_DETECT
 *     confidentiality = ISC_RET_CONFIDENTIALITY
 *     use_session_key = ISC_RET_USE_SESSION_KEY             # <<<<<<<<<<<<<<
 *     used_collected_creds = ISC_RET_USED_COLLECTED_CREDS
 *     used_supplied_creds = ISC_RET_USED_SUPPLIED_CREDS
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_USE_SESSION_KEY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_use_session_key, __pyx_t_2) < 0) __PYX_ERR(1, 213, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":214
 *     confidentiality = ISC_RET_CONFIDENTIALITY
 *     use_session_key = ISC_RET_USE_SESSION_KEY
 *     used_collected_creds = ISC_RET_USED_COLLECTED_CREDS             # <<<<<<<<<<<<<<
 *     used_supplied_creds = ISC_RET_USED_SUPPLIED_CREDS
 *     allocated_memory = ISC_RET_ALLOCATED_MEMORY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_USED_COLLECTED_CREDS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_used_collected_creds, __pyx_t_2) < 0) __PYX_ERR(1, 214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":215
 *     use_session_key = ISC_RET_USE_SESSION_KEY
 *     used_collected_creds = ISC_RET_USED_COLLECTED_CREDS
 *     used_supplied_creds = ISC_RET_USED_SUPPLIED_CREDS             # <<<<<<<<<<<<<<
 *     allocated_memory = ISC_RET_ALLOCATED_MEMORY
 *     used_dce_style = ISC_RET_USED_DCE_STYLE
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_USED_SUPPLIED_CREDS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_used_supplied_creds, __pyx_t_2) < 0) __PYX_ERR(1, 215, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":216
 *     used_collected_creds = ISC_RET_USED_COLLECTED_CREDS
 *     used_supplied_creds = ISC_RET_USED_SUPPLIED_CREDS
 *     allocated_memory = ISC_RET_ALLOCATED_MEMORY             # <<<<<<<<<<<<<<
 *     used_dce_style = ISC_RET_USED_DCE_STYLE
 *     datagram = ISC_RET_DATAGRAM
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_ALLOCATED_MEMORY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_allocated_memory, __pyx_t_2) < 0) __PYX_ERR(1, 216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":217
 *     used_supplied_creds = ISC_RET_USED_SUPPLIED_CREDS
 *     allocated_memory = ISC_RET_ALLOCATED_MEMORY
 *     used_dce_style = ISC_RET_USED_DCE_STYLE             # <<<<<<<<<<<<<<
 *     datagram = ISC_RET_DATAGRAM
 *     connection = ISC_RET_CONNECTION
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_USED_DCE_STYLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_used_dce_style, __pyx_t_2) < 0) __PYX_ERR(1, 217, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":218
 *     allocated_memory = ISC_RET_ALLOCATED_MEMORY
 *     used_dce_style = ISC_RET_USED_DCE_STYLE
 *     datagram = ISC_RET_DATAGRAM             # <<<<<<<<<<<<<<
 *     connection = ISC_RET_CONNECTION
 *     intermediate_return = ISC_RET_INTERMEDIATE_RETURN
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_DATAGRAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_datagram, __pyx_t_2) < 0) __PYX_ERR(1, 218, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":219
 *     used_dce_style = ISC_RET_USED_DCE_STYLE
 *     datagram = ISC_RET_DATAGRAM
 *     connection = ISC_RET_CONNECTION             # <<<<<<<<<<<<<<
 *     intermediate_return = ISC_RET_INTERMEDIATE_RETURN
 *     call_level = ISC_RET_CALL_LEVEL
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_CONNECTION); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_connection, __pyx_t_2) < 0) __PYX_ERR(1, 219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":220
 *     datagram = ISC_RET_DATAGRAM
 *     connection = ISC_RET_CONNECTION
 *     intermediate_return = ISC_RET_INTERMEDIATE_RETURN             # <<<<<<<<<<<<<<
 *     call_level = ISC_RET_CALL_LEVEL
 *     extended_error = ISC_RET_EXTENDED_ERROR
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_INTERMEDIATE_RETURN); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_intermediate_return, __pyx_t_2) < 0) __PYX_ERR(1, 220, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":221
 *     connection = ISC_RET_CONNECTION
 *     intermediate_return = ISC_RET_INTERMEDIATE_RETURN
 *     call_level = ISC_RET_CALL_LEVEL             # <<<<<<<<<<<<<<
 *     extended_error = ISC_RET_EXTENDED_ERROR
 *     stream = ISC_RET_STREAM
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_CALL_LEVEL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_call_level, __pyx_t_2) < 0) __PYX_ERR(1, 221, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":222
 *     intermediate_return = ISC_RET_INTERMEDIATE_RETURN
 *     call_level = ISC_RET_CALL_LEVEL
 *     extended_error = ISC_RET_EXTENDED_ERROR             # <<<<<<<<<<<<<<
 *     stream = ISC_RET_STREAM
 *     integrity = ISC_RET_INTEGRITY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_EXTENDED_ERROR); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_extended_error, __pyx_t_2) < 0) __PYX_ERR(1, 222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":223
 *     call_level = ISC_RET_CALL_LEVEL
 *     extended_error = ISC_RET_EXTENDED_ERROR
 *     stream = ISC_RET_STREAM             # <<<<<<<<<<<<<<
 *     integrity = ISC_RET_INTEGRITY
 *     identify = ISC_RET_IDENTIFY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_STREAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_stream, __pyx_t_2) < 0) __PYX_ERR(1, 223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":224
 *     extended_error = ISC_RET_EXTENDED_ERROR
 *     stream = ISC_RET_STREAM
 *     integrity = ISC_RET_INTEGRITY             # <<<<<<<<<<<<<<
 *     identify = ISC_RET_IDENTIFY
 *     null_session = ISC_RET_NULL_SESSION
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_INTEGRITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_integrity, __pyx_t_2) < 0) __PYX_ERR(1, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":225
 *     stream = ISC_RET_STREAM
 *     integrity = ISC_RET_INTEGRITY
 *     identify = ISC_RET_IDENTIFY             # <<<<<<<<<<<<<<
 *     null_session = ISC_RET_NULL_SESSION
 *     manual_cred_validation = ISC_RET_MANUAL_CRED_VALIDATION
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_IDENTIFY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_identify, __pyx_t_2) < 0) __PYX_ERR(1, 225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":226
 *     integrity = ISC_RET_INTEGRITY
 *     identify = ISC_RET_IDENTIFY
 *     null_session = ISC_RET_NULL_SESSION             # <<<<<<<<<<<<<<
 *     manual_cred_validation = ISC_RET_MANUAL_CRED_VALIDATION
 *     fragment_only = ISC_RET_FRAGMENT_ONLY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_NULL_SESSION); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_null_session, __pyx_t_2) < 0) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":227
 *     identify = ISC_RET_IDENTIFY
 *     null_session = ISC_RET_NULL_SESSION
 *     manual_cred_validation = ISC_RET_MANUAL_CRED_VALIDATION             # <<<<<<<<<<<<<<
 *     fragment_only = ISC_RET_FRAGMENT_ONLY
 *     forward_credentials = ISC_RET_FORWARD_CREDENTIALS
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_MANUAL_CRED_VALIDATION); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_manual_cred_validation, __pyx_t_2) < 0) __PYX_ERR(1, 227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":228
 *     null_session = ISC_RET_NULL_SESSION
 *     manual_cred_validation = ISC_RET_MANUAL_CRED_VALIDATION
 *     fragment_only = ISC_RET_FRAGMENT_ONLY             # <<<<<<<<<<<<<<
 *     forward_credentials = ISC_RET_FORWARD_CREDENTIALS
 *     used_http_style = ISC_RET_USED_HTTP_STYLE
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_FRAGMENT_ONLY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_fragment_only, __pyx_t_2) < 0) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":229
 *     manual_cred_validation = ISC_RET_MANUAL_CRED_VALIDATION
 *     fragment_only = ISC_RET_FRAGMENT_ONLY
 *     forward_credentials = ISC_RET_FORWARD_CREDENTIALS             # <<<<<<<<<<<<<<
 *     used_http_style = ISC_RET_USED_HTTP_STYLE
 *     no_additional_token = ISC_RET_NO_ADDITIONAL_TOKEN
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_FORWARD_CREDENTIALS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_forward_credentials, __pyx_t_2) < 0) __PYX_ERR(1, 229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":230
 *     fragment_only = ISC_RET_FRAGMENT_ONLY
 *     forward_credentials = ISC_RET_FORWARD_CREDENTIALS
 *     used_http_style = ISC_RET_USED_HTTP_STYLE             # <<<<<<<<<<<<<<
 *     no_additional_token = ISC_RET_NO_ADDITIONAL_TOKEN
 *     reauthentication = ISC_RET_REAUTHENTICATION
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_USED_HTTP_STYLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_used_http_style, __pyx_t_2) < 0) __PYX_ERR(1, 230, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":231
 *     forward_credentials = ISC_RET_FORWARD_CREDENTIALS
 *     used_http_style = ISC_RET_USED_HTTP_STYLE
 *     no_additional_token = ISC_RET_NO_ADDITIONAL_TOKEN             # <<<<<<<<<<<<<<
 *     reauthentication = ISC_RET_REAUTHENTICATION
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_NO_ADDITIONAL_TOKEN); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_no_additional_token, __pyx_t_2) < 0) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":232
 *     used_http_style = ISC_RET_USED_HTTP_STYLE
 *     no_additional_token = ISC_RET_NO_ADDITIONAL_TOKEN
 *     reauthentication = ISC_RET_REAUTHENTICATION             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_RET_REAUTHENTICATION); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_reauthentication, __pyx_t_2) < 0) __PYX_ERR(1, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":207
 * 
 * 
 * class ClientContextAttr:             # <<<<<<<<<<<<<<
 *     delegate = ISC_RET_DELEGATE
 *     mutual_auth = ISC_RET_MUTUAL_AUTH
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_ClientContextAttr, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ClientContextAttr, __pyx_t_2) < 0) __PYX_ERR(1, 207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":235
 * 
 * 
 * class ClientContextReq:             # <<<<<<<<<<<<<<
 *     allocate_memory = ISC_REQ_ALLOCATE_MEMORY
 *     confidentiality = ISC_REQ_CONFIDENTIALITY
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_ClientContextReq, __pyx_n_s_ClientContextReq, (PyObject *) NULL, __pyx_n_s_spnego__sspi_raw_sspi, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "spnego/_sspi_raw/sspi.pyx":236
 * 
 * class ClientContextReq:
 *     allocate_memory = ISC_REQ_ALLOCATE_MEMORY             # <<<<<<<<<<<<<<
 *     confidentiality = ISC_REQ_CONFIDENTIALITY
 *     connection = ISC_REQ_CONNECTION
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_ALLOCATE_MEMORY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_allocate_memory, __pyx_t_2) < 0) __PYX_ERR(1, 236, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":237
 * class ClientContextReq:
 *     allocate_memory = ISC_REQ_ALLOCATE_MEMORY
 *     confidentiality = ISC_REQ_CONFIDENTIALITY             # <<<<<<<<<<<<<<
 *     connection = ISC_REQ_CONNECTION
 *     delegate = ISC_REQ_DELEGATE
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_CONFIDENTIALITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_confidentiality, __pyx_t_2) < 0) __PYX_ERR(1, 237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":238
 *     allocate_memory = ISC_REQ_ALLOCATE_MEMORY
 *     confidentiality = ISC_REQ_CONFIDENTIALITY
 *     connection = ISC_REQ_CONNECTION             # <<<<<<<<<<<<<<
 *     delegate = ISC_REQ_DELEGATE
 *     extended_error = ISC_REQ_EXTENDED_ERROR
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_CONNECTION); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_connection, __pyx_t_2) < 0) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":239
 *     confidentiality = ISC_REQ_CONFIDENTIALITY
 *     connection = ISC_REQ_CONNECTION
 *     delegate = ISC_REQ_DELEGATE             # <<<<<<<<<<<<<<
 *     extended_error = ISC_REQ_EXTENDED_ERROR
 *     http = 0x10000000  # ISC_REQ_HTTP
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_DELEGATE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_delegate, __pyx_t_2) < 0) __PYX_ERR(1, 239, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":240
 *     connection = ISC_REQ_CONNECTION
 *     delegate = ISC_REQ_DELEGATE
 *     extended_error = ISC_REQ_EXTENDED_ERROR             # <<<<<<<<<<<<<<
 *     http = 0x10000000  # ISC_REQ_HTTP
 *     integrity = ISC_REQ_INTEGRITY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_EXTENDED_ERROR); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_extended_error, __pyx_t_2) < 0) __PYX_ERR(1, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":241
 *     delegate = ISC_REQ_DELEGATE
 *     extended_error = ISC_REQ_EXTENDED_ERROR
 *     http = 0x10000000  # ISC_REQ_HTTP             # <<<<<<<<<<<<<<
 *     integrity = ISC_REQ_INTEGRITY
 *     identify = ISC_REQ_IDENTIFY
 */
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_http, __pyx_int_268435456) < 0) __PYX_ERR(1, 241, __pyx_L1_error)

  /* "spnego/_sspi_raw/sspi.pyx":242
 *     extended_error = ISC_REQ_EXTENDED_ERROR
 *     http = 0x10000000  # ISC_REQ_HTTP
 *     integrity = ISC_REQ_INTEGRITY             # <<<<<<<<<<<<<<
 *     identify = ISC_REQ_IDENTIFY
 *     manual_cred_validation = ISC_REQ_MANUAL_CRED_VALIDATION
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_INTEGRITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_integrity, __pyx_t_2) < 0) __PYX_ERR(1, 242, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":243
 *     http = 0x10000000  # ISC_REQ_HTTP
 *     integrity = ISC_REQ_INTEGRITY
 *     identify = ISC_REQ_IDENTIFY             # <<<<<<<<<<<<<<
 *     manual_cred_validation = ISC_REQ_MANUAL_CRED_VALIDATION
 *     mutual_auth = ISC_REQ_MUTUAL_AUTH
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_IDENTIFY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_identify, __pyx_t_2) < 0) __PYX_ERR(1, 243, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":244
 *     integrity = ISC_REQ_INTEGRITY
 *     identify = ISC_REQ_IDENTIFY
 *     manual_cred_validation = ISC_REQ_MANUAL_CRED_VALIDATION             # <<<<<<<<<<<<<<
 *     mutual_auth = ISC_REQ_MUTUAL_AUTH
 *     no_integrity = ISC_REQ_NO_INTEGRITY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_MANUAL_CRED_VALIDATION); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_manual_cred_validation, __pyx_t_2) < 0) __PYX_ERR(1, 244, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":245
 *     identify = ISC_REQ_IDENTIFY
 *     manual_cred_validation = ISC_REQ_MANUAL_CRED_VALIDATION
 *     mutual_auth = ISC_REQ_MUTUAL_AUTH             # <<<<<<<<<<<<<<
 *     no_integrity = ISC_REQ_NO_INTEGRITY
 *     replay_detect = ISC_REQ_REPLAY_DETECT
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_MUTUAL_AUTH); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_mutual_auth, __pyx_t_2) < 0) __PYX_ERR(1, 245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":246
 *     manual_cred_validation = ISC_REQ_MANUAL_CRED_VALIDATION
 *     mutual_auth = ISC_REQ_MUTUAL_AUTH
 *     no_integrity = ISC_REQ_NO_INTEGRITY             # <<<<<<<<<<<<<<
 *     replay_detect = ISC_REQ_REPLAY_DETECT
 *     sequence_detect = ISC_REQ_SEQUENCE_DETECT
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_NO_INTEGRITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_no_integrity, __pyx_t_2) < 0) __PYX_ERR(1, 246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":247
 *     mutual_auth = ISC_REQ_MUTUAL_AUTH
 *     no_integrity = ISC_REQ_NO_INTEGRITY
 *     replay_detect = ISC_REQ_REPLAY_DETECT             # <<<<<<<<<<<<<<
 *     sequence_detect = ISC_REQ_SEQUENCE_DETECT
 *     stream = ISC_REQ_STREAM
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_REPLAY_DETECT); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_replay_detect, __pyx_t_2) < 0) __PYX_ERR(1, 247, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":248
 *     no_integrity = ISC_REQ_NO_INTEGRITY
 *     replay_detect = ISC_REQ_REPLAY_DETECT
 *     sequence_detect = ISC_REQ_SEQUENCE_DETECT             # <<<<<<<<<<<<<<
 *     stream = ISC_REQ_STREAM
 *     use_session_key = ISC_REQ_USE_SESSION_KEY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_SEQUENCE_DETECT); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_sequence_detect, __pyx_t_2) < 0) __PYX_ERR(1, 248, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":249
 *     replay_detect = ISC_REQ_REPLAY_DETECT
 *     sequence_detect = ISC_REQ_SEQUENCE_DETECT
 *     stream = ISC_REQ_STREAM             # <<<<<<<<<<<<<<
 *     use_session_key = ISC_REQ_USE_SESSION_KEY
 *     use_supplied_creds = ISC_REQ_USE_SUPPLIED_CREDS
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_STREAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_stream, __pyx_t_2) < 0) __PYX_ERR(1, 249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":250
 *     sequence_detect = ISC_REQ_SEQUENCE_DETECT
 *     stream = ISC_REQ_STREAM
 *     use_session_key = ISC_REQ_USE_SESSION_KEY             # <<<<<<<<<<<<<<
 *     use_supplied_creds = ISC_REQ_USE_SUPPLIED_CREDS
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_USE_SESSION_KEY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_use_session_key, __pyx_t_2) < 0) __PYX_ERR(1, 250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":251
 *     stream = ISC_REQ_STREAM
 *     use_session_key = ISC_REQ_USE_SESSION_KEY
 *     use_supplied_creds = ISC_REQ_USE_SUPPLIED_CREDS             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ISC_REQ_USE_SUPPLIED_CREDS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_use_supplied_creds, __pyx_t_2) < 0) __PYX_ERR(1, 251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":235
 * 
 * 
 * class ClientContextReq:             # <<<<<<<<<<<<<<
 *     allocate_memory = ISC_REQ_ALLOCATE_MEMORY
 *     confidentiality = ISC_REQ_CONFIDENTIALITY
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_ClientContextReq, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ClientContextReq, __pyx_t_2) < 0) __PYX_ERR(1, 235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":254
 * 
 * 
 * class CredentialUse:             # <<<<<<<<<<<<<<
 *     inbound = SECPKG_CRED_INBOUND
 *     outbound = SECPKG_CRED_OUTBOUND
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_CredentialUse, __pyx_n_s_CredentialUse, (PyObject *) NULL, __pyx_n_s_spnego__sspi_raw_sspi, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "spnego/_sspi_raw/sspi.pyx":255
 * 
 * class CredentialUse:
 *     inbound = SECPKG_CRED_INBOUND             # <<<<<<<<<<<<<<
 *     outbound = SECPKG_CRED_OUTBOUND
 *     both = SECPKG_CRED_BOTH
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECPKG_CRED_INBOUND); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_inbound, __pyx_t_2) < 0) __PYX_ERR(1, 255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":256
 * class CredentialUse:
 *     inbound = SECPKG_CRED_INBOUND
 *     outbound = SECPKG_CRED_OUTBOUND             # <<<<<<<<<<<<<<
 *     both = SECPKG_CRED_BOTH
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECPKG_CRED_OUTBOUND); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_outbound, __pyx_t_2) < 0) __PYX_ERR(1, 256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":257
 *     inbound = SECPKG_CRED_INBOUND
 *     outbound = SECPKG_CRED_OUTBOUND
 *     both = SECPKG_CRED_BOTH             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECPKG_CRED_BOTH); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_both, __pyx_t_2) < 0) __PYX_ERR(1, 257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":254
 * 
 * 
 * class CredentialUse:             # <<<<<<<<<<<<<<
 *     inbound = SECPKG_CRED_INBOUND
 *     outbound = SECPKG_CRED_OUTBOUND
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_CredentialUse, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CredentialUse, __pyx_t_2) < 0) __PYX_ERR(1, 254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":260
 * 
 * 
 * class SecBufferType:             # <<<<<<<<<<<<<<
 *     empty = SECBUFFER_EMPTY
 *     data = SECBUFFER_DATA
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_SecBufferType, __pyx_n_s_SecBufferType, (PyObject *) NULL, __pyx_n_s_spnego__sspi_raw_sspi, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "spnego/_sspi_raw/sspi.pyx":261
 * 
 * class SecBufferType:
 *     empty = SECBUFFER_EMPTY             # <<<<<<<<<<<<<<
 *     data = SECBUFFER_DATA
 *     token = SECBUFFER_TOKEN
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_EMPTY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_empty, __pyx_t_2) < 0) __PYX_ERR(1, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":262
 * class SecBufferType:
 *     empty = SECBUFFER_EMPTY
 *     data = SECBUFFER_DATA             # <<<<<<<<<<<<<<
 *     token = SECBUFFER_TOKEN
 *     pkg_params = SECBUFFER_PKG_PARAMS
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_DATA); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_data, __pyx_t_2) < 0) __PYX_ERR(1, 262, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":263
 *     empty = SECBUFFER_EMPTY
 *     data = SECBUFFER_DATA
 *     token = SECBUFFER_TOKEN             # <<<<<<<<<<<<<<
 *     pkg_params = SECBUFFER_PKG_PARAMS
 *     missing = SECBUFFER_MISSING
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_TOKEN); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_token, __pyx_t_2) < 0) __PYX_ERR(1, 263, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":264
 *     data = SECBUFFER_DATA
 *     token = SECBUFFER_TOKEN
 *     pkg_params = SECBUFFER_PKG_PARAMS             # <<<<<<<<<<<<<<
 *     missing = SECBUFFER_MISSING
 *     extra = SECBUFFER_EXTRA
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_PKG_PARAMS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_pkg_params, __pyx_t_2) < 0) __PYX_ERR(1, 264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":265
 *     token = SECBUFFER_TOKEN
 *     pkg_params = SECBUFFER_PKG_PARAMS
 *     missing = SECBUFFER_MISSING             # <<<<<<<<<<<<<<
 *     extra = SECBUFFER_EXTRA
 *     stream_trailer = SECBUFFER_STREAM_TRAILER
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_MISSING); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_missing, __pyx_t_2) < 0) __PYX_ERR(1, 265, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":266
 *     pkg_params = SECBUFFER_PKG_PARAMS
 *     missing = SECBUFFER_MISSING
 *     extra = SECBUFFER_EXTRA             # <<<<<<<<<<<<<<
 *     stream_trailer = SECBUFFER_STREAM_TRAILER
 *     stream_header = SECBUFFER_STREAM_HEADER
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_EXTRA); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_extra, __pyx_t_2) < 0) __PYX_ERR(1, 266, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":267
 *     missing = SECBUFFER_MISSING
 *     extra = SECBUFFER_EXTRA
 *     stream_trailer = SECBUFFER_STREAM_TRAILER             # <<<<<<<<<<<<<<
 *     stream_header = SECBUFFER_STREAM_HEADER
 *     negotiation_info = SECBUFFER_NEGOTIATION_INFO
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_STREAM_TRAILER); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_stream_trailer, __pyx_t_2) < 0) __PYX_ERR(1, 267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":268
 *     extra = SECBUFFER_EXTRA
 *     stream_trailer = SECBUFFER_STREAM_TRAILER
 *     stream_header = SECBUFFER_STREAM_HEADER             # <<<<<<<<<<<<<<
 *     negotiation_info = SECBUFFER_NEGOTIATION_INFO
 *     padding = SECBUFFER_PADDING
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_STREAM_HEADER); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_stream_header, __pyx_t_2) < 0) __PYX_ERR(1, 268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":269
 *     stream_trailer = SECBUFFER_STREAM_TRAILER
 *     stream_header = SECBUFFER_STREAM_HEADER
 *     negotiation_info = SECBUFFER_NEGOTIATION_INFO             # <<<<<<<<<<<<<<
 *     padding = SECBUFFER_PADDING
 *     stream = SECBUFFER_STREAM
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_NEGOTIATION_INFO); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_negotiation_info, __pyx_t_2) < 0) __PYX_ERR(1, 269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":270
 *     stream_header = SECBUFFER_STREAM_HEADER
 *     negotiation_info = SECBUFFER_NEGOTIATION_INFO
 *     padding = SECBUFFER_PADDING             # <<<<<<<<<<<<<<
 *     stream = SECBUFFER_STREAM
 *     mechlist = SECBUFFER_MECHLIST
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_PADDING); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_padding, __pyx_t_2) < 0) __PYX_ERR(1, 270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":271
 *     negotiation_info = SECBUFFER_NEGOTIATION_INFO
 *     padding = SECBUFFER_PADDING
 *     stream = SECBUFFER_STREAM             # <<<<<<<<<<<<<<
 *     mechlist = SECBUFFER_MECHLIST
 *     mechlist_signature = SECBUFFER_MECHLIST_SIGNATURE
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_STREAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_stream, __pyx_t_2) < 0) __PYX_ERR(1, 271, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":272
 *     padding = SECBUFFER_PADDING
 *     stream = SECBUFFER_STREAM
 *     mechlist = SECBUFFER_MECHLIST             # <<<<<<<<<<<<<<
 *     mechlist_signature = SECBUFFER_MECHLIST_SIGNATURE
 *     target = SECBUFFER_TARGET
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_MECHLIST); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_mechlist, __pyx_t_2) < 0) __PYX_ERR(1, 272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":273
 *     stream = SECBUFFER_STREAM
 *     mechlist = SECBUFFER_MECHLIST
 *     mechlist_signature = SECBUFFER_MECHLIST_SIGNATURE             # <<<<<<<<<<<<<<
 *     target = SECBUFFER_TARGET
 *     channel_bindings = SECBUFFER_CHANNEL_BINDINGS
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_MECHLIST_SIGNATURE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_mechlist_signature, __pyx_t_2) < 0) __PYX_ERR(1, 273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":274
 *     mechlist = SECBUFFER_MECHLIST
 *     mechlist_signature = SECBUFFER_MECHLIST_SIGNATURE
 *     target = SECBUFFER_TARGET             # <<<<<<<<<<<<<<
 *     channel_bindings = SECBUFFER_CHANNEL_BINDINGS
 *     change_pass_response = SECBUFFER_CHANGE_PASS_RESPONSE
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_TARGET); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_target, __pyx_t_2) < 0) __PYX_ERR(1, 274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":275
 *     mechlist_signature = SECBUFFER_MECHLIST_SIGNATURE
 *     target = SECBUFFER_TARGET
 *     channel_bindings = SECBUFFER_CHANNEL_BINDINGS             # <<<<<<<<<<<<<<
 *     change_pass_response = SECBUFFER_CHANGE_PASS_RESPONSE
 *     target_host = SECBUFFER_TARGET_HOST
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_CHANNEL_BINDINGS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_channel_bindings, __pyx_t_2) < 0) __PYX_ERR(1, 275, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":276
 *     target = SECBUFFER_TARGET
 *     channel_bindings = SECBUFFER_CHANNEL_BINDINGS
 *     change_pass_response = SECBUFFER_CHANGE_PASS_RESPONSE             # <<<<<<<<<<<<<<
 *     target_host = SECBUFFER_TARGET_HOST
 *     attrmask = SECBUFFER_ATTRMASK
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_CHANGE_PASS_RESPONSE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_change_pass_response, __pyx_t_2) < 0) __PYX_ERR(1, 276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":277
 *     channel_bindings = SECBUFFER_CHANNEL_BINDINGS
 *     change_pass_response = SECBUFFER_CHANGE_PASS_RESPONSE
 *     target_host = SECBUFFER_TARGET_HOST             # <<<<<<<<<<<<<<
 *     attrmask = SECBUFFER_ATTRMASK
 *     readonly = SECBUFFER_READONLY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_TARGET_HOST); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_target_host, __pyx_t_2) < 0) __PYX_ERR(1, 277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":278
 *     change_pass_response = SECBUFFER_CHANGE_PASS_RESPONSE
 *     target_host = SECBUFFER_TARGET_HOST
 *     attrmask = SECBUFFER_ATTRMASK             # <<<<<<<<<<<<<<
 *     readonly = SECBUFFER_READONLY
 *     readonly_with_checksum = SECBUFFER_READONLY_WITH_CHECKSUM
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_ATTRMASK); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_attrmask, __pyx_t_2) < 0) __PYX_ERR(1, 278, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":279
 *     target_host = SECBUFFER_TARGET_HOST
 *     attrmask = SECBUFFER_ATTRMASK
 *     readonly = SECBUFFER_READONLY             # <<<<<<<<<<<<<<
 *     readonly_with_checksum = SECBUFFER_READONLY_WITH_CHECKSUM
 *     reserved = SECBUFFER_RESERVED
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_READONLY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_readonly, __pyx_t_2) < 0) __PYX_ERR(1, 279, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":280
 *     attrmask = SECBUFFER_ATTRMASK
 *     readonly = SECBUFFER_READONLY
 *     readonly_with_checksum = SECBUFFER_READONLY_WITH_CHECKSUM             # <<<<<<<<<<<<<<
 *     reserved = SECBUFFER_RESERVED
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_READONLY_WITH_CHECKSUM); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_readonly_with_checksum, __pyx_t_2) < 0) __PYX_ERR(1, 280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":281
 *     readonly = SECBUFFER_READONLY
 *     readonly_with_checksum = SECBUFFER_READONLY_WITH_CHECKSUM
 *     reserved = SECBUFFER_RESERVED             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECBUFFER_RESERVED); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_reserved, __pyx_t_2) < 0) __PYX_ERR(1, 281, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":260
 * 
 * 
 * class SecBufferType:             # <<<<<<<<<<<<<<
 *     empty = SECBUFFER_EMPTY
 *     data = SECBUFFER_DATA
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_SecBufferType, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SecBufferType, __pyx_t_2) < 0) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":284
 * 
 * 
 * class SecPkgAttr:             # <<<<<<<<<<<<<<
 *     names = SECPKG_ATTR_NAMES
 *     package_info = SECPKG_ATTR_PACKAGE_INFO
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_SecPkgAttr, __pyx_n_s_SecPkgAttr, (PyObject *) NULL, __pyx_n_s_spnego__sspi_raw_sspi, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "spnego/_sspi_raw/sspi.pyx":285
 * 
 * class SecPkgAttr:
 *     names = SECPKG_ATTR_NAMES             # <<<<<<<<<<<<<<
 *     package_info = SECPKG_ATTR_PACKAGE_INFO
 *     session_key = SECPKG_ATTR_SESSION_KEY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECPKG_ATTR_NAMES); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_names, __pyx_t_2) < 0) __PYX_ERR(1, 285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":286
 * class SecPkgAttr:
 *     names = SECPKG_ATTR_NAMES
 *     package_info = SECPKG_ATTR_PACKAGE_INFO             # <<<<<<<<<<<<<<
 *     session_key = SECPKG_ATTR_SESSION_KEY
 *     sizes = SECPKG_ATTR_SIZES
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECPKG_ATTR_PACKAGE_INFO); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_package_info, __pyx_t_2) < 0) __PYX_ERR(1, 286, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":287
 *     names = SECPKG_ATTR_NAMES
 *     package_info = SECPKG_ATTR_PACKAGE_INFO
 *     session_key = SECPKG_ATTR_SESSION_KEY             # <<<<<<<<<<<<<<
 *     sizes = SECPKG_ATTR_SIZES
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECPKG_ATTR_SESSION_KEY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_session_key, __pyx_t_2) < 0) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":288
 *     package_info = SECPKG_ATTR_PACKAGE_INFO
 *     session_key = SECPKG_ATTR_SESSION_KEY
 *     sizes = SECPKG_ATTR_SIZES             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECPKG_ATTR_SIZES); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_sizes, __pyx_t_2) < 0) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":284
 * 
 * 
 * class SecPkgAttr:             # <<<<<<<<<<<<<<
 *     names = SECPKG_ATTR_NAMES
 *     package_info = SECPKG_ATTR_PACKAGE_INFO
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_SecPkgAttr, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SecPkgAttr, __pyx_t_2) < 0) __PYX_ERR(1, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":291
 * 
 * 
 * class SecStatus:             # <<<<<<<<<<<<<<
 *     SEC_E_BUFFER_TOO_SMALL = _SEC_E_BUFFER_TOO_SMALL
 *     SEC_E_CONTEXT_EXPIRED = _SEC_E_CONTEXT_EXPIRED
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_SecStatus, __pyx_n_s_SecStatus, (PyObject *) NULL, __pyx_n_s_spnego__sspi_raw_sspi, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "spnego/_sspi_raw/sspi.pyx":292
 * 
 * class SecStatus:
 *     SEC_E_BUFFER_TOO_SMALL = _SEC_E_BUFFER_TOO_SMALL             # <<<<<<<<<<<<<<
 *     SEC_E_CONTEXT_EXPIRED = _SEC_E_CONTEXT_EXPIRED
 *     SEC_E_CRYPTO_SYSTEM_INVALID = _SEC_E_CRYPTO_SYSTEM_INVALID
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_BUFFER_TOO_SMALL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_BUFFER_TOO_SMALL, __pyx_t_2) < 0) __PYX_ERR(1, 292, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":293
 * class SecStatus:
 *     SEC_E_BUFFER_TOO_SMALL = _SEC_E_BUFFER_TOO_SMALL
 *     SEC_E_CONTEXT_EXPIRED = _SEC_E_CONTEXT_EXPIRED             # <<<<<<<<<<<<<<
 *     SEC_E_CRYPTO_SYSTEM_INVALID = _SEC_E_CRYPTO_SYSTEM_INVALID
 *     SEC_E_INCOMPLETE_MESSAGE = _SEC_E_INCOMPLETE_MESSAGE
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_CONTEXT_EXPIRED); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_CONTEXT_EXPIRED, __pyx_t_2) < 0) __PYX_ERR(1, 293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":294
 *     SEC_E_BUFFER_TOO_SMALL = _SEC_E_BUFFER_TOO_SMALL
 *     SEC_E_CONTEXT_EXPIRED = _SEC_E_CONTEXT_EXPIRED
 *     SEC_E_CRYPTO_SYSTEM_INVALID = _SEC_E_CRYPTO_SYSTEM_INVALID             # <<<<<<<<<<<<<<
 *     SEC_E_INCOMPLETE_MESSAGE = _SEC_E_INCOMPLETE_MESSAGE
 *     SEC_E_INSUFFICIENT_MEMORY = _SEC_E_INSUFFICIENT_MEMORY
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_CRYPTO_SYSTEM_INVALID); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_CRYPTO_SYSTEM_INVALID, __pyx_t_2) < 0) __PYX_ERR(1, 294, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":295
 *     SEC_E_CONTEXT_EXPIRED = _SEC_E_CONTEXT_EXPIRED
 *     SEC_E_CRYPTO_SYSTEM_INVALID = _SEC_E_CRYPTO_SYSTEM_INVALID
 *     SEC_E_INCOMPLETE_MESSAGE = _SEC_E_INCOMPLETE_MESSAGE             # <<<<<<<<<<<<<<
 *     SEC_E_INSUFFICIENT_MEMORY = _SEC_E_INSUFFICIENT_MEMORY
 *     SEC_E_INTERNAL_ERROR = _SEC_E_INTERNAL_ERROR
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_INCOMPLETE_MESSAGE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_INCOMPLETE_MESSAGE, __pyx_t_2) < 0) __PYX_ERR(1, 295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":296
 *     SEC_E_CRYPTO_SYSTEM_INVALID = _SEC_E_CRYPTO_SYSTEM_INVALID
 *     SEC_E_INCOMPLETE_MESSAGE = _SEC_E_INCOMPLETE_MESSAGE
 *     SEC_E_INSUFFICIENT_MEMORY = _SEC_E_INSUFFICIENT_MEMORY             # <<<<<<<<<<<<<<
 *     SEC_E_INTERNAL_ERROR = _SEC_E_INTERNAL_ERROR
 *     SEC_E_INVALID_HANDLE = _SEC_E_INVALID_HANDLE
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_INSUFFICIENT_MEMORY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_INSUFFICIENT_MEMORY, __pyx_t_2) < 0) __PYX_ERR(1, 296, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":297
 *     SEC_E_INCOMPLETE_MESSAGE = _SEC_E_INCOMPLETE_MESSAGE
 *     SEC_E_INSUFFICIENT_MEMORY = _SEC_E_INSUFFICIENT_MEMORY
 *     SEC_E_INTERNAL_ERROR = _SEC_E_INTERNAL_ERROR             # <<<<<<<<<<<<<<
 *     SEC_E_INVALID_HANDLE = _SEC_E_INVALID_HANDLE
 *     SEC_E_INVALID_TOKEN = _SEC_E_INVALID_TOKEN
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_INTERNAL_ERROR); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_INTERNAL_ERROR, __pyx_t_2) < 0) __PYX_ERR(1, 297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":298
 *     SEC_E_INSUFFICIENT_MEMORY = _SEC_E_INSUFFICIENT_MEMORY
 *     SEC_E_INTERNAL_ERROR = _SEC_E_INTERNAL_ERROR
 *     SEC_E_INVALID_HANDLE = _SEC_E_INVALID_HANDLE             # <<<<<<<<<<<<<<
 *     SEC_E_INVALID_TOKEN = _SEC_E_INVALID_TOKEN
 *     SEC_E_LOGON_DENIED = _SEC_E_LOGON_DENIED
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_INVALID_HANDLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_INVALID_HANDLE, __pyx_t_2) < 0) __PYX_ERR(1, 298, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":299
 *     SEC_E_INTERNAL_ERROR = _SEC_E_INTERNAL_ERROR
 *     SEC_E_INVALID_HANDLE = _SEC_E_INVALID_HANDLE
 *     SEC_E_INVALID_TOKEN = _SEC_E_INVALID_TOKEN             # <<<<<<<<<<<<<<
 *     SEC_E_LOGON_DENIED = _SEC_E_LOGON_DENIED
 *     SEC_E_MESSAGE_ALTERED = _SEC_E_MESSAGE_ALTERED
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_INVALID_TOKEN); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_INVALID_TOKEN, __pyx_t_2) < 0) __PYX_ERR(1, 299, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":300
 *     SEC_E_INVALID_HANDLE = _SEC_E_INVALID_HANDLE
 *     SEC_E_INVALID_TOKEN = _SEC_E_INVALID_TOKEN
 *     SEC_E_LOGON_DENIED = _SEC_E_LOGON_DENIED             # <<<<<<<<<<<<<<
 *     SEC_E_MESSAGE_ALTERED = _SEC_E_MESSAGE_ALTERED
 *     SEC_E_NOT_OWNER = _SEC_E_NOT_OWNER
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_LOGON_DENIED); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_LOGON_DENIED, __pyx_t_2) < 0) __PYX_ERR(1, 300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":301
 *     SEC_E_INVALID_TOKEN = _SEC_E_INVALID_TOKEN
 *     SEC_E_LOGON_DENIED = _SEC_E_LOGON_DENIED
 *     SEC_E_MESSAGE_ALTERED = _SEC_E_MESSAGE_ALTERED             # <<<<<<<<<<<<<<
 *     SEC_E_NOT_OWNER = _SEC_E_NOT_OWNER
 *     SEC_E_NO_AUTHENTICATING_AUTHORITY = _SEC_E_NO_AUTHENTICATING_AUTHORITY
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_MESSAGE_ALTERED); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_MESSAGE_ALTERED, __pyx_t_2) < 0) __PYX_ERR(1, 301, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":302
 *     SEC_E_LOGON_DENIED = _SEC_E_LOGON_DENIED
 *     SEC_E_MESSAGE_ALTERED = _SEC_E_MESSAGE_ALTERED
 *     SEC_E_NOT_OWNER = _SEC_E_NOT_OWNER             # <<<<<<<<<<<<<<
 *     SEC_E_NO_AUTHENTICATING_AUTHORITY = _SEC_E_NO_AUTHENTICATING_AUTHORITY
 *     SEC_E_NO_CREDENTIALS = _SEC_E_NO_CREDENTIALS
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_NOT_OWNER); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_NOT_OWNER, __pyx_t_2) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":303
 *     SEC_E_MESSAGE_ALTERED = _SEC_E_MESSAGE_ALTERED
 *     SEC_E_NOT_OWNER = _SEC_E_NOT_OWNER
 *     SEC_E_NO_AUTHENTICATING_AUTHORITY = _SEC_E_NO_AUTHENTICATING_AUTHORITY             # <<<<<<<<<<<<<<
 *     SEC_E_NO_CREDENTIALS = _SEC_E_NO_CREDENTIALS
 *     SEC_E_OK = _SEC_E_OK
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_NO_AUTHENTICATING_AUTHORITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_NO_AUTHENTICATING_AUTHORIT, __pyx_t_2) < 0) __PYX_ERR(1, 303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":304
 *     SEC_E_NOT_OWNER = _SEC_E_NOT_OWNER
 *     SEC_E_NO_AUTHENTICATING_AUTHORITY = _SEC_E_NO_AUTHENTICATING_AUTHORITY
 *     SEC_E_NO_CREDENTIALS = _SEC_E_NO_CREDENTIALS             # <<<<<<<<<<<<<<
 *     SEC_E_OK = _SEC_E_OK
 *     SEC_E_OUT_OF_SEQUENCE = _SEC_E_OUT_OF_SEQUENCE
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_NO_CREDENTIALS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_NO_CREDENTIALS, __pyx_t_2) < 0) __PYX_ERR(1, 304, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":305
 *     SEC_E_NO_AUTHENTICATING_AUTHORITY = _SEC_E_NO_AUTHENTICATING_AUTHORITY
 *     SEC_E_NO_CREDENTIALS = _SEC_E_NO_CREDENTIALS
 *     SEC_E_OK = _SEC_E_OK             # <<<<<<<<<<<<<<
 *     SEC_E_OUT_OF_SEQUENCE = _SEC_E_OUT_OF_SEQUENCE
 *     SEC_E_QOP_NOT_SUPPORTED = _SEC_E_QOP_NOT_SUPPORTED
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_OK); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_OK, __pyx_t_2) < 0) __PYX_ERR(1, 305, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":306
 *     SEC_E_NO_CREDENTIALS = _SEC_E_NO_CREDENTIALS
 *     SEC_E_OK = _SEC_E_OK
 *     SEC_E_OUT_OF_SEQUENCE = _SEC_E_OUT_OF_SEQUENCE             # <<<<<<<<<<<<<<
 *     SEC_E_QOP_NOT_SUPPORTED = _SEC_E_QOP_NOT_SUPPORTED
 *     SEC_E_SECPKG_NOT_FOUND = _SEC_E_SECPKG_NOT_FOUND
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_OUT_OF_SEQUENCE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_OUT_OF_SEQUENCE, __pyx_t_2) < 0) __PYX_ERR(1, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":307
 *     SEC_E_OK = _SEC_E_OK
 *     SEC_E_OUT_OF_SEQUENCE = _SEC_E_OUT_OF_SEQUENCE
 *     SEC_E_QOP_NOT_SUPPORTED = _SEC_E_QOP_NOT_SUPPORTED             # <<<<<<<<<<<<<<
 *     SEC_E_SECPKG_NOT_FOUND = _SEC_E_SECPKG_NOT_FOUND
 *     SEC_E_TARGET_UNKNOWN = _SEC_E_TARGET_UNKNOWN
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_QOP_NOT_SUPPORTED); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_QOP_NOT_SUPPORTED, __pyx_t_2) < 0) __PYX_ERR(1, 307, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":308
 *     SEC_E_OUT_OF_SEQUENCE = _SEC_E_OUT_OF_SEQUENCE
 *     SEC_E_QOP_NOT_SUPPORTED = _SEC_E_QOP_NOT_SUPPORTED
 *     SEC_E_SECPKG_NOT_FOUND = _SEC_E_SECPKG_NOT_FOUND             # <<<<<<<<<<<<<<
 *     SEC_E_TARGET_UNKNOWN = _SEC_E_TARGET_UNKNOWN
 *     SEC_E_UNKNOWN_CREDENTIALS = _SEC_E_UNKNOWN_CREDENTIALS
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_SECPKG_NOT_FOUND); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_SECPKG_NOT_FOUND, __pyx_t_2) < 0) __PYX_ERR(1, 308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":309
 *     SEC_E_QOP_NOT_SUPPORTED = _SEC_E_QOP_NOT_SUPPORTED
 *     SEC_E_SECPKG_NOT_FOUND = _SEC_E_SECPKG_NOT_FOUND
 *     SEC_E_TARGET_UNKNOWN = _SEC_E_TARGET_UNKNOWN             # <<<<<<<<<<<<<<
 *     SEC_E_UNKNOWN_CREDENTIALS = _SEC_E_UNKNOWN_CREDENTIALS
 *     SEC_E_UNSUPPORTED_FUNCTION = _SEC_E_UNSUPPORTED_FUNCTION
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_TARGET_UNKNOWN); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_TARGET_UNKNOWN, __pyx_t_2) < 0) __PYX_ERR(1, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":310
 *     SEC_E_SECPKG_NOT_FOUND = _SEC_E_SECPKG_NOT_FOUND
 *     SEC_E_TARGET_UNKNOWN = _SEC_E_TARGET_UNKNOWN
 *     SEC_E_UNKNOWN_CREDENTIALS = _SEC_E_UNKNOWN_CREDENTIALS             # <<<<<<<<<<<<<<
 *     SEC_E_UNSUPPORTED_FUNCTION = _SEC_E_UNSUPPORTED_FUNCTION
 *     SEC_E_WRONG_PRINCIPAL = _SEC_E_WRONG_PRINCIPAL
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_UNKNOWN_CREDENTIALS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_UNKNOWN_CREDENTIALS, __pyx_t_2) < 0) __PYX_ERR(1, 310, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":311
 *     SEC_E_TARGET_UNKNOWN = _SEC_E_TARGET_UNKNOWN
 *     SEC_E_UNKNOWN_CREDENTIALS = _SEC_E_UNKNOWN_CREDENTIALS
 *     SEC_E_UNSUPPORTED_FUNCTION = _SEC_E_UNSUPPORTED_FUNCTION             # <<<<<<<<<<<<<<
 *     SEC_E_WRONG_PRINCIPAL = _SEC_E_WRONG_PRINCIPAL
 *     SEC_I_COMPLETE_AND_CONTINUE = _SEC_I_COMPLETE_AND_CONTINUE
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_UNSUPPORTED_FUNCTION); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_UNSUPPORTED_FUNCTION, __pyx_t_2) < 0) __PYX_ERR(1, 311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":312
 *     SEC_E_UNKNOWN_CREDENTIALS = _SEC_E_UNKNOWN_CREDENTIALS
 *     SEC_E_UNSUPPORTED_FUNCTION = _SEC_E_UNSUPPORTED_FUNCTION
 *     SEC_E_WRONG_PRINCIPAL = _SEC_E_WRONG_PRINCIPAL             # <<<<<<<<<<<<<<
 *     SEC_I_COMPLETE_AND_CONTINUE = _SEC_I_COMPLETE_AND_CONTINUE
 *     SEC_I_COMPLETE_NEEDED = _SEC_I_COMPLETE_NEEDED
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_E_WRONG_PRINCIPAL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_E_WRONG_PRINCIPAL, __pyx_t_2) < 0) __PYX_ERR(1, 312, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":313
 *     SEC_E_UNSUPPORTED_FUNCTION = _SEC_E_UNSUPPORTED_FUNCTION
 *     SEC_E_WRONG_PRINCIPAL = _SEC_E_WRONG_PRINCIPAL
 *     SEC_I_COMPLETE_AND_CONTINUE = _SEC_I_COMPLETE_AND_CONTINUE             # <<<<<<<<<<<<<<
 *     SEC_I_COMPLETE_NEEDED = _SEC_I_COMPLETE_NEEDED
 *     SEC_I_CONTINUE_NEEDED = _SEC_I_CONTINUE_NEEDED
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_I_COMPLETE_AND_CONTINUE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_I_COMPLETE_AND_CONTINUE, __pyx_t_2) < 0) __PYX_ERR(1, 313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":314
 *     SEC_E_WRONG_PRINCIPAL = _SEC_E_WRONG_PRINCIPAL
 *     SEC_I_COMPLETE_AND_CONTINUE = _SEC_I_COMPLETE_AND_CONTINUE
 *     SEC_I_COMPLETE_NEEDED = _SEC_I_COMPLETE_NEEDED             # <<<<<<<<<<<<<<
 *     SEC_I_CONTINUE_NEEDED = _SEC_I_CONTINUE_NEEDED
 *     SEC_I_INCOMPLETE_CREDENTIALS = _SEC_I_INCOMPLETE_CREDENTIALS
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_I_COMPLETE_NEEDED); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_I_COMPLETE_NEEDED, __pyx_t_2) < 0) __PYX_ERR(1, 314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":315
 *     SEC_I_COMPLETE_AND_CONTINUE = _SEC_I_COMPLETE_AND_CONTINUE
 *     SEC_I_COMPLETE_NEEDED = _SEC_I_COMPLETE_NEEDED
 *     SEC_I_CONTINUE_NEEDED = _SEC_I_CONTINUE_NEEDED             # <<<<<<<<<<<<<<
 *     SEC_I_INCOMPLETE_CREDENTIALS = _SEC_I_INCOMPLETE_CREDENTIALS
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_I_CONTINUE_NEEDED); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_I_CONTINUE_NEEDED, __pyx_t_2) < 0) __PYX_ERR(1, 315, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":316
 *     SEC_I_COMPLETE_NEEDED = _SEC_I_COMPLETE_NEEDED
 *     SEC_I_CONTINUE_NEEDED = _SEC_I_CONTINUE_NEEDED
 *     SEC_I_INCOMPLETE_CREDENTIALS = _SEC_I_INCOMPLETE_CREDENTIALS             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_LONG(SEC_I_INCOMPLETE_CREDENTIALS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_SEC_I_INCOMPLETE_CREDENTIALS, __pyx_t_2) < 0) __PYX_ERR(1, 316, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":291
 * 
 * 
 * class SecStatus:             # <<<<<<<<<<<<<<
 *     SEC_E_BUFFER_TOO_SMALL = _SEC_E_BUFFER_TOO_SMALL
 *     SEC_E_CONTEXT_EXPIRED = _SEC_E_CONTEXT_EXPIRED
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_SecStatus, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SecStatus, __pyx_t_2) < 0) __PYX_ERR(1, 291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":319
 * 
 * 
 * class ServerContextAttr:             # <<<<<<<<<<<<<<
 *     delegate = ASC_RET_DELEGATE
 *     mutual_auth = ASC_RET_MUTUAL_AUTH
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_ServerContextAttr, __pyx_n_s_ServerContextAttr, (PyObject *) NULL, __pyx_n_s_spnego__sspi_raw_sspi, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "spnego/_sspi_raw/sspi.pyx":320
 * 
 * class ServerContextAttr:
 *     delegate = ASC_RET_DELEGATE             # <<<<<<<<<<<<<<
 *     mutual_auth = ASC_RET_MUTUAL_AUTH
 *     replay_detect = ASC_RET_REPLAY_DETECT
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_DELEGATE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_delegate, __pyx_t_2) < 0) __PYX_ERR(1, 320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":321
 * class ServerContextAttr:
 *     delegate = ASC_RET_DELEGATE
 *     mutual_auth = ASC_RET_MUTUAL_AUTH             # <<<<<<<<<<<<<<
 *     replay_detect = ASC_RET_REPLAY_DETECT
 *     sequence_detect = ASC_RET_SEQUENCE_DETECT
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_MUTUAL_AUTH); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_mutual_auth, __pyx_t_2) < 0) __PYX_ERR(1, 321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":322
 *     delegate = ASC_RET_DELEGATE
 *     mutual_auth = ASC_RET_MUTUAL_AUTH
 *     replay_detect = ASC_RET_REPLAY_DETECT             # <<<<<<<<<<<<<<
 *     sequence_detect = ASC_RET_SEQUENCE_DETECT
 *     confidentiality = ASC_RET_CONFIDENTIALITY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_REPLAY_DETECT); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_replay_detect, __pyx_t_2) < 0) __PYX_ERR(1, 322, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":323
 *     mutual_auth = ASC_RET_MUTUAL_AUTH
 *     replay_detect = ASC_RET_REPLAY_DETECT
 *     sequence_detect = ASC_RET_SEQUENCE_DETECT             # <<<<<<<<<<<<<<
 *     confidentiality = ASC_RET_CONFIDENTIALITY
 *     use_session_key = ASC_RET_USE_SESSION_KEY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_SEQUENCE_DETECT); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_sequence_detect, __pyx_t_2) < 0) __PYX_ERR(1, 323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":324
 *     replay_detect = ASC_RET_REPLAY_DETECT
 *     sequence_detect = ASC_RET_SEQUENCE_DETECT
 *     confidentiality = ASC_RET_CONFIDENTIALITY             # <<<<<<<<<<<<<<
 *     use_session_key = ASC_RET_USE_SESSION_KEY
 *     allocated_memory = ASC_RET_ALLOCATED_MEMORY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_CONFIDENTIALITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_confidentiality, __pyx_t_2) < 0) __PYX_ERR(1, 324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":325
 *     sequence_detect = ASC_RET_SEQUENCE_DETECT
 *     confidentiality = ASC_RET_CONFIDENTIALITY
 *     use_session_key = ASC_RET_USE_SESSION_KEY             # <<<<<<<<<<<<<<
 *     allocated_memory = ASC_RET_ALLOCATED_MEMORY
 *     used_dce_style = ASC_RET_USED_DCE_STYLE
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_USE_SESSION_KEY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_use_session_key, __pyx_t_2) < 0) __PYX_ERR(1, 325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":326
 *     confidentiality = ASC_RET_CONFIDENTIALITY
 *     use_session_key = ASC_RET_USE_SESSION_KEY
 *     allocated_memory = ASC_RET_ALLOCATED_MEMORY             # <<<<<<<<<<<<<<
 *     used_dce_style = ASC_RET_USED_DCE_STYLE
 *     datagram = ASC_RET_DATAGRAM
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_ALLOCATED_MEMORY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_allocated_memory, __pyx_t_2) < 0) __PYX_ERR(1, 326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":327
 *     use_session_key = ASC_RET_USE_SESSION_KEY
 *     allocated_memory = ASC_RET_ALLOCATED_MEMORY
 *     used_dce_style = ASC_RET_USED_DCE_STYLE             # <<<<<<<<<<<<<<
 *     datagram = ASC_RET_DATAGRAM
 *     connection = ASC_RET_CONNECTION
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_USED_DCE_STYLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_used_dce_style, __pyx_t_2) < 0) __PYX_ERR(1, 327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":328
 *     allocated_memory = ASC_RET_ALLOCATED_MEMORY
 *     used_dce_style = ASC_RET_USED_DCE_STYLE
 *     datagram = ASC_RET_DATAGRAM             # <<<<<<<<<<<<<<
 *     connection = ASC_RET_CONNECTION
 *     call_level = ASC_RET_CALL_LEVEL
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_DATAGRAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_datagram, __pyx_t_2) < 0) __PYX_ERR(1, 328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":329
 *     used_dce_style = ASC_RET_USED_DCE_STYLE
 *     datagram = ASC_RET_DATAGRAM
 *     connection = ASC_RET_CONNECTION             # <<<<<<<<<<<<<<
 *     call_level = ASC_RET_CALL_LEVEL
 *     third_leg_failed = ASC_RET_THIRD_LEG_FAILED
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_CONNECTION); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_connection, __pyx_t_2) < 0) __PYX_ERR(1, 329, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":330
 *     datagram = ASC_RET_DATAGRAM
 *     connection = ASC_RET_CONNECTION
 *     call_level = ASC_RET_CALL_LEVEL             # <<<<<<<<<<<<<<
 *     third_leg_failed = ASC_RET_THIRD_LEG_FAILED
 *     extended_error = ASC_RET_EXTENDED_ERROR
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_CALL_LEVEL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_call_level, __pyx_t_2) < 0) __PYX_ERR(1, 330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":331
 *     connection = ASC_RET_CONNECTION
 *     call_level = ASC_RET_CALL_LEVEL
 *     third_leg_failed = ASC_RET_THIRD_LEG_FAILED             # <<<<<<<<<<<<<<
 *     extended_error = ASC_RET_EXTENDED_ERROR
 *     stream = ASC_RET_STREAM
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_THIRD_LEG_FAILED); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_third_leg_failed, __pyx_t_2) < 0) __PYX_ERR(1, 331, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":332
 *     call_level = ASC_RET_CALL_LEVEL
 *     third_leg_failed = ASC_RET_THIRD_LEG_FAILED
 *     extended_error = ASC_RET_EXTENDED_ERROR             # <<<<<<<<<<<<<<
 *     stream = ASC_RET_STREAM
 *     integrity = ASC_RET_INTEGRITY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_EXTENDED_ERROR); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_extended_error, __pyx_t_2) < 0) __PYX_ERR(1, 332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":333
 *     third_leg_failed = ASC_RET_THIRD_LEG_FAILED
 *     extended_error = ASC_RET_EXTENDED_ERROR
 *     stream = ASC_RET_STREAM             # <<<<<<<<<<<<<<
 *     integrity = ASC_RET_INTEGRITY
 *     licensing = ASC_RET_LICENSING
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_STREAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_stream, __pyx_t_2) < 0) __PYX_ERR(1, 333, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":334
 *     extended_error = ASC_RET_EXTENDED_ERROR
 *     stream = ASC_RET_STREAM
 *     integrity = ASC_RET_INTEGRITY             # <<<<<<<<<<<<<<
 *     licensing = ASC_RET_LICENSING
 *     identify = ASC_RET_IDENTIFY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_INTEGRITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_integrity, __pyx_t_2) < 0) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":335
 *     stream = ASC_RET_STREAM
 *     integrity = ASC_RET_INTEGRITY
 *     licensing = ASC_RET_LICENSING             # <<<<<<<<<<<<<<
 *     identify = ASC_RET_IDENTIFY
 *     null_session = ASC_RET_NULL_SESSION
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_LICENSING); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_licensing, __pyx_t_2) < 0) __PYX_ERR(1, 335, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":336
 *     integrity = ASC_RET_INTEGRITY
 *     licensing = ASC_RET_LICENSING
 *     identify = ASC_RET_IDENTIFY             # <<<<<<<<<<<<<<
 *     null_session = ASC_RET_NULL_SESSION
 *     allow_non_user_logons = ASC_RET_ALLOW_NON_USER_LOGONS
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_IDENTIFY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_identify, __pyx_t_2) < 0) __PYX_ERR(1, 336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":337
 *     licensing = ASC_RET_LICENSING
 *     identify = ASC_RET_IDENTIFY
 *     null_session = ASC_RET_NULL_SESSION             # <<<<<<<<<<<<<<
 *     allow_non_user_logons = ASC_RET_ALLOW_NON_USER_LOGONS
 *     allow_context_replay = ASC_RET_ALLOW_CONTEXT_REPLAY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_NULL_SESSION); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_null_session, __pyx_t_2) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":338
 *     identify = ASC_RET_IDENTIFY
 *     null_session = ASC_RET_NULL_SESSION
 *     allow_non_user_logons = ASC_RET_ALLOW_NON_USER_LOGONS             # <<<<<<<<<<<<<<
 *     allow_context_replay = ASC_RET_ALLOW_CONTEXT_REPLAY
 *     fragment_only = ASC_RET_FRAGMENT_ONLY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_ALLOW_NON_USER_LOGONS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_allow_non_user_logons, __pyx_t_2) < 0) __PYX_ERR(1, 338, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":339
 *     null_session = ASC_RET_NULL_SESSION
 *     allow_non_user_logons = ASC_RET_ALLOW_NON_USER_LOGONS
 *     allow_context_replay = ASC_RET_ALLOW_CONTEXT_REPLAY             # <<<<<<<<<<<<<<
 *     fragment_only = ASC_RET_FRAGMENT_ONLY
 *     no_token = ASC_RET_NO_TOKEN
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_ALLOW_CONTEXT_REPLAY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_allow_context_replay, __pyx_t_2) < 0) __PYX_ERR(1, 339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":340
 *     allow_non_user_logons = ASC_RET_ALLOW_NON_USER_LOGONS
 *     allow_context_replay = ASC_RET_ALLOW_CONTEXT_REPLAY
 *     fragment_only = ASC_RET_FRAGMENT_ONLY             # <<<<<<<<<<<<<<
 *     no_token = ASC_RET_NO_TOKEN
 *     no_additional_token = ASC_RET_NO_ADDITIONAL_TOKEN
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_FRAGMENT_ONLY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_fragment_only, __pyx_t_2) < 0) __PYX_ERR(1, 340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":341
 *     allow_context_replay = ASC_RET_ALLOW_CONTEXT_REPLAY
 *     fragment_only = ASC_RET_FRAGMENT_ONLY
 *     no_token = ASC_RET_NO_TOKEN             # <<<<<<<<<<<<<<
 *     no_additional_token = ASC_RET_NO_ADDITIONAL_TOKEN
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_NO_TOKEN); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_no_token, __pyx_t_2) < 0) __PYX_ERR(1, 341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":342
 *     fragment_only = ASC_RET_FRAGMENT_ONLY
 *     no_token = ASC_RET_NO_TOKEN
 *     no_additional_token = ASC_RET_NO_ADDITIONAL_TOKEN             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_RET_NO_ADDITIONAL_TOKEN); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_no_additional_token, __pyx_t_2) < 0) __PYX_ERR(1, 342, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":319
 * 
 * 
 * class ServerContextAttr:             # <<<<<<<<<<<<<<
 *     delegate = ASC_RET_DELEGATE
 *     mutual_auth = ASC_RET_MUTUAL_AUTH
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_ServerContextAttr, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ServerContextAttr, __pyx_t_2) < 0) __PYX_ERR(1, 319, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":345
 * 
 * 
 * class ServerContextReq:             # <<<<<<<<<<<<<<
 *     delegate = ASC_REQ_DELEGATE
 *     mutual_auth = ASC_REQ_MUTUAL_AUTH
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_ServerContextReq, __pyx_n_s_ServerContextReq, (PyObject *) NULL, __pyx_n_s_spnego__sspi_raw_sspi, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "spnego/_sspi_raw/sspi.pyx":346
 * 
 * class ServerContextReq:
 *     delegate = ASC_REQ_DELEGATE             # <<<<<<<<<<<<<<
 *     mutual_auth = ASC_REQ_MUTUAL_AUTH
 *     replay_detect = ASC_REQ_REPLAY_DETECT
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_DELEGATE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_delegate, __pyx_t_2) < 0) __PYX_ERR(1, 346, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":347
 * class ServerContextReq:
 *     delegate = ASC_REQ_DELEGATE
 *     mutual_auth = ASC_REQ_MUTUAL_AUTH             # <<<<<<<<<<<<<<
 *     replay_detect = ASC_REQ_REPLAY_DETECT
 *     sequence_detect = ASC_REQ_SEQUENCE_DETECT
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_MUTUAL_AUTH); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_mutual_auth, __pyx_t_2) < 0) __PYX_ERR(1, 347, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":348
 *     delegate = ASC_REQ_DELEGATE
 *     mutual_auth = ASC_REQ_MUTUAL_AUTH
 *     replay_detect = ASC_REQ_REPLAY_DETECT             # <<<<<<<<<<<<<<
 *     sequence_detect = ASC_REQ_SEQUENCE_DETECT
 *     confidentiality = ASC_REQ_CONFIDENTIALITY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_REPLAY_DETECT); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_replay_detect, __pyx_t_2) < 0) __PYX_ERR(1, 348, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":349
 *     mutual_auth = ASC_REQ_MUTUAL_AUTH
 *     replay_detect = ASC_REQ_REPLAY_DETECT
 *     sequence_detect = ASC_REQ_SEQUENCE_DETECT             # <<<<<<<<<<<<<<
 *     confidentiality = ASC_REQ_CONFIDENTIALITY
 *     use_session_key = ASC_REQ_USE_SESSION_KEY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_SEQUENCE_DETECT); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_sequence_detect, __pyx_t_2) < 0) __PYX_ERR(1, 349, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":350
 *     replay_detect = ASC_REQ_REPLAY_DETECT
 *     sequence_detect = ASC_REQ_SEQUENCE_DETECT
 *     confidentiality = ASC_REQ_CONFIDENTIALITY             # <<<<<<<<<<<<<<
 *     use_session_key = ASC_REQ_USE_SESSION_KEY
 *     allocate_memory = ASC_REQ_ALLOCATE_MEMORY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_CONFIDENTIALITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_confidentiality, __pyx_t_2) < 0) __PYX_ERR(1, 350, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":351
 *     sequence_detect = ASC_REQ_SEQUENCE_DETECT
 *     confidentiality = ASC_REQ_CONFIDENTIALITY
 *     use_session_key = ASC_REQ_USE_SESSION_KEY             # <<<<<<<<<<<<<<
 *     allocate_memory = ASC_REQ_ALLOCATE_MEMORY
 *     use_dce_style = ASC_REQ_USE_DCE_STYLE
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_USE_SESSION_KEY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_use_session_key, __pyx_t_2) < 0) __PYX_ERR(1, 351, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":352
 *     confidentiality = ASC_REQ_CONFIDENTIALITY
 *     use_session_key = ASC_REQ_USE_SESSION_KEY
 *     allocate_memory = ASC_REQ_ALLOCATE_MEMORY             # <<<<<<<<<<<<<<
 *     use_dce_style = ASC_REQ_USE_DCE_STYLE
 *     datagram = ASC_REQ_DATAGRAM
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_ALLOCATE_MEMORY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_allocate_memory, __pyx_t_2) < 0) __PYX_ERR(1, 352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":353
 *     use_session_key = ASC_REQ_USE_SESSION_KEY
 *     allocate_memory = ASC_REQ_ALLOCATE_MEMORY
 *     use_dce_style = ASC_REQ_USE_DCE_STYLE             # <<<<<<<<<<<<<<
 *     datagram = ASC_REQ_DATAGRAM
 *     connection = ASC_REQ_CONNECTION
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_USE_DCE_STYLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_use_dce_style, __pyx_t_2) < 0) __PYX_ERR(1, 353, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":354
 *     allocate_memory = ASC_REQ_ALLOCATE_MEMORY
 *     use_dce_style = ASC_REQ_USE_DCE_STYLE
 *     datagram = ASC_REQ_DATAGRAM             # <<<<<<<<<<<<<<
 *     connection = ASC_REQ_CONNECTION
 *     call_level = ASC_REQ_CALL_LEVEL
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_DATAGRAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_datagram, __pyx_t_2) < 0) __PYX_ERR(1, 354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":355
 *     use_dce_style = ASC_REQ_USE_DCE_STYLE
 *     datagram = ASC_REQ_DATAGRAM
 *     connection = ASC_REQ_CONNECTION             # <<<<<<<<<<<<<<
 *     call_level = ASC_REQ_CALL_LEVEL
 *     fragment_supplied = ASC_REQ_FRAGMENT_SUPPLIED
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_CONNECTION); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_connection, __pyx_t_2) < 0) __PYX_ERR(1, 355, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":356
 *     datagram = ASC_REQ_DATAGRAM
 *     connection = ASC_REQ_CONNECTION
 *     call_level = ASC_REQ_CALL_LEVEL             # <<<<<<<<<<<<<<
 *     fragment_supplied = ASC_REQ_FRAGMENT_SUPPLIED
 *     extended_error = ASC_REQ_EXTENDED_ERROR
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_CALL_LEVEL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_call_level, __pyx_t_2) < 0) __PYX_ERR(1, 356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":357
 *     connection = ASC_REQ_CONNECTION
 *     call_level = ASC_REQ_CALL_LEVEL
 *     fragment_supplied = ASC_REQ_FRAGMENT_SUPPLIED             # <<<<<<<<<<<<<<
 *     extended_error = ASC_REQ_EXTENDED_ERROR
 *     stream = ASC_REQ_STREAM
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_FRAGMENT_SUPPLIED); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_fragment_supplied, __pyx_t_2) < 0) __PYX_ERR(1, 357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":358
 *     call_level = ASC_REQ_CALL_LEVEL
 *     fragment_supplied = ASC_REQ_FRAGMENT_SUPPLIED
 *     extended_error = ASC_REQ_EXTENDED_ERROR             # <<<<<<<<<<<<<<
 *     stream = ASC_REQ_STREAM
 *     integrity = ASC_REQ_INTEGRITY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_EXTENDED_ERROR); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_extended_error, __pyx_t_2) < 0) __PYX_ERR(1, 358, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":359
 *     fragment_supplied = ASC_REQ_FRAGMENT_SUPPLIED
 *     extended_error = ASC_REQ_EXTENDED_ERROR
 *     stream = ASC_REQ_STREAM             # <<<<<<<<<<<<<<
 *     integrity = ASC_REQ_INTEGRITY
 *     licensing = ASC_REQ_LICENSING
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_STREAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_stream, __pyx_t_2) < 0) __PYX_ERR(1, 359, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":360
 *     extended_error = ASC_REQ_EXTENDED_ERROR
 *     stream = ASC_REQ_STREAM
 *     integrity = ASC_REQ_INTEGRITY             # <<<<<<<<<<<<<<
 *     licensing = ASC_REQ_LICENSING
 *     identify = ASC_REQ_IDENTIFY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_INTEGRITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_integrity, __pyx_t_2) < 0) __PYX_ERR(1, 360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":361
 *     stream = ASC_REQ_STREAM
 *     integrity = ASC_REQ_INTEGRITY
 *     licensing = ASC_REQ_LICENSING             # <<<<<<<<<<<<<<
 *     identify = ASC_REQ_IDENTIFY
 *     allow_null_session = ASC_REQ_ALLOW_NULL_SESSION
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_LICENSING); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_licensing, __pyx_t_2) < 0) __PYX_ERR(1, 361, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":362
 *     integrity = ASC_REQ_INTEGRITY
 *     licensing = ASC_REQ_LICENSING
 *     identify = ASC_REQ_IDENTIFY             # <<<<<<<<<<<<<<
 *     allow_null_session = ASC_REQ_ALLOW_NULL_SESSION
 *     allow_non_user_logons = ASC_REQ_ALLOW_NON_USER_LOGONS
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_IDENTIFY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_identify, __pyx_t_2) < 0) __PYX_ERR(1, 362, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":363
 *     licensing = ASC_REQ_LICENSING
 *     identify = ASC_REQ_IDENTIFY
 *     allow_null_session = ASC_REQ_ALLOW_NULL_SESSION             # <<<<<<<<<<<<<<
 *     allow_non_user_logons = ASC_REQ_ALLOW_NON_USER_LOGONS
 *     allow_context_replay = ASC_REQ_ALLOW_CONTEXT_REPLAY
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_ALLOW_NULL_SESSION); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_allow_null_session, __pyx_t_2) < 0) __PYX_ERR(1, 363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":364
 *     identify = ASC_REQ_IDENTIFY
 *     allow_null_session = ASC_REQ_ALLOW_NULL_SESSION
 *     allow_non_user_logons = ASC_REQ_ALLOW_NON_USER_LOGONS             # <<<<<<<<<<<<<<
 *     allow_context_replay = ASC_REQ_ALLOW_CONTEXT_REPLAY
 *     fragment_to_fit = ASC_REQ_FRAGMENT_TO_FIT
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_ALLOW_NON_USER_LOGONS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_allow_non_user_logons, __pyx_t_2) < 0) __PYX_ERR(1, 364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":365
 *     allow_null_session = ASC_REQ_ALLOW_NULL_SESSION
 *     allow_non_user_logons = ASC_REQ_ALLOW_NON_USER_LOGONS
 *     allow_context_replay = ASC_REQ_ALLOW_CONTEXT_REPLAY             # <<<<<<<<<<<<<<
 *     fragment_to_fit = ASC_REQ_FRAGMENT_TO_FIT
 *     no_token = ASC_REQ_NO_TOKEN
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_ALLOW_CONTEXT_REPLAY); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_allow_context_replay, __pyx_t_2) < 0) __PYX_ERR(1, 365, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":366
 *     allow_non_user_logons = ASC_REQ_ALLOW_NON_USER_LOGONS
 *     allow_context_replay = ASC_REQ_ALLOW_CONTEXT_REPLAY
 *     fragment_to_fit = ASC_REQ_FRAGMENT_TO_FIT             # <<<<<<<<<<<<<<
 *     no_token = ASC_REQ_NO_TOKEN
 *     proxy_bindings = ASC_REQ_PROXY_BINDINGS
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_FRAGMENT_TO_FIT); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_fragment_to_fit, __pyx_t_2) < 0) __PYX_ERR(1, 366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":367
 *     allow_context_replay = ASC_REQ_ALLOW_CONTEXT_REPLAY
 *     fragment_to_fit = ASC_REQ_FRAGMENT_TO_FIT
 *     no_token = ASC_REQ_NO_TOKEN             # <<<<<<<<<<<<<<
 *     proxy_bindings = ASC_REQ_PROXY_BINDINGS
 *     allow_missing_binding = ASC_REQ_ALLOW_MISSING_BINDINGS
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_NO_TOKEN); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_no_token, __pyx_t_2) < 0) __PYX_ERR(1, 367, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":368
 *     fragment_to_fit = ASC_REQ_FRAGMENT_TO_FIT
 *     no_token = ASC_REQ_NO_TOKEN
 *     proxy_bindings = ASC_REQ_PROXY_BINDINGS             # <<<<<<<<<<<<<<
 *     allow_missing_binding = ASC_REQ_ALLOW_MISSING_BINDINGS
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_PROXY_BINDINGS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_proxy_bindings, __pyx_t_2) < 0) __PYX_ERR(1, 368, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":369
 *     no_token = ASC_REQ_NO_TOKEN
 *     proxy_bindings = ASC_REQ_PROXY_BINDINGS
 *     allow_missing_binding = ASC_REQ_ALLOW_MISSING_BINDINGS             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(ASC_REQ_ALLOW_MISSING_BINDINGS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_allow_missing_binding, __pyx_t_2) < 0) __PYX_ERR(1, 369, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":345
 * 
 * 
 * class ServerContextReq:             # <<<<<<<<<<<<<<
 *     delegate = ASC_REQ_DELEGATE
 *     mutual_auth = ASC_REQ_MUTUAL_AUTH
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_ServerContextReq, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ServerContextReq, __pyx_t_2) < 0) __PYX_ERR(1, 345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":372
 * 
 * 
 * class SSPIQoP:             # <<<<<<<<<<<<<<
 *     wrap_no_encrypt = SECQOP_WRAP_NO_ENCRYPT
 * 
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_SSPIQoP, __pyx_n_s_SSPIQoP, (PyObject *) NULL, __pyx_n_s_spnego__sspi_raw_sspi, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "spnego/_sspi_raw/sspi.pyx":373
 * 
 * class SSPIQoP:
 *     wrap_no_encrypt = SECQOP_WRAP_NO_ENCRYPT             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECQOP_WRAP_NO_ENCRYPT); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_wrap_no_encrypt, __pyx_t_2) < 0) __PYX_ERR(1, 373, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":372
 * 
 * 
 * class SSPIQoP:             # <<<<<<<<<<<<<<
 *     wrap_no_encrypt = SECQOP_WRAP_NO_ENCRYPT
 * 
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_SSPIQoP, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SSPIQoP, __pyx_t_2) < 0) __PYX_ERR(1, 372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":376
 * 
 * 
 * class TargetDataRep:             # <<<<<<<<<<<<<<
 *     native = SECURITY_NATIVE_DREP
 *     network = SECURITY_NETWORK_DREP
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_TargetDataRep, __pyx_n_s_TargetDataRep, (PyObject *) NULL, __pyx_n_s_spnego__sspi_raw_sspi, (PyObject *) NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "spnego/_sspi_raw/sspi.pyx":377
 * 
 * class TargetDataRep:
 *     native = SECURITY_NATIVE_DREP             # <<<<<<<<<<<<<<
 *     network = SECURITY_NETWORK_DREP
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECURITY_NATIVE_DREP); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_native, __pyx_t_2) < 0) __PYX_ERR(1, 377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":378
 * class TargetDataRep:
 *     native = SECURITY_NATIVE_DREP
 *     network = SECURITY_NETWORK_DREP             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_long(SECURITY_NETWORK_DREP); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_network, __pyx_t_2) < 0) __PYX_ERR(1, 378, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":376
 * 
 * 
 * class TargetDataRep:             # <<<<<<<<<<<<<<
 *     native = SECURITY_NATIVE_DREP
 *     network = SECURITY_NETWORK_DREP
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_TargetDataRep, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TargetDataRep, __pyx_t_2) < 0) __PYX_ERR(1, 376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":381
 * 
 * 
 * SecPkgAttrSizes = namedtuple('SecPkgAttrSizes', ['max_token', 'max_signature', 'block_size', 'security_trailer'])             # <<<<<<<<<<<<<<
 * SecPkgInfo = namedtuple('SecPkgInfo', ['capabilities', 'version', 'rpcid', 'max_token', 'name', 'comment'])
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_u_max_token);
  __Pyx_GIVEREF(__pyx_n_u_max_token);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_u_max_token);
  __Pyx_INCREF(__pyx_n_u_max_signature);
  __Pyx_GIVEREF(__pyx_n_u_max_signature);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_u_max_signature);
  __Pyx_INCREF(__pyx_n_u_block_size);
  __Pyx_GIVEREF(__pyx_n_u_block_size);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_u_block_size);
  __Pyx_INCREF(__pyx_n_u_security_trailer);
  __Pyx_GIVEREF(__pyx_n_u_security_trailer);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_u_security_trailer);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_u_SecPkgAttrSizes);
  __Pyx_GIVEREF(__pyx_n_u_SecPkgAttrSizes);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_SecPkgAttrSizes);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SecPkgAttrSizes, __pyx_t_2) < 0) __PYX_ERR(1, 381, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":382
 * 
 * SecPkgAttrSizes = namedtuple('SecPkgAttrSizes', ['max_token', 'max_signature', 'block_size', 'security_trailer'])
 * SecPkgInfo = namedtuple('SecPkgInfo', ['capabilities', 'version', 'rpcid', 'max_token', 'name', 'comment'])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_u_capabilities);
  __Pyx_GIVEREF(__pyx_n_u_capabilities);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_capabilities);
  __Pyx_INCREF(__pyx_n_u_version);
  __Pyx_GIVEREF(__pyx_n_u_version);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_u_version);
  __Pyx_INCREF(__pyx_n_u_rpcid);
  __Pyx_GIVEREF(__pyx_n_u_rpcid);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_u_rpcid);
  __Pyx_INCREF(__pyx_n_u_max_token);
  __Pyx_GIVEREF(__pyx_n_u_max_token);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_u_max_token);
  __Pyx_INCREF(__pyx_n_u_name_2);
  __Pyx_GIVEREF(__pyx_n_u_name_2);
  PyList_SET_ITEM(__pyx_t_3, 4, __pyx_n_u_name_2);
  __Pyx_INCREF(__pyx_n_u_comment);
  __Pyx_GIVEREF(__pyx_n_u_comment);
  PyList_SET_ITEM(__pyx_t_3, 5, __pyx_n_u_comment);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_SecPkgInfo);
  __Pyx_GIVEREF(__pyx_n_u_SecPkgInfo);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_SecPkgInfo);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SecPkgInfo, __pyx_t_3) < 0) __PYX_ERR(1, 382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":409
 * cdef class SecBufferDesc:
 * 
 *     def __cinit__(SecBufferDesc self, list buffers not None, unsigned long version=SECBUFFER_VERSION):             # <<<<<<<<<<<<<<
 *         self._c_value = NativeSecBufferDesc(version, <unsigned long>len(buffers), NULL)
 *         self._buffers = buffers
 */
  __pyx_k__5 = SECBUFFER_VERSION;

  /* "spnego/_sspi_raw/sspi.pyx":623
 * 
 * def accept_security_context(Credential credential not None, SecurityContext context not None,
 *     SecBufferDesc input_buffer=None, unsigned long context_req=0, unsigned long target_data_rep=SECURITY_NATIVE_DREP,             # <<<<<<<<<<<<<<
 *     SecBufferDesc output_buffer=None):
 * 
 */
  __pyx_k__16 = SECURITY_NATIVE_DREP;

  /* "spnego/_sspi_raw/sspi.pyx":622
 * 
 * 
 * def accept_security_context(Credential credential not None, SecurityContext context not None,             # <<<<<<<<<<<<<<
 *     SecBufferDesc input_buffer=None, unsigned long context_req=0, unsigned long target_data_rep=SECURITY_NATIVE_DREP,
 *     SecBufferDesc output_buffer=None):
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_1accept_security_context, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_accept_security_context, __pyx_t_3) < 0) __PYX_ERR(1, 622, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":652
 * 
 * def acquire_credentials_handle(unicode principal, unicode package not None,
 *     unsigned long credential_use=SECPKG_CRED_OUTBOUND, _AuthIdentityBase auth_data=None):             # <<<<<<<<<<<<<<
 * 
 *     cdef WideChar w_principal = WideChar.from_text(principal)
 */
  __pyx_k__17 = SECPKG_CRED_OUTBOUND;

  /* "spnego/_sspi_raw/sspi.pyx":651
 * 
 * 
 * def acquire_credentials_handle(unicode principal, unicode package not None,             # <<<<<<<<<<<<<<
 *     unsigned long credential_use=SECPKG_CRED_OUTBOUND, _AuthIdentityBase auth_data=None):
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_3acquire_credentials_handle, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_acquire_credentials_handle, __pyx_t_3) < 0) __PYX_ERR(1, 651, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":671
 * 
 * 
 * def decrypt_message(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):             # <<<<<<<<<<<<<<
 *     cdef unsigned long qop = 0;
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_5decrypt_message, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_decrypt_message, __pyx_t_3) < 0) __PYX_ERR(1, 671, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":681
 * 
 * 
 * def encrypt_message(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0,             # <<<<<<<<<<<<<<
 *     unsigned long qop=0):
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_7encrypt_message, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_encrypt_message, __pyx_t_3) < 0) __PYX_ERR(1, 681, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":691
 * def initialize_security_context(Credential credential not None, SecurityContext context not None,
 *     unicode target_name not None, unsigned long context_req=0, SecBufferDesc input_buffer=None,
 *     unsigned long target_data_rep=SECURITY_NATIVE_DREP, SecBufferDesc output_buffer=None):             # <<<<<<<<<<<<<<
 * 
 *     cdef PCtxtHandle input_context = &context.handle if context.init else NULL
 */
  __pyx_k__18 = SECURITY_NATIVE_DREP;

  /* "spnego/_sspi_raw/sspi.pyx":689
 * 
 * 
 * def initialize_security_context(Credential credential not None, SecurityContext context not None,             # <<<<<<<<<<<<<<
 *     unicode target_name not None, unsigned long context_req=0, SecBufferDesc input_buffer=None,
 *     unsigned long target_data_rep=SECURITY_NATIVE_DREP, SecBufferDesc output_buffer=None):
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_9initialize_security_context, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_initialize_security_context, __pyx_t_3) < 0) __PYX_ERR(1, 689, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":719
 * 
 * 
 * def make_signature(SecurityContext context not None, unsigned long qop, SecBufferDesc message not None,             # <<<<<<<<<<<<<<
 *     unsigned long seq_no=0):
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_11make_signature, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_make_signature, __pyx_t_3) < 0) __PYX_ERR(1, 719, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":727
 * 
 * 
 * def query_context_attributes(SecurityContext context not None, unsigned long attribute):             # <<<<<<<<<<<<<<
 *     if attribute not in [SecPkgAttr.names, SecPkgAttr.package_info, SecPkgAttr.session_key, SecPkgAttr.sizes]:
 *         raise NotImplementedError("Only names, package_info, session_key, or sizes is implemented")
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_13query_context_attributes, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_query_context_attributes, __pyx_t_3) < 0) __PYX_ERR(1, 727, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":744
 * 
 * 
 * def _query_context_names(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_Names info
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_15_query_context_names, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_query_context_names, __pyx_t_3) < 0) __PYX_ERR(1, 744, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":756
 *         FreeContextBuffer(<void*>info.sUserName)
 * 
 * def _query_context_package_info(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_PackageInfoW raw_info
 *     cdef PSecPkgInfoW info
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_17_query_context_package_info, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_query_context_package_info, __pyx_t_3) < 0) __PYX_ERR(1, 756, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":772
 * 
 * 
 * def _query_context_session_key(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_SessionKey info
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_19_query_context_session_key, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_query_context_session_key, __pyx_t_3) < 0) __PYX_ERR(1, 772, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":785
 * 
 * 
 * def _query_context_sizes(SecurityContext context not None):             # <<<<<<<<<<<<<<
 *     cdef SecPkgContext_Sizes sizes
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_21_query_context_sizes, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_query_context_sizes, __pyx_t_3) < 0) __PYX_ERR(1, 785, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":795
 * 
 * 
 * def verify_signature(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):             # <<<<<<<<<<<<<<
 *     cdef unsigned long qop = 0
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_23verify_signature, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_verify_signature, __pyx_t_3) < 0) __PYX_ERR(1, 795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__AuthIdentityBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_6spnego_9_sspi_raw_4sspi_25__pyx_unpickle__AuthIdentityBase, NULL, __pyx_n_s_spnego__sspi_raw_sspi); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__AuthIdentityBase, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "spnego/_sspi_raw/sspi.pyx":1
 * # Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>             # <<<<<<<<<<<<<<
 * # MIT License (see LICENSE or https://opensource.org/licenses/MIT)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init spnego._sspi_raw.sspi", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init spnego._sspi_raw.sspi");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value) {
    const unsigned long neg_one = (unsigned long) ((unsigned long) 0 - (unsigned long) 1), const_zero = (unsigned long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_LONG(LONG value) {
    const LONG neg_one = (LONG) ((LONG) 0 - (LONG) 1), const_zero = (LONG) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(LONG),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value) {
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG) ((unsigned PY_LONG_LONG) 0 - (unsigned PY_LONG_LONG) 1), const_zero = (unsigned PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned PY_LONG_LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned PY_LONG_LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned PY_LONG_LONG),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_short(unsigned short value) {
    const unsigned short neg_one = (unsigned short) ((unsigned short) 0 - (unsigned short) 1), const_zero = (unsigned short) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned short) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned short) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned short) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned short) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned short) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned short),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *x) {
    const unsigned long neg_one = (unsigned long) ((unsigned long) 0 - (unsigned long) 1), const_zero = (unsigned long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned long, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 2 * PyLong_SHIFT) {
                            return (unsigned long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 3 * PyLong_SHIFT) {
                            return (unsigned long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 4 * PyLong_SHIFT) {
                            return (unsigned long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned long) ((((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned long) ((((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned long) ((((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned long) -1;
        }
    } else {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned long) -1;
        val = __Pyx_PyInt_As_unsigned_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned long");
    return (unsigned long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned long");
    return (unsigned long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
